#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//BITMAPFILEHEADER  ptr index
#define BF_TYPE 0
#define BF_SIZE 2
#define BF_RESERVED1 6
#define BF_RESERVED2 8
#define BF_OFFBITS 10

//WINBMPINFOHEADER  ptr index
#define BI_SIZE 14
#define BI_WIDTH 18
#define BI_HEIGHT 22
#define BI_PLANES 26
#define BI_BITCOUNT 28
#define BI_COMPRESSION 30
#define BI_SIZEIMAGE 34
#define BI_XPELSPERMETER 38
#define BI_YPELSPERMETER 42
#define BI_CLRUSED 46
#define BI_CLRIMPORTANT 50

//Palette entry  ptr index
#define BF_PALETTE 54

//DIB  ptr index
#define BF_DIBP8 1078
#define BF_DIBP24 54

#undef TRUE
#undef FALSE
#define TRUE 1
#define FALSE 0

static unsigned long realsize = 0;

static unsigned char *galloc(unsigned char *mptr, unsigned long cursize, unsigned long gsize)
{
	unsigned char *mmptr;
	if(!mptr)
		return NULL;
	if((cursize + gsize) <= realsize)
		return mptr;
	if(gsize <= 0x1000) {
		mmptr = (unsigned char *) malloc (realsize + 0x1000);
		realsize += 0x1000;
	}
	else {
		mmptr = (unsigned char *) malloc (realsize + gsize);
		realsize += gsize;
	}
	if(!mmptr) {
		free ((char *) mptr);
		return NULL;
	}
	memmove(mmptr, mptr, cursize);
	free ((char *) mptr);
	return mmptr;
}

void realsizeinit(unsigned long initsize)
{
	realsize = initsize;
}

static unsigned char CheckBitmapfileHeader(unsigned char *pBitmapImage)
{
	//Check bitmap file header.
	if(!pBitmapImage ||
		*(unsigned short *)(&pBitmapImage[BF_TYPE]) != 0x4D42 ||
		*(unsigned short *)(&pBitmapImage[BF_RESERVED1]) != 0 ||
		*(unsigned short *)(&pBitmapImage[BF_RESERVED2]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_SIZE]) != 40 ||
		*(unsigned short *)(&pBitmapImage[BI_PLANES]) != 1 ||
		*(unsigned long *)(&pBitmapImage[BI_COMPRESSION]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_XPELSPERMETER]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_YPELSPERMETER]) != 0 ||
		!((*(unsigned long *)(&pBitmapImage[BF_OFFBITS]) == 54 && *(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 24 &&
		*(unsigned long *)(&pBitmapImage[BI_CLRUSED]) == 0 && *(unsigned long *)(&pBitmapImage[BI_CLRIMPORTANT]) == 0) ||
		(*(unsigned long *)(&pBitmapImage[BF_OFFBITS]) == 1078 && *(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 8 &&
		*(unsigned long *)(&pBitmapImage[BI_CLRUSED]) == 256 && *(unsigned long *)(&pBitmapImage[BI_CLRIMPORTANT]) == 256)))
		return FALSE;
	return TRUE;
}

unsigned char *Bitmap2pcx(unsigned char *pBitmapImage, unsigned char *pcxpalette, unsigned long *ImageDataSize)
{
	/* lookup table for scaling 6 bit colors up to 8 bits */
	static unsigned char RgbScale6[64] =
	{
		0,   4,   8,   12,  16,  20,  24,  28,
		32,  36,  40,  44,  48,  52,  56,  60,
		64,  68,  72,  76,  80,  85,  89,  93,
		97,  101, 105, 109, 113, 117, 121, 125,
		129, 133, 137, 141, 145, 149, 153, 157,
		161, 165, 170, 174, 178, 182, 186, 190,
		194, 198, 202, 206, 210, 214, 218, 222,
		226, 230, 234, 238, 242, 246, 250, 255
	};

	int i , x, y, runcount;
	unsigned char runchar;
	unsigned char ch, putc;

	//Check bitmap file header.
	if(!CheckBitmapfileHeader(pBitmapImage) || !pcxpalette || !ImageDataSize) {
		return NULL;
	}

	unsigned char planes = ((*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 24) ? 3 : 1);

	unsigned char *pPCXImage = (unsigned char *) malloc (16 + 48 + 10 + 54);
	if(!pPCXImage)
		return NULL;

	unsigned short pcxWidth = (unsigned short)(*(unsigned long *)(&pBitmapImage[BI_WIDTH]));
	unsigned short pcxHeight = (unsigned short)(*(unsigned long *)(&pBitmapImage[BI_HEIGHT]));

	//PCX info header (16 byte)
	*(unsigned char *)(&pPCXImage[0]) = 10;//pcxiManufacturer
	*(unsigned char *)(&pPCXImage[1]) = 5;//pcxiVersion
	*(unsigned char *)(&pPCXImage[2]) = 1;//pcxiRunLengthEncoding
	*(unsigned char *)(&pPCXImage[3]) = 8;//pcxi8BitsPerPixel
	*(unsigned short *)(&pPCXImage[4]) = 0;//pcxiXmin
	*(unsigned short *)(&pPCXImage[6]) = 0;//pcxiYmin
	*(unsigned short *)(&pPCXImage[8]) = pcxWidth - 1;//pcxiXmax
	*(unsigned short *)(&pPCXImage[10]) = pcxHeight - 1;//pcxiYmax
	*(unsigned short *)(&pPCXImage[12]) = 320;//pcxiHDpi
	*(unsigned short *)(&pPCXImage[14]) = 200;//pcxiVDpi

	for (i = 0; i < 16; i++) {
		*(unsigned char *)(&pPCXImage[16 + i * 3]) = RgbScale6[pcxpalette[i * 4]];
		*(unsigned char *)(&pPCXImage[16 + i * 3 + 1]) = RgbScale6[pcxpalette[i * 4 + 1]];
		*(unsigned char *)(&pPCXImage[16 + i * 3 + 2]) = RgbScale6[pcxpalette[i * 4 + 2]];
	}

	//PCX info header (10 byte)
	*(unsigned char *)(&pPCXImage[64]) = 0;//pcxiReserved
	*(unsigned char *)(&pPCXImage[65]) = planes;//planes
	*(unsigned short *)(&pPCXImage[66]) = pcxWidth;//pcxiWidth
	*(unsigned short *)(&pPCXImage[68]) = 1;//pcxiColorPalette
	*(unsigned short *)(&pPCXImage[70]) = pcxWidth;//pcxiHscreen
	*(unsigned short *)(&pPCXImage[72]) = pcxHeight;//pcxiVscreen
	memset((void *)&pPCXImage[74], 0, 54);

	unsigned long iputc = 128;
	realsizeinit(128);

	for (y = 0; y < pcxHeight; y++) {
		runcount = 0;
		runchar = 0;
		for (x = 0; x < (pcxWidth * planes); x++) {
			if (*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 8)
				ch = pBitmapImage[BF_DIBP8 + (pcxHeight - 1 - y) * pcxWidth + x];
			else {
				if (x < pcxWidth)
					ch = pBitmapImage[BF_DIBP24 + (pcxHeight - 1 - y) * pcxWidth * 3 + x * 3 + 2];
				else if (x < (pcxWidth * 2))
					ch = pBitmapImage[BF_DIBP24 + (pcxHeight - 1 - y) * pcxWidth * 3 + (x - pcxWidth) * 3 + 1];
				else
					ch = pBitmapImage[BF_DIBP24 + (pcxHeight - 1 - y) * pcxWidth * 3 + (x - pcxWidth * 2) * 3];
			}
			if (runcount == 0) {
				runcount = 1;
				runchar = ch;
			}
			else {
				if ((ch != runchar) || (runcount >= 0x3f)) {
					if ((runcount > 1) || ((runchar & 0xC0) == 0xC0)) {
						putc = 0xC0 | runcount;
						if(!(pPCXImage = galloc(pPCXImage, iputc, 1)))
							return NULL;
						pPCXImage[iputc] = putc;
						++iputc;
					}
					putc = runchar;
					if(!(pPCXImage = galloc(pPCXImage, iputc, 1)))
						return NULL;
					pPCXImage[iputc] = putc;
					++iputc;
					runcount = 1;
					runchar = ch;
				}
				else
					runcount++;
			}
		}
		if ((runcount > 1) || ((runchar & 0xC0) == 0xC0)) {
			putc = 0xC0 | runcount;
			if(!(pPCXImage = galloc(pPCXImage, iputc, 1)))
				return NULL;
			pPCXImage[iputc] = putc;
			++iputc;
		}
		putc = runchar;
		if(!(pPCXImage = galloc(pPCXImage, iputc, 1)))
			return NULL;
		pPCXImage[iputc] = putc;
		++iputc;
	}

	if (*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 8) {
		putc = 12;
		if(!(pPCXImage = galloc(pPCXImage, iputc, 1)))
			return NULL;
		pPCXImage[iputc] = putc;
		++iputc;
		if(!(pPCXImage = galloc(pPCXImage, iputc, 768)))
			return NULL;
		for (i = 0; i < 256; i++) {
			pPCXImage[iputc + i * 3] = RgbScale6[pcxpalette[i * 4]];
			pPCXImage[iputc + i * 3 + 1] = RgbScale6[pcxpalette[i * 4 + 1]];
			pPCXImage[iputc + i * 3 + 2] = RgbScale6[pcxpalette[i * 4 + 2]];
		}
		iputc += 768;
	}
	*ImageDataSize = iputc;
	return pPCXImage;
}

/*****************************************************************************
 *
 * convert bmp2gif
 *                  return  pGIFfileimagedata, ImageDataSize
 *****************************************************************************/

static unsigned char cmap[256*4];	/* Output colormap */
static unsigned char *pbmpf = NULL;

static unsigned char GetPixel(int x, int y){
	return pbmpf[BF_DIBP8 + (*(unsigned long *)(&pbmpf[BI_WIDTH])) * (*(unsigned long *)(&pbmpf[BI_HEIGHT]) - 1 - y) + x];
}

static unsigned char *GetpGIFdataAndSize(unsigned long *ImageDataSize);

static void formoutputcmap(unsigned char *pBitmapImage);
typedef unsigned char (* ifunptr)(int, int);
static int GIFEncode(unsigned long GWidth, unsigned long GHeight,
	int GInterlace, int Background, int Transparent,
	int BitsPerPixel, unsigned char *gifpalette, ifunptr GetPixel, char *gifcomment);

unsigned char *Bitmap2gif(unsigned char *pBitmapImage, unsigned long *ImageDataSize, char *gifcomment)
{
	if(!CheckBitmapfileHeader(pBitmapImage) || *(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) != 8)
		return NULL;

	pbmpf = pBitmapImage;
	formoutputcmap(pBitmapImage);

	int GInterlace= 0;
	int Background = 0;
	int Transparent = -1;

	int gifErr = GIFEncode(*(unsigned long *)(&pBitmapImage[BI_WIDTH]), *(unsigned long *)(&pBitmapImage[BI_HEIGHT]),
		GInterlace, Background, Transparent,
		(int)(*(unsigned short *)(&pBitmapImage[BI_BITCOUNT])), cmap, GetPixel, gifcomment);
	if(!gifErr)
		return NULL;
	else
		return GetpGIFdataAndSize(ImageDataSize);
}

/*****************************************************************************
 *
 * GIFEncode(GHeight, GWidth, GInterlace, Background, Transparent,
 *	      BitsPerPixel, Palette, GetPixel, gifcomment)
 *
 *****************************************************************************/

static void formoutputcmap(unsigned char *pBitmapImage){
	for(int i = 0; i < 256; i++) {
		cmap[i * 3]= pBitmapImage[BF_PALETTE + 4 * i + 2];
		cmap[i * 3 + 1]= pBitmapImage[BF_PALETTE + 4 * i + 1];
		cmap[i * 3 + 2]= pBitmapImage[BF_PALETTE + 4 * i];
	}
}

static unsigned char *pwdata = NULL;
static unsigned long wdatasize = 0;

static unsigned char mem_write_data(unsigned char *prdata, unsigned long rdatasize, unsigned long rdatacount)
{
	if((wdatasize + rdatasize * rdatacount) > realsize) {
		unsigned char *newptr;
		if(rdatasize * rdatacount <= 0x1000) {
			newptr = (unsigned char *) malloc (realsize + 0x1000);
			realsize += 0x1000;
		}
		else {
			newptr = (unsigned char *) malloc (realsize + rdatasize * rdatacount);
			realsize += (rdatasize * rdatacount);
		}
		if(!newptr) {
			if(pwdata)
				free ((char *) pwdata);
			return FALSE;
		}
		if(pwdata) {
			memmove(newptr, pwdata, wdatasize);
			free ((char *) pwdata);
		}
		pwdata = newptr;
	}
	memmove(pwdata + wdatasize, prdata, rdatasize * rdatacount);
	wdatasize += (rdatasize * rdatacount);
	return TRUE;
}

static unsigned char *GetpGIFdataAndSize(unsigned long *ImageDataSize)
{
	*ImageDataSize = wdatasize;
	return pwdata;
}

static unsigned long Width, Height;
static unsigned long curx, cury;
static unsigned long CountDown;
static int Pass = 0;
static int Interlace;

static int compress ( int init_bits, ifunptr ReadValue );
static int output ( int code );
static int cl_block ( void );
static void cl_hash ( long int hsize );
static void char_init ( void );
static int char_out ( int c );
static int flush_char ( void );


/*
 * Bump the 'curx' and 'cury' to point to the next pixel
 */
static void BumpPixel() {
	/* Bump the current X position */
	++curx;

	/*
	 * If we are at the end of a scan line, set curx back to the beginning
	 * If we are interlaced, bump the cury to the appropriate spot,
	 * otherwise, just increment it.
	 */
	if( curx == Width ) {
		curx = 0;
		if( !Interlace )
			++cury;
		else {
		     switch( Pass ) {
		       case 0:
			  cury += 8;
			  if( cury >= Height ) {
				++Pass;
				cury = 4;
			  } break;
		       case 1:
			  cury += 8;
			  if( cury >= Height ) {
				++Pass;
				cury = 2;
			  } break;
		       case 2:
			  cury += 4;
			  if( cury >= Height ) {
			     ++Pass;
			     cury = 1;
			  } break;
		       case 3:
			  cury += 2;
			  break;
			}
		}
	}
}

/*
 * Return the next pixel from the image
 */
static int GIFNextPixel(ifunptr getpixel){
	int r;

	if( CountDown == 0 )
		return EOF;
	--CountDown;

	r = ( * getpixel )( curx, cury );
	BumpPixel();
	return r;
}

static void gifencinit();

/* public */
static int GIFEncode(unsigned long GWidth, unsigned long GHeight,
	int GInterlace, int Background, int Transparent,
	int BitsPerPixel, unsigned char *gifpalette, ifunptr GetPixel, char *gifcomment ){

	int B;
	unsigned long RWidth, RHeight;
	int LeftOfs, TopOfs;
	int Resolution;
	int ColorMapSize;
	int InitCodeSize;
	int i;

	unsigned char putc;
	unsigned short putw;
	unsigned long putl;
	char gifcomm[0xff + 1];
	if(gifcomment)
		strncpy(gifcomm, gifcomment, 0xff);
	else
		gifcomm[0] = 0;
	gifcomm[0xff] = 0;
	gifencinit();
	
	Interlace = GInterlace;

	ColorMapSize = 1 << BitsPerPixel;

	RWidth = Width = GWidth;
	RHeight = Height = GHeight;
	LeftOfs = TopOfs = 0;

	Resolution = BitsPerPixel;

	/* Calculate number of bits we are expecting */
	CountDown = Width * Height;

	/* Indicate which pass we are on (if interlace) */
	Pass = 0;

	/* The initial code size */
	if( BitsPerPixel <= 1 )
		InitCodeSize = 2;
	else
		InitCodeSize = BitsPerPixel;

	/* Set up the current x and y position */
	curx = cury = 0;

	realsizeinit(0);

	/* Write the Magic header */
	if(!mem_write_data((unsigned char *)(Transparent < 0 ? "GIF87a" : "GIF89a"), 1, 6))
		return FALSE;

	/* Write out the screen width and height */
	if(!mem_write_data((unsigned char *)(&RWidth), 2, 1))
		return FALSE;
	if(!mem_write_data((unsigned char *)(&RHeight), 2, 1))
		return FALSE;

	/* Indicate that there is a global colour map */
	B = 0x80;	/* Yes, there is a color map */

	/* OR in the resolution */
	B |= (Resolution - 1) << 5;

	/* OR in the Bits per Pixel */
	B |= (BitsPerPixel - 1);

	/* Write it out */
	if(!mem_write_data((unsigned char *)(&B), 1, 1))
		return FALSE;

	/* Write out the Background colour */
	if(!mem_write_data((unsigned char *)(&Background), 1, 1))
		return FALSE;

	/* Byte of 0's (future expansion) */
	putc = 0;
	if(!mem_write_data(&putc, 1, 1))
		return FALSE;

	/* Write out the Global Colour Map */
	for( i=0; i<ColorMapSize; ++i ) {
		if(!mem_write_data((unsigned char *)gifpalette, 1, 3))
			return FALSE;
		// Red
		// Green
		// Blue
		gifpalette += 3;
	}

	/* Write out extension for transparent colour index, if necessary. */
	if ( Transparent >= 0 ) {
		putc = '!';
		if(!mem_write_data(&putc, 1, 1))
			return FALSE;
		putc = 0xf9;
		if(!mem_write_data(&putc, 1, 1))
			return FALSE;
		putc = 0x4;
		if(!mem_write_data(&putc, 1, 1))
			return FALSE;
		putc = 1;
		if(!mem_write_data(&putc, 1, 1))
			return FALSE;
		putw = 0;
		if(!mem_write_data((unsigned char *)(&putw), 2, 1))
			return FALSE;
		if(!mem_write_data((unsigned char *)(&Transparent), 1, 1))
			return FALSE;
		putc = 0;
		if(!mem_write_data(&putc, 1, 1))
			return FALSE;
	}

	if (gifcomm[0]){
		putc = '!';
		if(!mem_write_data(&putc, 1, 1))
			return FALSE;
		putc = 0xfe;
		if(!mem_write_data(&putc, 1, 1))
			return FALSE;
		putl = (unsigned long)strlen(gifcomm);
		if(putl > 0xff)
			putl = 0xff;
		putc = (unsigned char)(putl & 0xff);
		if(!mem_write_data(&putc, 1, 1))
			return FALSE;
		if(!mem_write_data((unsigned char *)gifcomm, 1, (unsigned long)(putc + 1)))
			return FALSE;
	}

	/* Write an Image separator */
	putc = ',';
	if(!mem_write_data(&putc, 1, 1))
		return FALSE;

	/* Write the Image header */
	if(!mem_write_data((unsigned char *)(&LeftOfs), 2, 1))
		return FALSE;
	if(!mem_write_data((unsigned char *)(&TopOfs), 2, 1))
		return FALSE;
	if(!mem_write_data((unsigned char *)(&Width), 2, 1))
		return FALSE;
	if(!mem_write_data((unsigned char *)(&Height), 2, 1))
		return FALSE;

	/* Write out whether or not the image is interlaced */
	if( Interlace ) {
		putc = 0x40;
		if(!mem_write_data(&putc, 1, 1))
			return FALSE;
	}
	else {
		putc = 0;
		if(!mem_write_data(&putc, 1, 1))
			return FALSE;
	}

	/* Write out the initial code size */
	if(!mem_write_data((unsigned char *)(&InitCodeSize), 1, 1))
		return FALSE;

	/* Go and actually compress the data */
	if(!compress( InitCodeSize+1, GetPixel ))
		return FALSE;

	/* Write out a Zero-length packet (to end the series) */
	putc = 0;
	if(!mem_write_data(&putc, 1, 1))
		return FALSE;

	/* Write the GIF file terminator */
	putc = ';';
	if(!mem_write_data(&putc, 1, 1))
		return FALSE;

	return TRUE;
}

/***************************************************************************
 *
 *  GIFCOMPR.C	     - GIF Image compression routines
 *
 *  Lempel-Ziv compression based on 'compress'.  GIF modifications by
 *  David Rowley (mgardi@watdcsu.waterloo.edu)
 *
 ***************************************************************************/

/* General DEFINEs  */

#define BITS	12
#define HSIZE  5003	       /* 80% occupancy */

/*
 *
 * GIF Image compression - modified 'compress'
 *
 * Based on: compress.c - File compression ala IEEE Computer, June 1984.
 *
 * By Authors:	Spencer W. Thomas	(decvax!harpo!utah-cs!utah-gr!thomas)
 *		Jim McKie		(decvax!mcvax!jim)
 *		Steve Davies		(decvax!vax135!petsd!peora!srd)
 *		Ken Turkowski		(decvax!decwrl!turtlevax!ken)
 *		James A. Woods		(decvax!ihnp4!ames!jaw)
 *		Joe Orost		(decvax!vax135!petsd!joe)
 *
 */

static int n_bits;			  /* number of bits/code */
static int maxbits = BITS;		  /* user settable max # bits/code */
static int maxcode;		  /* maximum code, given n_bits */
static int maxmaxcode = (int)1 << BITS; /* should NEVER generate this code */
#ifdef COMPATIBLE		/* But wrong! */
# define MAXCODE(n_bits)	((int) 1 << (n_bits) - 1)
#else /*COMPATIBLE*/
# define MAXCODE(n_bits)	(((int) 1 << (n_bits)) - 1)
#endif /*COMPATIBLE*/

static long int htab [HSIZE];
static unsigned short codetab [HSIZE];
#define HashTabOf(i)	   htab[i]
#define CodeTabOf(i)	codetab[i]

static int hsize = HSIZE;		       /* for dynamic table sizing */

/*
 * To save much memory, we overlay the table used by compress() with those
 * used by decompress().  The tab_prefix table is the same size and type
 * as the codetab.  The tab_suffix table needs 2**BITS characters.  We
 * get this from the beginning of htab.  The output stack uses the rest
 * of htab, and contains characters.  There is plenty of room for any
 * possible stack (stack used to be 8000 characters).
 */

static int free_ent = 0;		       /* first unused entry */

/*
 * block compression parameters -- after all codes are used up,
 * and compression rate changes, start over.
 */
static int clear_flg = 0;

static int offset;
static long int in_count = 1;		 /* length of input */
static long int out_count = 0;		 /* # of codes output (for debugging) */

/*
 * compress stdin to stdout
 *
 * Algorithm:  use open addressing double hashing (no chaining) on the
 * prefix code / next character combination.  We do a variant of Knuth's
 * algorithm D (vol. 3, sec. 6.4) along with G. Knott's relatively-prime
 * secondary probe.  Here, the modular division first probe is gives way
 * to a faster exclusive-or manipulation.  Also do block compression with
 * an adaptive reset, whereby the code table is cleared when the compression
 * ratio decreases, but after the table fills.	The variable-length output
 * codes are re-sized at this point, and a special CLEAR code is generated
 * for the decompressor.  Late addition:  construct the table according to
 * file size for noticeable speed improvement on small files.  Please direct
 * questions about this implementation to ames!jaw.
 */

static int g_init_bits;

static int ClearCode;
static int EOFCode;

static int compress(int init_bits, ifunptr ReadValue ){
    register long fcode;
    register int i /* = 0 */;
    register int c;
    register int ent;
    register int disp;
    register int hsize_reg;
    register int hshift;

    /*
     * Set up the globals:  g_init_bits - initial number of bits
     *			    g_outfile	- pointer to output file
     */
    g_init_bits = init_bits;

    /*
     * Set up the necessary values
     */
    offset = 0;
    out_count = 0;
    clear_flg = 0;
    in_count = 1;
    maxcode = MAXCODE(n_bits = g_init_bits);

    ClearCode = (1 << (init_bits - 1));
    EOFCode = ClearCode + 1;
    free_ent = ClearCode + 2;

    char_init();

    ent = GIFNextPixel( ReadValue );

    hshift = 0;
    for ( fcode = (long) hsize;  fcode < 65536L; fcode *= 2L )
	++hshift;
    hshift = 8 - hshift;		/* set hash code range bound */

    hsize_reg = hsize;
    cl_hash( (long int) hsize_reg);		/* clear hash table */

    if(!output( (int)ClearCode ))
    	return FALSE;

    while ( (c = GIFNextPixel( ReadValue )) != EOF ) {	/* } */

	++in_count;

	fcode = (long) (((long) c << maxbits) + ent);
	i = (((int)c << hshift) ^ ent);	/* xor hashing */

	if ( HashTabOf (i) == fcode ) {
	    ent = CodeTabOf (i);
	    continue;
	} else if ( (long)HashTabOf (i) < 0 )	   /* empty slot */
	    goto nomatch;
	disp = hsize_reg - i;		/* secondary hash (after G. Knott) */
	if ( i == 0 )
	    disp = 1;
probe:
	if ( (i -= disp) < 0 )
	    i += hsize_reg;

	if ( HashTabOf (i) == fcode ) {
	    ent = CodeTabOf (i);
	    continue;
	}
	if ( (long)HashTabOf (i) > 0 )
	    goto probe;
nomatch:
	if(!output ( (int) ent ))
		return FALSE;
	++out_count;
	ent = c;
	if ( free_ent < maxmaxcode ) {	/* } */
	    CodeTabOf (i) = free_ent++; /* code -> hashtable */
	    HashTabOf (i) = fcode;
	} 
	else {
		if(!cl_block())
			return FALSE;
	}
    }
    /*
     * Put out the final code.
     */
    if(!output( (int)ent ))
    	return FALSE;
    ++out_count;
    if(!output( (int) EOFCode ))
    	return FALSE;
	return TRUE;
}

/*****************************************************************
 * TAG( output )
 *
 * Output the given code.
 * Inputs:
 *	code:	A n_bits-bit integer.  If == -1, then EOF.  This assumes
 *		that n_bits =< (long)wordsize - 1.
 * Outputs:
 *	Outputs code to the file.
 * Assumptions:
 *	Chars are 8 bits long.
 * Algorithm:
 *	Maintain a BITS character long buffer (so that 8 codes will
 * fit in it exactly).	Use the VAX insv instruction to insert each
 * code in turn.  When the buffer fills up empty it and start over.
 */

static unsigned long cur_accum = 0;
static int cur_bits = 0;

static unsigned long masks[] = { 0x0000, 0x0001, 0x0003, 0x0007, 0x000F,
				  0x001F, 0x003F, 0x007F, 0x00FF,
				  0x01FF, 0x03FF, 0x07FF, 0x0FFF,
				  0x1FFF, 0x3FFF, 0x7FFF, 0xFFFF };

static int output(int code )
{
    cur_accum &= masks[ cur_bits ];

    if( cur_bits > 0 )
	cur_accum |= ((long)code << cur_bits);
    else
	cur_accum = code;

    cur_bits += n_bits;

    while( cur_bits >= 8 ) {
	if(!char_out( (unsigned int)(cur_accum & 0xff) ))
		return FALSE;
	cur_accum >>= 8;
	cur_bits -= 8;
    }

    /*
     * If the next entry is going to be too big for the code size,
     * then increase it, if possible.
     */
   if ( free_ent > maxcode || clear_flg ) {

	    if( clear_flg ) {

		maxcode = MAXCODE (n_bits = g_init_bits);
		clear_flg = 0;

	    } else {

		++n_bits;
		if ( n_bits == maxbits )
		    maxcode = maxmaxcode;
		else
		    maxcode = MAXCODE(n_bits);
	    }
	}

    if( code == EOFCode ) {
	/*
	 * At EOF, write the rest of the buffer.
	 */
		while( cur_bits > 0 ) {
			if(!char_out( (unsigned int)(cur_accum & 0xff) ))
				return FALSE;
			cur_accum >>= 8;
			cur_bits -= 8;
		}
	
		if(!flush_char())
			return FALSE;
	}

	return TRUE;
}

/*
 * Clear out the hash table
 */
static int cl_block ()		/* table clear for block compress */
{

	cl_hash ( (long int) hsize );
	free_ent = ClearCode + 2;
	clear_flg = 1;

	if(!output( (int)ClearCode ))
		return FALSE;
	return TRUE;
}

static void
cl_hash(long int hsize)		/* reset code table */
{

	register long int *htab_p = htab+hsize;

	register long i;
	register long m1 = -1;

	i = hsize - 16;
	do {				/* might use Sys V memset(3) here */
		*(htab_p-16) = m1;
		*(htab_p-15) = m1;
		*(htab_p-14) = m1;
		*(htab_p-13) = m1;
		*(htab_p-12) = m1;
		*(htab_p-11) = m1;
		*(htab_p-10) = m1;
		*(htab_p-9) = m1;
		*(htab_p-8) = m1;
		*(htab_p-7) = m1;
		*(htab_p-6) = m1;
		*(htab_p-5) = m1;
		*(htab_p-4) = m1;
		*(htab_p-3) = m1;
		*(htab_p-2) = m1;
		*(htab_p-1) = m1;
		htab_p -= 16;
	} while ((i -= 16) >= 0);

	for ( i += 16; i > 0; --i )
		*--htab_p = m1;
}

/******************************************************************************
 *
 * GIF Specific routines
 *
 ******************************************************************************/

/*
 * Number of characters so far in this 'packet'
 */
static int a_count;

/*
 * Set up the 'byte output' routine
 */
static void char_init(){
	a_count = 0;
}

/*
 * Define the storage for the packet accumulator
 */
static char accum[ 256 ];

static void
gifencinit(){
	pwdata = NULL;
	wdatasize = 0;
	cur_accum = 0;
	cur_bits = 0;
	for(int i = 0; i < 256; i++)
		accum[i] = 0;
}

/*
 * Add a character to the end of the current packet, and if it is 254
 * characters, flush the packet to disk.
 */
static int
char_out( int c ) {
	accum[ a_count++ ] = c;
	if( a_count >= 254 ) {
		if(!flush_char())
			return FALSE;
	}
	return TRUE;
}
/*
 * Flush the packet to disk, and reset the accumulator
 */
static int flush_char(){
	if( a_count > 0 ) {
		if(!mem_write_data((unsigned char *)(&a_count), 1, 1))
			return FALSE;
		if(!mem_write_data((unsigned char *)accum, 1, (unsigned long)a_count))
			return FALSE;
		a_count = 0;
	}
	return TRUE;
}
