#include "../snes9x/snes9x.h"
#include "../snes9x/ppu.h"

#define MULBRIGHTNESS 140

extern int S9xCheckPixform();
static unsigned char S9xWriteBMP(FILE *fp);
static unsigned char S9xWritePCX(FILE *fp);
static unsigned char S9xWriteGIF(FILE *fp);
static unsigned char S9xWritePNG(FILE *fp);

unsigned char S9xSaveScreenImage(char *filename)
{
	//Check Resolution
	if(!(IPPU.RenderedScreenWidth == 256 && IPPU.RenderedScreenHeight == 224)
		&& !(IPPU.RenderedScreenWidth == 256 && IPPU.RenderedScreenHeight == 239)
		&& !(IPPU.RenderedScreenWidth == 256 && IPPU.RenderedScreenHeight == 448)
		&& !(IPPU.RenderedScreenWidth == 256 && IPPU.RenderedScreenHeight == 478)
		&& !(IPPU.RenderedScreenWidth == 512 && IPPU.RenderedScreenHeight == 224)
		&& !(IPPU.RenderedScreenWidth == 512 && IPPU.RenderedScreenHeight == 239)
		&& !(IPPU.RenderedScreenWidth == 512 && IPPU.RenderedScreenHeight == 448)
		&& !(IPPU.RenderedScreenWidth == 512 && IPPU.RenderedScreenHeight == 478)
		)
		return FALSE;

	//Check ext
	static char extstr[_MAX_EXT];
	char *extptr = strrchr((const char *)filename, '.');
	if(extptr && strlen(extptr) < _MAX_EXT)
		strcpy(extstr, extptr);
	else
		return FALSE;
	strlwr(extstr);

	//File open
	FILE *fp;
	fp = fopen(filename, "wb");
	if(!fp)
		return FALSE;

	unsigned char Err;
	//Select output format
	if(!strcmp(extstr, ".pcx"))
		Err = S9xWritePCX(fp);
	else if(!strcmp(extstr, ".gif"))
		Err = S9xWriteGIF(fp);
	else if(!strcmp(extstr, ".png"))
		Err = S9xWritePNG(fp);
	else
		Err = S9xWriteBMP(fp);
	if(!Err)
		remove(filename);
	return Err;
}

static int dstiR;
static int dstiG;
static int dstiB;
static int SecondShift;
static int ThirdShift;

void GetRgbPixformForWriteImageFile()
{
	//Check pixel format
	int pixform = S9xCheckPixform();

	if(pixform == RGB565) {
		dstiR = 2;
		dstiG = 1;
		dstiB = 0;
		SecondShift = 3;
		ThirdShift = 8;
	}
	else if(pixform == RGB555) {
		dstiR = 2;
		dstiG = 1;
		dstiB = 0;
		SecondShift = 2;
		ThirdShift = 7;
	}
	else if(pixform == BGR565) {
		dstiR = 0;
		dstiG = 1;
		dstiB = 2;
		SecondShift = 3;
		ThirdShift = 8;
	}
	else if(pixform == BGR555) {
		dstiR = 0;
		dstiG = 1;
		dstiB = 2;
		SecondShift = 2;
		ThirdShift = 7;
	}
	else if(pixform == GBR565) {
		dstiR = 1;
		dstiG = 0;
		dstiB = 2;
		SecondShift = 3;
		ThirdShift = 8;
	}
	else if(pixform == GBR555) {
		dstiR = 1;
		dstiG = 0;
		dstiB = 2;
		SecondShift = 2;
		ThirdShift = 7;
	}
	else {
		dstiR = 2;
		dstiG = 1;
		dstiB = 0;
		SecondShift = 3;
		ThirdShift = 8;
	}
}

//BITMAPFILEHEADER  ptr index
#define BF_TYPE 0
#define BF_SIZE 2
#define BF_RESERVED1 6
#define BF_RESERVED2 8
#define BF_OFFBITS 10

//WINBMPINFOHEADER  ptr index
#define BI_SIZE 14
#define BI_WIDTH 18
#define BI_HEIGHT 22
#define BI_PLANES 26
#define BI_BITCOUNT 28
#define BI_COMPRESSION 30
#define BI_SIZEIMAGE 34
#define BI_XPELSPERMETER 38
#define BI_YPELSPERMETER 42
#define BI_CLRUSED 46
#define BI_CLRIMPORTANT 50

//Palette entry  ptr index
#define BF_PALETTE 54

//DIB  ptr index
#define BF_DIBP8 1078
#define BF_DIBP24 54

extern unsigned char *ConvertBitmapImage24To8(unsigned char *pBitmapImage);
extern unsigned char *ConvertBitmapImage24To8IfWithin(unsigned char *pBitmapImage);

extern unsigned char *x2wBitmapImage(unsigned char *pBitmapImage);
extern unsigned char *x2hBitmapImage(unsigned char *pBitmapImage);

static unsigned char *S9xCreateBitmapImage(unsigned char Force8)
{
	unsigned char *pbmp = (unsigned char *) malloc (14 + 40 + (0) + (IPPU.RenderedScreenWidth * IPPU.RenderedScreenHeight * 3));
	if(!pbmp)
		return NULL;

	int SrcPixelByteSize = (2);
	int DstPixelByteSize = (3);

	//BITMAPFILEHEADER (14 byte)
	*(unsigned short *)(&pbmp[BF_TYPE]) = 0x4D42;
	*(unsigned long *)(&pbmp[BF_SIZE]) = 54 + (0) + (IPPU.RenderedScreenWidth * IPPU.RenderedScreenHeight * DstPixelByteSize);
	*(unsigned short *)(&pbmp[BF_RESERVED1]) = 0;
	*(unsigned short *)(&pbmp[BF_RESERVED2]) = 0;
	*(unsigned long *)(&pbmp[BF_OFFBITS]) = (54);

	//WINBMPINFOHEADER (40 byte)
	*(unsigned long *)(&pbmp[BI_SIZE]) = 40;
	*(unsigned long *)(&pbmp[BI_WIDTH]) = IPPU.RenderedScreenWidth;
	*(unsigned long *)(&pbmp[BI_HEIGHT]) = IPPU.RenderedScreenHeight;
	*(unsigned short *)(&pbmp[BI_PLANES]) = 1;
	*(unsigned short *)(&pbmp[BI_BITCOUNT]) = 24;
	*(unsigned long *)(&pbmp[BI_COMPRESSION]) = 0;
	*(unsigned long *)(&pbmp[BI_SIZEIMAGE]) = IPPU.RenderedScreenWidth * IPPU.RenderedScreenHeight * DstPixelByteSize;
	*(unsigned long *)(&pbmp[BI_XPELSPERMETER]) = 0;
	*(unsigned long *)(&pbmp[BI_YPELSPERMETER]) = 0;
	*(unsigned long *)(&pbmp[BI_CLRUSED]) = (0);
	*(unsigned long *)(&pbmp[BI_CLRIMPORTANT]) = (0);

	unsigned char *pScreenBuffer = &GFX.Screen[GFX.Pitch * (IPPU.RenderedScreenHeight - 1)];

	int dibpinc = 0;

	{
		//Create palette
		unsigned short Brightness = IPPU.MaxBrightness * MULBRIGHTNESS;
		for (int i = 0; i < 256; i++)
		{
			pbmp[BF_PALETTE + i * 4] = (((PPU.CGDATA [i] >> 10) & 0x1F) * Brightness) >> 8;
			pbmp[BF_PALETTE + i * 4 + 1] = (((PPU.CGDATA [i] >>  5) & 0x1F) * Brightness) >> 8;
			pbmp[BF_PALETTE + i * 4 + 2] = (((PPU.CGDATA [i] >>  0) & 0x1F) * Brightness) >> 8;
			pbmp[BF_PALETTE + i * 4 + 3] = 0;
		}
		for(int Height = IPPU.RenderedScreenHeight; Height; Height--, pScreenBuffer -= (GFX.Pitch + IPPU.RenderedScreenWidth * SrcPixelByteSize))
			for(int Width = IPPU.RenderedScreenWidth; Width; Width--, pScreenBuffer += SrcPixelByteSize, dibpinc += DstPixelByteSize)
				pbmp[BF_DIBP8 + dibpinc] = *pScreenBuffer;
	}

	if((Force8 || Options.ReduceTheColorTo256) && *(unsigned short *)(&pbmp[BI_BITCOUNT]) == 24)
//		unsigned char *pbmp = ConvertBitmapImage24To8(pbmp);
		pbmp = ConvertBitmapImage24To8IfWithin(pbmp);
	else if(!Force8 && Options.ReduceTheColorTo256IfWithin && *(unsigned short *)(&pbmp[BI_BITCOUNT]) == 24)
		pbmp = ConvertBitmapImage24To8IfWithin(pbmp);

	if(Options.ImagefileHiresStretch && pbmp && IPPU.RenderedScreenWidth == 512 && IPPU.RenderedScreenHeight < 448)
		pbmp = x2hBitmapImage(pbmp);
	else if(Options.ImagefileHiresStretch && pbmp && IPPU.RenderedScreenWidth == 256 && IPPU.RenderedScreenHeight >= 448)
		pbmp = x2wBitmapImage(pbmp);
	return pbmp;
}

extern unsigned char *bmp_compress_dib_rle(unsigned char *pBitmapImage);

static unsigned char S9xWriteBMP(FILE *fp)
{
	if(!fp)
		return FALSE;

	unsigned char *pBitmapImage = S9xCreateBitmapImage(FALSE);
	if(!pBitmapImage) {
		fclose(fp);
		return FALSE;
	}

	//RLE compression, 8bit color only
	unsigned char *pBitmapImageRLE;
	if(Options.RLEcompression256bitmap && (*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 8)) {
		pBitmapImageRLE = bmp_compress_dib_rle(pBitmapImage);
		if(pBitmapImageRLE) {
			free ((char *) pBitmapImage);
			pBitmapImage = pBitmapImageRLE;
		}
	}

	int Err = fwrite(pBitmapImage, *(unsigned long *)(&pBitmapImage[BF_SIZE]), 1, fp);
	fclose(fp);
	free ((char *) pBitmapImage);
	if(Err)
		return TRUE;
	else
		return FALSE;
}

extern unsigned char *Bitmap2pcx(unsigned char *pBitmapImage, unsigned char *pcxpalette, unsigned long *ImageDataSize);

static unsigned char S9xWritePCX(FILE *fp)
{
	if(!fp)
		return FALSE;
	unsigned char *pBitmapImage = S9xCreateBitmapImage(FALSE);
	if(!pBitmapImage) {
		fclose(fp);
		return FALSE;
	}
	//Create PCX palette
	unsigned char pcxpalette[1024];
	if(*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 8) {
		for (int i = 0; i < 256; i++)
		{
			pcxpalette[i * 4 + 2] = pBitmapImage[BF_PALETTE + i * 4] >> 2;
			pcxpalette[i * 4 + 1] = pBitmapImage[BF_PALETTE + i * 4 + 1] >> 2;
			pcxpalette[i * 4] = pBitmapImage[BF_PALETTE + i * 4 + 2] >> 2;
			pcxpalette[i * 4 + 3] = 0;
		}
	}
	else {
		unsigned short Brightness = IPPU.MaxBrightness * MULBRIGHTNESS;
		for (int i = 0; i < 256; i++)
		{
			pcxpalette[i * 4] = (((PPU.CGDATA [i] >>  0) & 0x1F) * Brightness) >> 10;
			pcxpalette[i * 4 + 1] = (((PPU.CGDATA [i] >>  5) & 0x1F) * Brightness) >> 10;
			pcxpalette[i * 4 + 2] = (((PPU.CGDATA [i] >> 10) & 0x1F) * Brightness) >> 10;
			pcxpalette[i * 4 + 3] = 0;
		}
	}

	unsigned long ImageDataSize;
	unsigned char *pPCXImage = Bitmap2pcx(pBitmapImage, pcxpalette, &ImageDataSize);
	free ((char *) pBitmapImage);

	if(!pPCXImage) {
		fclose(fp);
		return FALSE;
	}
	int Err = fwrite((const void *)pPCXImage, ImageDataSize, 1, fp);
	fclose(fp);
       free ((char *) pPCXImage);
	if(Err)
		return TRUE;
	else
		return FALSE;
}

extern unsigned char *Bitmap2gif(unsigned char *pBitmapImage, unsigned long *ImageDataSize, char *gifcomment);

static unsigned char S9xWriteGIF(FILE *fp)
{
	if(!fp)
		return FALSE;
	unsigned char *pBitmapImage = S9xCreateBitmapImage(TRUE);
	if(!pBitmapImage) {
		fclose(fp);
		return FALSE;
	}
	if(*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) != 8) {
		fclose(fp);
		free ((char *) pBitmapImage);
		return FALSE;
	}
	unsigned long ImageDataSize;
	unsigned char *pGIFImage = Bitmap2gif(pBitmapImage, &ImageDataSize, "Generated by Snes9xpp");
	free ((char *) pBitmapImage);

	if(!pGIFImage) {
		fclose(fp);
		return FALSE;
	}
	int Err = fwrite((const void *)pGIFImage, ImageDataSize, 1, fp);
	fclose(fp);
	free ((char *) pGIFImage);
	if(Err)
		return TRUE;
	else
		return FALSE;
}

unsigned char bmp2png(unsigned char *pBitmapImage, FILE *fp, char *png_comment_key_str, char *png_comment_text_str);

static unsigned char S9xWritePNG(FILE *fp)
{
	if(!fp)
		return FALSE;
	unsigned char *pBitmapImage = S9xCreateBitmapImage(FALSE);
	if(!pBitmapImage) {
		fclose(fp);
		return FALSE;
	}

	unsigned char pngErr = bmp2png(pBitmapImage, fp, "Emulator", "Snes9xpp");
	fclose(fp);
	free ((char *) pBitmapImage);
	return pngErr;
}
