#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lha/lhaccess.h"

#ifndef __WIN32__
typedef int int32;
#else
typedef long int32;
#endif

#undef TRUE
#undef FALSE
#define TRUE 1
#define FALSE 0

//#define MAXFILESIZE 0x600200
#define MAXFILESIZE 0x800200

unsigned char LoadLha(unsigned char *databuf, const char *archname, int32 *TotalFileSize,
                      int32 *headers, unsigned char ForceNoHeader, unsigned char ForceHeader);
unsigned char LoadCab(unsigned char *databuf, const char *archname, int32 *TotalFileSize,
                      int32 *headers, unsigned char ForceNoHeader, unsigned char ForceHeader);

unsigned char LoadArch(unsigned char *databuf, const char *archname, int32 *TotalFileSize,
                       int32 *headers, unsigned char ForceNoHeader, unsigned char ForceHeader)
{
	//Check ext
	static char extstr[1024];
	char *extptr = strrchr((const char *)archname, '.');
	if(extptr && strlen(extptr) < 1024)
		strcpy(extstr, extptr);
	else
		return FALSE;
	strlwr(extstr);

	*TotalFileSize = 0;
	*headers = 0;

	//Select output format
	if(!strcmp(extstr, ".lzh"))
		return LoadLha(databuf, archname, TotalFileSize, headers, ForceNoHeader, ForceHeader);
	else if(!strcmp(extstr, ".cab"))
		return LoadCab(databuf, archname, TotalFileSize, headers, ForceNoHeader, ForceHeader);
	else
		return FALSE;
}

unsigned char LoadLha(unsigned char *databuf, const char *archname, int32 *TotalFileSize,
                      int32 *headers, unsigned char ForceNoHeader, unsigned char ForceHeader)
{
	ROOT root;
	int errret;
	if(islzh((char *)archname))
		return FALSE;
	errret = open_root((char *)archname, &root);
	if(errret) {
		if(errret == 2)
			close_root( &root );
		return FALSE;
	}
	if(get_all_header(&root)) {
		close_root( &root );
		return FALSE;
	}
	int filenum;
	for(filenum = 1; filenum <= root.number; filenum++) {
		if(lzh_fopen(&root , filenum)) {
			close_root( &root );
			return FALSE;
		}
		if(((root.now->filesize % 0x2000) != 512 && (root.now->filesize % 0x2000) != 0) ||
			((root.now->filesize % 0x2000) == 512 && root.now->filesize < 0x8200) ||
			((root.now->filesize % 0x2000) == 0 && root.now->filesize < 0x8000))
			continue;
		else
			break;
	}
	if(filenum > root.number || root.now->filesize > MAXFILESIZE) {
		close_root( &root );
		return FALSE;
	}
	long readbyte = lzh_fread(databuf , (long)(root.now->filesize));
	if(readbyte != (long)(root.now->filesize)) {
		close_root( &root );
		return FALSE;
	}
	long calc_size = readbyte / 0x2000;
	calc_size *= 0x2000;
	if ((readbyte - calc_size == 512 && !ForceNoHeader) || ForceHeader)
	{
		memmove (databuf, databuf + 512, calc_size);
		(*headers)++;
		readbyte -= 512;
	}
	(*TotalFileSize) += (int32)readbyte;
	close_root( &root );
	return TRUE;
}

#include <sys/types.h>
#include "cabexlib.h"

unsigned char LoadCab(unsigned char *databuf, const char *archname, int32 *TotalFileSize,
                      int32 *headers, unsigned char ForceNoHeader, unsigned char ForceHeader)
{
	struct cabinet *basecab, *cab;
	struct file *filelist, *fi;
	char *fullname, *tail;
	ULONG calc_size;
	unsigned char loadcabresult = FALSE;
	basecab = load_cab((char *)archname);
	if (!basecab)
		return FALSE;
	if (basecab->flags & cfheadPREV_CABINET) {
		unload_cab(basecab);
		return FALSE;
	}

	tail = rindex(archname, SLASH_CHAR);

	for (cab = basecab; cab; cab = cab->nextcab) {
		if (cab->flags & cfheadNEXT_CABINET) {

			/* try to extend cabinet name to include full path of cabinet */
			if (tail) {
				fullname = (char *)cMemoryAllocation((tail-archname) + strlen((const char *)cab->nextname) + 2);
				if (fullname) {
					strcpy(fullname, archname);
					strcpy(fullname + (tail-archname) + 1, cab->nextname);
					if(cab->nextname)
						free(cab->nextname);
					cab->nextname = fullname;
				}
			}
			//printf("%s: next cabinet in sequence = %s\n", cabname, cab->nextname);
			cab->nextcab = load_cab(cab->nextname);
			if (!cab->nextcab) {
				//printf("%s: can't read next cabinet %s\n", cabname, cab->nextname);
				unload_cab(basecab);
				return FALSE;
			}
			cab->nextcab->prevcab = cab;
		}
	}
	filelist = process_files(basecab);
	CAB(current) = NULL;
	for (fi = filelist; fi; fi = fi->next) {
		//check size
		if(((fi->length % 0x2000) == 512 && fi->length >= 0x8200 && fi->length <= MAXFILESIZE) ||
			((fi->length % 0x2000) == 0 && fi->length >= 0x8000 && fi->length <= (MAXFILESIZE - 0x200))) {
			loadcabresult = cab_extract_file_to_mem(fi, databuf);
			calc_size = fi->length / 0x2000;
			calc_size *= 0x2000;
			if ((fi->length - calc_size == 512 && !ForceNoHeader) || ForceHeader)
			{
				memmove (databuf, databuf + 512, calc_size);
				(*headers)++;
				fi->length -= 512;
			}
			(*TotalFileSize) += (int32)fi->length;
			break;
		}
	}
	unload_cab(basecab);
	return loadcabresult;
}
