/*********************************************************************************************/
/*                                                                                           */
/*********************************************************************************************/

#include "../snes9x.h"
#include "../memmap.h"
#include "../ppu.h"
#include "../display.h"
#include "../apu.h"
#include "../gfx.h"
#include "../soundux.h"
#include "../snapshot.h"

#define MxST_REPLAY_VERSION 1
#define REPLAY_COUNT_ADD_SKIP 0x10
#define REPLAY_CHECK_COUNT_ADD_SKIP 0x4
#define REPLAY_CHECK_CHEAT_KEY_ADD_SKIP 0x10
#define REPLAY_FRAME_CHECK 300
#define REPLAY_DATA_TOTAL (12)
#define REPLAY_SAME_ADD (0xc)

/*********************************************************************************************/
/*                                                                                           */
/*********************************************************************************************/

#define REPLAY_RECODE_A_MACRO \
	for (int i = 0; i < REPLAY_DATA_TOTAL; i += 1) \
	{ \
	  XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_COUNT+i] = XppMovie.replay_data_temp [i] = XppMovie.replay_data[i]; \
	}

#define REPLAY_RECODE_B_MACRO \
	for (int i = 0; i < REPLAY_DATA_TOTAL; i += 1) \
	{ \
	  if(XppMovie.replay_data_temp [i] != XppMovie.replay_data[i]) change_byte = true; \
	}

#define REPLAY_RECODE_C_MACRO \
	XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_COUNT+REPLAY_SAME_ADD] = (uint8)(XppMovie.replay_data_count); \
	XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_COUNT+REPLAY_SAME_ADD+1] = (uint8)(XppMovie.replay_data_count >> 8); \
	XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_COUNT+REPLAY_SAME_ADD+2] = (uint8)(XppMovie.replay_data_count >> 16); \
	XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_COUNT+REPLAY_SAME_ADD+3] = (uint8)(XppMovie.replay_data_count >> 24);

#define REPLAY_PLAY_A_MACRO \
	for (int i = 0; i < REPLAY_DATA_TOTAL; i += 1) \
	{ \
	  XppMovie.replay_data[i] = XppMovie.replay_data_temp [i] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_COUNT+i]; \
	}

#define REPLAY_PLAY_B_MACRO \
	for (int i = 0; i < REPLAY_DATA_TOTAL; i += 1) \
	{ \
	  XppMovie.replay_data[i] = XppMovie.replay_data_temp [i]; \
	}

#define REPLAY_PLAY_C_MACRO \
	XppMovie.replay_data_count  = ((XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_COUNT+REPLAY_SAME_ADD]) + \
	                                (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_COUNT+REPLAY_SAME_ADD+1] << 8) + \
	                                (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_COUNT+REPLAY_SAME_ADD+2] << 16) + \
	                                (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_COUNT+REPLAY_SAME_ADD+3] << 24));

#define REPLAY_SET_EXT(slot,ma,mb,ra,rb) \
	XppMovie.movie_ext_snap[slot] = ma; \
	XppMovie.movie_ext_trace[slot] = mb; \
	XppMovie.resume_ext_snap[slot] = ra; \
	XppMovie.resume_ext_trace[slot] = rb;


void REPLAY_BYTE_SET ()
{
	if(XppMovie.Replay_Type == 0)
	{
		XppMovie.replay_data[0] = Memory.FillRAM[0x4218];
		XppMovie.replay_data[1] = Memory.FillRAM[0x4219];
		XppMovie.replay_data[2] = Memory.FillRAM[0x421a];
		XppMovie.replay_data[3] = Memory.FillRAM[0x421b];
		XppMovie.replay_data[4] = Memory.FillRAM[0x421c];
		XppMovie.replay_data[5] = Memory.FillRAM[0x421d];
		XppMovie.replay_data[6] = Memory.FillRAM[0x421e];
		XppMovie.replay_data[7] = Memory.FillRAM[0x421f];

		XppMovie.replay_data[8] = XppMovie.replay_data_dummy;
		XppMovie.replay_data[9] = XppMovie.replay_data_dummy;
		XppMovie.replay_data[10] = XppMovie.replay_data_dummy;
		XppMovie.replay_data[11] = XppMovie.replay_data_dummy;
	}
	else
	{
		XppMovie.replay_data[0] = (uint8)((IPPU.Joypads [0] & 0xff));
		XppMovie.replay_data[1] = (uint8)((IPPU.Joypads [0] & 0xff00) >> 8);
		XppMovie.replay_data[2] = (uint8)((IPPU.Joypads [1] & 0xff));
		XppMovie.replay_data[3] = (uint8)((IPPU.Joypads [1] & 0xff00) >> 8);
		XppMovie.replay_data[4] = (uint8)((IPPU.Joypads [2] & 0xff));
		XppMovie.replay_data[5] = (uint8)((IPPU.Joypads [2] & 0xff00) >> 8);
		XppMovie.replay_data[6] = (uint8)((IPPU.Joypads [3] & 0xff));
		XppMovie.replay_data[7] = (uint8)((IPPU.Joypads [3] & 0xff00) >> 8);
		XppMovie.replay_data[8] = (uint8)((IPPU.Joypads [4] & 0xff));
		XppMovie.replay_data[9] = (uint8)((IPPU.Joypads [4] & 0xff00) >> 8);

		XppMovie.replay_data[10] = XppMovie.replay_data_dummy;
		XppMovie.replay_data[11] = XppMovie.replay_data_dummy;
	}

//   XppMovie.replay_data[**] = (uint8)(XppMovie.replay_data_dummy);
//   XppMovie.replay_data[**] = (uint8)(XppMovie.replay_data_dummy >> 8);

}

void REPLAY_BYTE_GET ()
{

	if(XppMovie.Replay_Type == 0)
	{
		Memory.FillRAM[0x4218] = XppMovie.replay_data[0];
		Memory.FillRAM[0x4219] = XppMovie.replay_data[1];
		Memory.FillRAM[0x421a] = XppMovie.replay_data[2];
		Memory.FillRAM[0x421b] = XppMovie.replay_data[3];
		Memory.FillRAM[0x421c] = XppMovie.replay_data[4];
		Memory.FillRAM[0x421d] = XppMovie.replay_data[5];
		Memory.FillRAM[0x421e] = XppMovie.replay_data[6];
		Memory.FillRAM[0x421f] = XppMovie.replay_data[7];

		XppMovie.replay_data_dummy = XppMovie.replay_data[8];
		XppMovie.replay_data_dummy = XppMovie.replay_data[9];
		XppMovie.replay_data_dummy = XppMovie.replay_data[10];
		XppMovie.replay_data_dummy = XppMovie.replay_data[11];
	}
	else
	{
		IPPU.Joypads [0] = ((XppMovie.replay_data[0] + (XppMovie.replay_data[1] << 8)) | 0x80000000L);
		IPPU.Joypads [1] = ((XppMovie.replay_data[2] + (XppMovie.replay_data[3] << 8)) | 0x80000000L);
		IPPU.Joypads [2] = ((XppMovie.replay_data[4] + (XppMovie.replay_data[5] << 8)) | 0x80000000L);
		IPPU.Joypads [3] = ((XppMovie.replay_data[6] + (XppMovie.replay_data[7] << 8)) | 0x80000000L);
		IPPU.Joypads [4] = ((XppMovie.replay_data[8] + (XppMovie.replay_data[9] << 8)) | 0x80000000L);

		XppMovie.replay_data_dummy = XppMovie.replay_data[10];
		XppMovie.replay_data_dummy = XppMovie.replay_data[11];
	}

//   XppMovie.replay_data_dummy = XppMovie.replay_data[**] + (XppMovie.replay_data[**] << 8);

}

/*********************************************************************************************/
/*                                                                                           */
/*********************************************************************************************/

void XPP_REPLAY_START ()
{
	if(XppMovie.Replay_status == MOVIE_RECODE)
	{
		bool8 change_byte = false;

		REPLAY_BYTE_SET ();
		REPLAY_RECODE_B_MACRO

		if(XppMovie.Replay_count == 0)
		{

			bool8 replay_disable_shutdown=false;
			bool8 replay_irqcount=false;

			ZeroMemory (XppMovie.REPLAY_RAM, REPLAY_RAM_SIZE);
			for (int i = 0; i < REPLAY_DATA_TOTAL; i += 1)
			{
				XppMovie.replay_data_temp [i]=0;
			}

			XppMovie.REPLAY_RAM_CHECK_COUNT = REPLAY_RAM_SIZE - REPLAY_FRAME_CHECK_SIZE;
			XppMovie.Replay_frame_count_check = 0;
			XppMovie.Replay_count_fix = 0;
			XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-15] = XppMovie.Replay_Version = MxST_REPLAY_VERSION; // MxST Replay Version

			if(!Settings.Shutdown) replay_disable_shutdown = true;
			if(!Xppsystem.DisableIRQCycleCount) replay_irqcount = true;

			XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-16] = replay_disable_shutdown; // Check CpuShutdown
			XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-17] = replay_irqcount; // Check DisableIRQCycleCount

			if(XppMovie.Replay_Recode_Mode == 1) IPPU.Controller = SNES_MULTIPLAYER5;

			if(IPPU.Controller == SNES_MULTIPLAYER5) XppMovie.Replay_Type = 1;

			XppMovie.replay_data_count = 0;

			XppMovie.Replay_frame_count = 0;
		}
		else
		if(XppMovie.Replay_count == 1)
		{
			REPLAY_RECODE_A_MACRO
			XppMovie.replay_data_count = 0;
		}
		else
		if(XppMovie.replay_data_count > 0x7FFFFFFF || change_byte)
		{
			XppMovie.replay_data_count  -= 1;
			REPLAY_RECODE_C_MACRO

			XppMovie.REPLAY_RAM_COUNT += REPLAY_COUNT_ADD_SKIP;

			REPLAY_RECODE_A_MACRO
			XppMovie.replay_data_count = 0;
		}

		XppMovie.replay_data_count += 1;
		XppMovie.Replay_count += 1;

		if((XppMovie.REPLAY_RAM_COUNT > REPLAY_RAM_SIZE - (REPLAY_FRAME_CHECK_SIZE + REPLAY_CHEAT_KEY_SIZE + 0x8000)) || XppMovie.Replay_count > 0xfffffff0 || XppMovie.Replay_frame_count > 0xfffffff0) XPP_REPLAY_END ();
	}
	else
	if(XppMovie.Replay_status == MOVIE_PLAY)
	{
		if(XppMovie.Replay_count == 0)
		{
			bool8 replay_disable_shutdown=false;
			bool8 replay_irqcount=false;

			XppMovie.replay_data_count = 0;

			XppMovie.REPLAY_RAM_CHECK_COUNT = REPLAY_RAM_SIZE - REPLAY_FRAME_CHECK_SIZE;
			XppMovie.Replay_frame_count_check = 0;
			XppMovie.Replay_count_fix = 0;
			XppMovie.Replay_Version = XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-15];

			replay_disable_shutdown = XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-16]; // Check CpuShutdown
			replay_irqcount = XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-17]; // Check DisableIRQCycleCount
			
			if(!((!replay_disable_shutdown && Settings.Shutdown) || (replay_disable_shutdown && !Settings.Shutdown)) && !XppMovie.file_open_error)
			{
						GFX.InfoStringTimeout = 200;
						GFX.InfoString=("Warning - Replay file and the setting are different. (Shutdown)");
			}

			if(!((!replay_irqcount && Xppsystem.DisableIRQCycleCount) || (replay_irqcount && !Xppsystem.DisableIRQCycleCount)) && !XppMovie.file_open_error)
			{
						GFX.InfoStringTimeout = 200;
						GFX.InfoString=("Warning - Replay file and the setting are different. (IRQCycleCount)");
			}

			if(XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-14] == 0xff)
			{
				XppMovie.Replay_Type = 1;
				IPPU.Controller = SNES_MULTIPLAYER5;
			}

			XppMovie.REPLAY_RAM_COUNT_MAX = XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-3]
			+ (XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-2] << 8)
			+ (XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-1] << 16);

			XppMovie.Replay_frame_count_max = XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-12]
			+ (XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-11] << 8)
			+ (XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-10] << 16)
			+ (XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-9] << 24);

			XppMovie.Replay_frame_count = 0;
			REPLAY_BYTE_SET ();
		}
		else
		if(XppMovie.replay_data_count == 0)
		{
			REPLAY_PLAY_A_MACRO
			REPLAY_PLAY_C_MACRO
			XppMovie.REPLAY_RAM_COUNT += REPLAY_COUNT_ADD_SKIP;
		}
		else
		{
			REPLAY_PLAY_B_MACRO
			XppMovie.replay_data_count  -= 1;
		}

		REPLAY_BYTE_GET ();

		XppMovie.Replay_count += 1;

		if((XppMovie.Replay_Version == 0 && (XppMovie.REPLAY_RAM_COUNT_MAX + REPLAY_COUNT_ADD_SKIP) < XppMovie.REPLAY_RAM_COUNT) ||
		(XppMovie.Replay_Version > 0 && XppMovie.Replay_frame_count_max < XppMovie.Replay_frame_count))
		{
			XPP_REPLAY_END ();
			GFX.InfoStringTimeout = 100;
			GFX.InfoString=("Movie Play Finished");
		}

	}

}


void XPP_REPLAY_END ()
{
	if(XppMovie.Replay_status == MOVIE_RECODE)
	{
		REPLAY_BYTE_SET ();

		REPLAY_RECODE_A_MACRO
		REPLAY_RECODE_C_MACRO

		if(XppMovie.Replay_Type == 1)
		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-14] = 0xff;

		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-12] = (uint8)(XppMovie.Replay_frame_count);
		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-11] = (uint8)(XppMovie.Replay_frame_count >> 8);
		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-10] = (uint8)(XppMovie.Replay_frame_count >> 16);
		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-9] = (uint8)(XppMovie.Replay_frame_count >> 24);

		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-8] = (uint8)(XppMovie.Replay_count);
		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-7] = (uint8)(XppMovie.Replay_count >> 8);
		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-6] = (uint8)(XppMovie.Replay_count >> 16);
		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-5] = (uint8)(XppMovie.Replay_count >> 24);

		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-3] = (uint8)(XppMovie.REPLAY_RAM_COUNT);
		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-2] = (uint8)(XppMovie.REPLAY_RAM_COUNT >> 8);
		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-1] = (uint8)(XppMovie.REPLAY_RAM_COUNT >> 16);

		XPP_REPLAY_SAVE (XppMovie.movie_ext_trace[XppMovie.Replay_Slot]);

		XppMovie.Replay_status = MOVIE_NONE;
		XppMovie.Replay_count = 0;
		XppMovie.replay_data_count = 0;
		XppMovie.REPLAY_RAM_COUNT_MAX = 0;
		XppMovie.REPLAY_RAM_COUNT = 0;
		XppMovie.Replay_Type = 0;
		XppMovie.Replay_frame_count = 0;
		XppMovie.Replay_frame_count_max = 0;
	}
	else
	if(XppMovie.Replay_status == MOVIE_PLAY)
	{
		XppMovie.Replay_status = MOVIE_NONE;
		XppMovie.Replay_count = 0;
		XppMovie.replay_data_count = 0;
		XppMovie.REPLAY_RAM_COUNT_MAX = 0;
		XppMovie.REPLAY_RAM_COUNT = 0;
		XppMovie.Replay_Type = 0;
		XppMovie.Replay_frame_count = 0;
		XppMovie.Replay_frame_count_max = 0;
	}
}

/*********************************************************************************************/
/*                                                                                           */
/*********************************************************************************************/

void XPP_REPLAY_RESUME (int mode)
{

	if(mode==REPLAY_LOAD)
	{
		bool8 replay_disable_shutdown=false;
		bool8 replay_irqcount=false;

		XppMovie.Replay_status = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x1];
		XppMovie.Replay_Type = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x2];
		XppMovie.REPLAY_RAM_COUNT = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x4] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5] << 16);

		XppMovie.Replay_count = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x8] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x9] << 24);

		XppMovie.replay_data_count = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0xa]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0xb] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0xc] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0xd] << 24);

		XppMovie.replay_data[0] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x10];
		XppMovie.replay_data[1] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x11];
		XppMovie.replay_data[2] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x12];
		XppMovie.replay_data[3] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x13];
		XppMovie.replay_data[4] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x14];
		XppMovie.replay_data[5] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x15];
		XppMovie.replay_data[6] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x16];
		XppMovie.replay_data[7] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x17];
		XppMovie.replay_data[8] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x18];
		XppMovie.replay_data[9] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x19];
		XppMovie.replay_data[10] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x1a];
		XppMovie.replay_data[11] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x1b];

		XppMovie.replay_data_temp[0] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x20];
		XppMovie.replay_data_temp[1] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x21];
		XppMovie.replay_data_temp[2] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x22];
		XppMovie.replay_data_temp[3] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x23];
		XppMovie.replay_data_temp[4] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x24];
		XppMovie.replay_data_temp[5] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x25];
		XppMovie.replay_data_temp[6] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x26];
		XppMovie.replay_data_temp[7] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x27];
		XppMovie.replay_data_temp[8] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x28];
		XppMovie.replay_data_temp[9] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x29];
		XppMovie.replay_data_temp[10] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x2a];
		XppMovie.replay_data_temp[11] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x2b];

		XppMovie.Replay_frame_count = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x30]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x31] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x32] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x33] << 24);

		XppMovie.Replay_frame_count_check = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x34]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x35] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x36] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x37] << 24);

		XppMovie.REPLAY_RAM_CHECK_COUNT = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x38]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x39] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3a] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3b] << 24);

		XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3c]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3d] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3e] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3f] << 24);

		Xppcode.CheatNum = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x40];
		Xppcode.CheatNumRock = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x41];
		Xppcode.CheatNumRock_Old = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x42];
		Xppcode.CheatFlag = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x43];
		Xppcode.CheatNumBackFlag = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x44];

		CPU.WaitCounter = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x45]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x46] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x47] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x48] << 24);

		CPU.InDMA = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x49];

		CPU.NMICycleCount = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x4a]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x4b] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x4c] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x4d] << 24);

		IPPU.Joypads[0] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x50]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x51] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x52] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x53] << 24);

		IPPU.Joypads[1] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x54]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x55] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x56] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x57] << 24);

		IPPU.Joypads[2] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x58]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x59] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5a] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5b] << 24);

		IPPU.Joypads[3] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5c]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5d] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5e] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5f] << 24);

		IPPU.Joypads[4] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x60]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x61] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x62] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x63] << 24);

		IPPU.SuperScope = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x64]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x65] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x66] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x67] << 24);

		IPPU.Mouse[0] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x68]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x69] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6a] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6b] << 24);

		IPPU.Mouse[1] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6c]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6d] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6e] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6f] << 24);

		PPU.ClipCounts[4] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x70];
		PPU.ClipCounts[5] = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x71];
		PPU.LastSprite = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x72];
		PPU.BG_Forced = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x73];
		PPU.RecomputeClipWindows = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x74];
		PPU.OpenBus1 = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x75];
		PPU.OpenBus2 = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x76];
		PPU.RangeTimeOver = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x77];

		PPU.OAMWriteRegister = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x78]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x79] << 8);

		PPU.BGnxOFSbyte = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7a];

		IPPU.VRAMReadBuffer = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7b]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7c] << 8);

		PPU.OBJThroughMain = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7d];
		PPU.OBJThroughSub = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7e];
		PPU.OBJAddition = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7f];

		IAPU.NextAPUTimerPos = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x80]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x81] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x82] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x83] << 24);

		IAPU.APUTimerCounter = XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x84]
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x85] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x86] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x87] << 24);

		XppMovie.REPLAY_RAM_COUNT_MAX = XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-3]
		+ (XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-2] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-1] << 16);

		XppMovie.Replay_frame_count_max = XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-12]
		+ (XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-11] << 8)
		+ (XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-10] << 16)
		+ (XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-9] << 24);

		XppMovie.Replay_frame_count_cheat = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT]
		+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+1] << 8)
		+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+2] << 16)
		+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+3] << 24);

		XppMovie.replay_cheat_data [0] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+4];
		XppMovie.replay_cheat_data [1] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+5];
		XppMovie.replay_cheat_data [2] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+6];
		XppMovie.replay_cheat_data [3] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+7];
		XppMovie.replay_cheat_data [4] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+8];
		XppMovie.replay_cheat_data [5] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+9];

		replay_disable_shutdown = XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-16]; // Check CpuShutdown
		replay_irqcount = XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-17]; // Check DisableIRQCycleCount

		if(!((!replay_disable_shutdown && Settings.Shutdown) || (replay_disable_shutdown && !Settings.Shutdown)) && !XppMovie.file_open_error)
		{
					GFX.InfoStringTimeout = 200;
					GFX.InfoString=("Warning - Replay file and the setting are different. (Shutdown)");
		}

		if(!((!replay_irqcount && Xppsystem.DisableIRQCycleCount) || (replay_irqcount && !Xppsystem.DisableIRQCycleCount)) && !XppMovie.file_open_error)
		{
					GFX.InfoStringTimeout = 200;
					GFX.InfoString=("Warning - Replay file and the setting are different. (IRQCycleCount)");
		}

		XppMovie.Replay_Version = XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-15]; //Replay Version

		if(XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-14] == 0xff)
		{
				XppMovie.Replay_Type = 1;
				IPPU.Controller = SNES_MULTIPLAYER5;
		}

	}
	else if(mode==REPLAY_SAVE)
	{
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x1] = XppMovie.Replay_status;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x2] = XppMovie.Replay_Type;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3] = (uint8)(XppMovie.REPLAY_RAM_COUNT);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x4] = (uint8)(XppMovie.REPLAY_RAM_COUNT >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5] = (uint8)(XppMovie.REPLAY_RAM_COUNT >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6] = (uint8)(XppMovie.Replay_count);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7] = (uint8)(XppMovie.Replay_count >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x8] = (uint8)(XppMovie.Replay_count >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x9] = (uint8)(XppMovie.Replay_count >> 24);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0xa] = (uint8)(XppMovie.replay_data_count);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0xb] = (uint8)(XppMovie.replay_data_count >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0xc] = (uint8)(XppMovie.replay_data_count >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0xd] = (uint8)(XppMovie.replay_data_count >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x10] = XppMovie.replay_data[0];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x11] = XppMovie.replay_data[1];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x12] = XppMovie.replay_data[2];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x13] = XppMovie.replay_data[3];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x14] = XppMovie.replay_data[4];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x15] = XppMovie.replay_data[5];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x16] = XppMovie.replay_data[6];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x17] = XppMovie.replay_data[7];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x18] = XppMovie.replay_data[8];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x19] = XppMovie.replay_data[9];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x1a] = XppMovie.replay_data[10];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x1b] = XppMovie.replay_data[11];

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x20] = XppMovie.replay_data_temp[0];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x21] = XppMovie.replay_data_temp[1];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x22] = XppMovie.replay_data_temp[2];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x23] = XppMovie.replay_data_temp[3];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x24] = XppMovie.replay_data_temp[4];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x25] = XppMovie.replay_data_temp[5];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x26] = XppMovie.replay_data_temp[6];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x27] = XppMovie.replay_data_temp[7];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x28] = XppMovie.replay_data_temp[8];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x29] = XppMovie.replay_data_temp[9];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x2a] = XppMovie.replay_data_temp[10];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x2b] = XppMovie.replay_data_temp[11];

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x30] = (uint8)(XppMovie.Replay_frame_count);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x31] = (uint8)(XppMovie.Replay_frame_count >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x32] = (uint8)(XppMovie.Replay_frame_count >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x33] = (uint8)(XppMovie.Replay_frame_count >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x34] = (uint8)(XppMovie.Replay_frame_count_check);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x35] = (uint8)(XppMovie.Replay_frame_count_check >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x36] = (uint8)(XppMovie.Replay_frame_count_check >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x37] = (uint8)(XppMovie.Replay_frame_count_check >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x38] = (uint8)(XppMovie.REPLAY_RAM_CHECK_COUNT);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x39] = (uint8)(XppMovie.REPLAY_RAM_CHECK_COUNT >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3a] = (uint8)(XppMovie.REPLAY_RAM_CHECK_COUNT >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3b] = (uint8)(XppMovie.REPLAY_RAM_CHECK_COUNT >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3c] = (uint8)(XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3d] = (uint8)(XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3e] = (uint8)(XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x3f] = (uint8)(XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x40] = Xppcode.CheatNum;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x41] = Xppcode.CheatNumRock;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x42] = Xppcode.CheatNumRock_Old;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x43] = Xppcode.CheatFlag;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x44] = Xppcode.CheatNumBackFlag;

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x45] = (uint8)(CPU.WaitCounter);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x46] = (uint8)(CPU.WaitCounter >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x47] = (uint8)(CPU.WaitCounter >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x48] = (uint8)(CPU.WaitCounter >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x49] = CPU.InDMA;

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x4a] = (uint8)(CPU.NMICycleCount);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x4b] = (uint8)(CPU.NMICycleCount >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x4c] = (uint8)(CPU.NMICycleCount >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x4d] = (uint8)(CPU.NMICycleCount >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x50] = (uint8)(IPPU.Joypads[0]);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x51] = (uint8)(IPPU.Joypads[0] >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x52] = (uint8)(IPPU.Joypads[0] >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x53] = (uint8)(IPPU.Joypads[0] >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x54] = (uint8)(IPPU.Joypads[1]);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x55] = (uint8)(IPPU.Joypads[1] >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x56] = (uint8)(IPPU.Joypads[1] >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x57] = (uint8)(IPPU.Joypads[1] >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x58] = (uint8)(IPPU.Joypads[2]);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x59] = (uint8)(IPPU.Joypads[2] >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5a] = (uint8)(IPPU.Joypads[2] >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5b] = (uint8)(IPPU.Joypads[2] >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5c] = (uint8)(IPPU.Joypads[3]);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5d] = (uint8)(IPPU.Joypads[3] >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5e] = (uint8)(IPPU.Joypads[3] >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x5f] = (uint8)(IPPU.Joypads[3] >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x60] = (uint8)(IPPU.Joypads[4]);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x61] = (uint8)(IPPU.Joypads[4] >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x62] = (uint8)(IPPU.Joypads[4] >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x63] = (uint8)(IPPU.Joypads[4] >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x64] = (uint8)(IPPU.SuperScope);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x65] = (uint8)(IPPU.SuperScope >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x66] = (uint8)(IPPU.SuperScope >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x67] = (uint8)(IPPU.SuperScope >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x68] = (uint8)(IPPU.Mouse[0]);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x69] = (uint8)(IPPU.Mouse[0] >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6a] = (uint8)(IPPU.Mouse[0] >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6b] = (uint8)(IPPU.Mouse[0] >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6c] = (uint8)(IPPU.Mouse[1]);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6d] = (uint8)(IPPU.Mouse[1] >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6e] = (uint8)(IPPU.Mouse[1] >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x6f] = (uint8)(IPPU.Mouse[1] >> 24);

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x70] = PPU.ClipCounts[4];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x71] = PPU.ClipCounts[5];
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x72] = PPU.LastSprite;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x73] = PPU.BG_Forced;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x74] = PPU.RecomputeClipWindows;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x75] = PPU.OpenBus1;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x76] = PPU.OpenBus2;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x77] = PPU.RangeTimeOver;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x78] = (uint8)(PPU.OAMWriteRegister);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x79] = (uint8)(PPU.OAMWriteRegister >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7a] = PPU.BGnxOFSbyte;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7b] = (uint8)(IPPU.VRAMReadBuffer);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7c] = (uint8)(IPPU.VRAMReadBuffer >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7d] = PPU.OBJThroughMain;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7e] = PPU.OBJThroughSub;
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x7f] = PPU.OBJAddition;

		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x80] = (uint8)(IAPU.NextAPUTimerPos);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x81] = (uint8)(IAPU.NextAPUTimerPos >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x82] = (uint8)(IAPU.NextAPUTimerPos >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x83] = (uint8)(IAPU.NextAPUTimerPos >> 24);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x84] = (uint8)(IAPU.APUTimerCounter);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x85] = (uint8)(IAPU.APUTimerCounter >> 8);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x86] = (uint8)(IAPU.APUTimerCounter >> 16);
		XppMovie.REPLAY_RAM[REPLAY_HEADER_POINT-0x87] = (uint8)(IAPU.APUTimerCounter >> 24);

		if(XppMovie.Replay_Type == 1)
		XppMovie.REPLAY_RAM[REPLAY_RAM_SIZE-14] = 0xff;
	}
	else if(mode==REPLAY_CLEAR)
	{
		for (int i = 1; i < 0x100; i += 1)
		{
			XppMovie.REPLAY_RAM [REPLAY_HEADER_POINT-i]=0;
		}
	}
}

void XPP_REPLAY_CHECK ()
{
	if(XppMovie.Replay_status == MOVIE_RECODE && XppMovie.Replay_Version != 0)
	{
		if(XppMovie.Replay_frame_count_check == REPLAY_FRAME_CHECK && XppMovie.REPLAY_RAM_CHECK_COUNT < REPLAY_RAM_SIZE-0x1000)
		{
			XppMovie.Replay_frame_count_check = 0;
			XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHECK_COUNT+0x0] = (uint8)(XppMovie.Replay_count);
			XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHECK_COUNT+0x1] = (uint8)(XppMovie.Replay_count >> 8);
			XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHECK_COUNT+0x2] = (uint8)(XppMovie.Replay_count >> 16);
			XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHECK_COUNT+0x3] = (uint8)(XppMovie.Replay_count >> 24);

			XppMovie.REPLAY_RAM_CHECK_COUNT += REPLAY_CHECK_COUNT_ADD_SKIP;
		}
		XppMovie.Replay_frame_count_check++;
	}
	else
	if(XppMovie.Replay_status == MOVIE_PLAY && XppMovie.Replay_Version != 0)
	{
		if(XppMovie.Replay_frame_count_check == REPLAY_FRAME_CHECK && XppMovie.REPLAY_RAM_CHECK_COUNT < REPLAY_RAM_SIZE-0x1000)
		{
			uint32 XppMovie_Replay_count_temp;

			XppMovie.Replay_frame_count_check = 0;

			XppMovie_Replay_count_temp = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHECK_COUNT+0x0]
			+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHECK_COUNT+0x1] << 8)
			+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHECK_COUNT+0x2] << 16)
			+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHECK_COUNT+0x3] << 24);

#if 0 // DEBUG
			if(XppMovie.Replay_count != XppMovie_Replay_count_temp)
			{
				XppMovie.Replay_count_fix++;

#if 1 // DEBUG
				if (Options.DisplayScreenBufferAll)
				{
					GFX.InfoStringTimeout = 120;
					sprintf (String,"Missing Movie Sync / Replay_count [%10d-%10d / Movie Sync Fix %5d]",XppMovie.Replay_count,XppMovie_Replay_count_temp,XppMovie.Replay_count_fix);
					GFX.InfoString=(String);
				}
#endif
				XppMovie.Replay_count = XppMovie_Replay_count_temp;
			}
#endif
			XppMovie.REPLAY_RAM_CHECK_COUNT += REPLAY_CHECK_COUNT_ADD_SKIP;
		}
		XppMovie.Replay_frame_count_check++;

		if(XppMovie.Replay_Version > 0 && XppMovie.Replay_frame_count_max < XppMovie.Replay_frame_count)
		{
			XPP_REPLAY_END ();
			GFX.InfoStringTimeout = 100;
			GFX.InfoString=("Movie Play Finished");
		}
	}

	XPP_REPLAY_CHEAT_KEY (0);
}

void XPP_REPLAY_CHEAT_KEY (int codetype)
{
	if(XppMovie.Replay_Version > 0)
	{
		if(XppMovie.Replay_status == MOVIE_RECODE)
		{
			if(XppMovie.Replay_frame_count == 1)
			{
				XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT = REPLAY_RAM_SIZE - REPLAY_FRAME_CHECK_SIZE - REPLAY_CHEAT_KEY_SIZE;
			}
			else if(codetype == ROMCHEAT_CODE || codetype == ROMCHEAT_SELECT)
			{
				XppMovie.replay_cheat_data [0] = codetype;
				XppMovie.replay_cheat_data [1] = Xppcode.CheatNum;
				XppMovie.replay_cheat_data [2] = Xppcode.CheatNumRock;
				XppMovie.replay_cheat_data [3] = Xppcode.CheatNumRock_Old;
				XppMovie.replay_cheat_data [4] = Xppcode.CheatFlag;
				XppMovie.replay_cheat_data [5] = Xppcode.CheatNumBackFlag;

				if(XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT < REPLAY_RAM_SIZE - REPLAY_FRAME_CHECK_SIZE)
				{
					XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT]   = (uint8)(XppMovie.Replay_frame_count);
					XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+1] = (uint8)(XppMovie.Replay_frame_count >> 8);
					XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+2] = (uint8)(XppMovie.Replay_frame_count >> 16);
					XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+3] = (uint8)(XppMovie.Replay_frame_count >> 24);
					XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+4] = XppMovie.replay_cheat_data [0];
					XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+5] = XppMovie.replay_cheat_data [1];
					XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+6] = XppMovie.replay_cheat_data [2];
					XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+7] = XppMovie.replay_cheat_data [3];
					XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+8] = XppMovie.replay_cheat_data [4];
					XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+9] = XppMovie.replay_cheat_data [5];

					XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT += REPLAY_CHECK_CHEAT_KEY_ADD_SKIP;
				}
			}
		}
		else if(XppMovie.Replay_status == MOVIE_PLAY && codetype == 0)
		{
			if(XppMovie.Replay_frame_count == 1)
			{
				XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT = REPLAY_RAM_SIZE - REPLAY_FRAME_CHECK_SIZE - REPLAY_CHEAT_KEY_SIZE;

				XppMovie.Replay_frame_count_cheat = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT]
				+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+1] << 8)
				+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+2] << 16)
				+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+3] << 24);
				XppMovie.replay_cheat_data [0] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+4];
				XppMovie.replay_cheat_data [1] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+5];
				XppMovie.replay_cheat_data [2] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+6];
				XppMovie.replay_cheat_data [3] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+7];
				XppMovie.replay_cheat_data [4] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+8];
				XppMovie.replay_cheat_data [5] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+9];
			}
			else if(XppMovie.Replay_frame_count_cheat == XppMovie.Replay_frame_count)
			{
				codetype = XppMovie.replay_cheat_data [0];
				Xppcode.CheatNum = XppMovie.replay_cheat_data [1];
				Xppcode.CheatNumRock = XppMovie.replay_cheat_data [2];
				Xppcode.CheatNumRock_Old = XppMovie.replay_cheat_data [3];
				Xppcode.CheatFlag = XppMovie.replay_cheat_data [4];
				Xppcode.CheatNumBackFlag = XppMovie.replay_cheat_data [5];

				if (codetype == ROMCHEAT_CODE)
				{
					if (Xppcode.CheatNumRock == 1)
					{
						Xppcode.CheatNumRock = 0;
						GFX.InfoStringTimeout = 100;
						GFX.InfoString=("Rock Code Off");
					}
					else
					codetype = ROMCHEAT_RUN;
				}

				if ((Memory.HiROM && ROM [0xffd9] == 0)||(Memory.LoROM && ROM [0x7fd9] == 0 || Settings.SFMTURBO))
				ROMCHEAT_CODE_JAPAN (codetype);
				else if ((Memory.HiROM && ROM [0xffd9] == 1)||(Memory.LoROM && ROM [0x7fd9] == 1))
				ROMCHEAT_CODE_USA (codetype);

				XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT += REPLAY_CHECK_CHEAT_KEY_ADD_SKIP;

				if(XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT < REPLAY_RAM_SIZE - REPLAY_FRAME_CHECK_SIZE)
				{
					XppMovie.Replay_frame_count_cheat = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT]
					+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+1] << 8)
					+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+2] << 16)
					+ (XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+3] << 24);
					XppMovie.replay_cheat_data [0] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+4];
					XppMovie.replay_cheat_data [1] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+5];
					XppMovie.replay_cheat_data [2] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+6];
					XppMovie.replay_cheat_data [3] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+7];
					XppMovie.replay_cheat_data [4] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+8];
					XppMovie.replay_cheat_data [5] = XppMovie.REPLAY_RAM[XppMovie.REPLAY_RAM_CHEAT_KEY_COUNT+9];
				}
			}
		}
	}
}

void XPP_REPLAY_EXT_SET ()
{
	REPLAY_SET_EXT(0,".m0s",".m0t",".r0s",".r0t")
	REPLAY_SET_EXT(1,".m1s",".m1t",".r1s",".r1t")
	REPLAY_SET_EXT(2,".m2s",".m2t",".r2s",".r2t")
	REPLAY_SET_EXT(3,".m3s",".m3t",".r3s",".r3t")
	REPLAY_SET_EXT(4,".m4s",".m4t",".r4s",".r4t")
	REPLAY_SET_EXT(5,".m5s",".m5t",".r5s",".r5t")
	REPLAY_SET_EXT(6,".m6s",".m6t",".r6s",".r6t")
	REPLAY_SET_EXT(7,".m7s",".m7t",".r7s",".r7t")
	REPLAY_SET_EXT(8,".m8s",".m8t",".r8s",".r8t")
	REPLAY_SET_EXT(9,".m9s",".m9t",".r9s",".r9t")
}

void XPP_REPLAY_TIME_H_M_S (uint32 time)
{
	XppMovie.time_hour = 0;
	XppMovie.time_min = 0;
	XppMovie.time_sec = 0;

	if(time>=3600)
	{
		XppMovie.time_hour=time/3600;
		time = time - (XppMovie.time_hour * 3600);
	}

	if(time>=60)
	{
		XppMovie.time_min=time/60;
		time = time - (XppMovie.time_min * 60);
	}

	XppMovie.time_sec = time;

}

void XPP_REPLAY_LOAD (const char *Xpp_ext)
{
	STREAM f_replay;
	if ((f_replay = OPEN_STREAM (S9xGetFilename (Xpp_ext), "rb")))
	{
		READ_STREAM (XppMovie.REPLAY_RAM, REPLAY_RAM_SIZE, f_replay);
		CLOSE_STREAM(f_replay);
	}
	else
	{
		XppMovie.file_open_error = TRUE;
	}
}

void XPP_REPLAY_SAVE (const char *Xpp_ext)
{
	STREAM f_replay;
	if ((f_replay = OPEN_STREAM (S9xGetFilename (Xpp_ext), "wb")))
	{
		WRITE_STREAM (XppMovie.REPLAY_RAM, REPLAY_RAM_SIZE, f_replay);
		CLOSE_STREAM(f_replay);
	}
}

