#include <shlobj.h>
#include "../common/commutil.h"
#include "../Snes9xpp/globl.h"
#include "../snes9x/Snes9x.h"
#include "../Snes9x/soundux.h"
#include "../Snes9x/ppu.h"
#include "../Snes9x/apu.h"
#include "../Snes9x/memmap.h"
#include "snes9xw.h"
#include "directx.h"
#include "joywin.h"
#include "Render.h"
#include "cfgdlg.h"

/***********************************************************************************************
 Global Dialog
************************************************************************************************/ 

static char tmpcharbuf[256];

/***********************************************************************************************
 Cfg Dialog
************************************************************************************************/ 

HWND hProp;
WNDPROC PropProc;
UINT StartPageIndex = 0;
UINT ConfigChangePageIndex = 0;
unsigned char errConfigSwitch = FALSE;

LRESULT CALLBACK JoypadDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK SNESEmulationDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK DirectoryDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK SettingsDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK OtherDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam);
bool SetJoyButtonCode(HWND hFocus, WORD KeyCode);

void S9xwSettingsLoad();
void SNESEmulationSettingsLoad();
void JoySettingsLoad();
void DirectorySettingsLoad();
void OtherSettingsLoad();

void S9xwSettingsSave();
void SNESEmulationSettingsSave();
void JoySettingsSave();
void DirectorySettingsSave();
void OtherSettingsSave();

void S9xwDefaultSettingsLoad();
void SNESEmulationDefaultSettingsLoad();
void JoyDefaultSettingsLoad();
void OtherDefaultSettingsLoad();

static unsigned char ConfigDlgSettingsSaved;
static unsigned char ConfigDlgDefaultSettingsLoaded;

/***********************************************************************************************
 Set Dialog
************************************************************************************************/ 

extern struct SSoundRates SoundRates[];
extern bool Draw1FrameFlag;

//proc
static void SettingsDlgCtrlSet(HWND hWnd);
static void Scale2X16RadioCtrlSet(HWND hWnd);

//sound settings
static int GUI_SoundMode_cpy;
static int Settings_SoundBufferSize_cpy;
static int Settings_SoundMixInterval_cpy;

//Display settings
static int DoubleRenderingModes16_cpy;
static unsigned char ForceWindowedDirectDrawClipper_cpy;
static unsigned char dRendering_cpy;
static unsigned char DirectX_DisableHardwareScale_cpy;
static bool GUI_DoubleBuffered_cpy;
static bool DirectX_OffScreenVRAM_cpy;
static bool DirectX_BackBufferVRAM_cpy;

//Thread
static DWORD GUI_PriorityClass_cpy;
static int GUI_ThreadPriority_cpy;

static unsigned char Options_CpuSleep_cpy;
static unsigned char CPU_DisableSpinWaitLoop_cpy;

//Valuable extern
extern bool8 Options_GFXengine2_cpy;

extern int DoubleRenderingModes16;
extern unsigned char dRendering;
extern unsigned char ForceWindowedDirectDrawClipper;
extern bool8 errConfigSwitch;
extern bool8 TransparencyFlag;
extern bool8 TransparencyFlag_cpy;

//proc extern
extern bool8 SetupSound( long rate, bool8 sixteen_bit, bool8 stereo);
extern bool8 S9xwSwitchScreenMode(bool8 Windowed);
extern void CurFullscreenDisplayModeIndexChange();

extern void UpdateBackBuffer();

/*****************
 default settings
*****************/
extern int GUI_SoundMode_def;
extern int Settings_SoundBufferSize_def;
extern int Settings_SoundMixInterval_def;
extern int DoubleRenderingModes16_def;
extern unsigned char ForceWindowedDirectDrawClipper_def;
extern unsigned char dRendering_def;
extern bool GUI_DoubleBuffered_def;
extern DWORD GUI_PriorityClass_def;
extern int GUI_ThreadPriority_def;

extern bool8 Options_CpuSleep_def;
extern bool8 CPU_DisableSpinWaitLoop_def;


/***********************************************************************************************
 Snes Dialog
************************************************************************************************/ 

extern bool8 TransparencyFlag;

//proc
static void SNESEmulationDlgCtrlSet(HWND hWnd);
static void SneseDlgSetSubClass(HWND hWnd);
static void SneseDlgReleaseSubClass();

//graphics
bool8 TransparencyFlag_cpy;
bool8 Options_GFXengine2_cpy;
static bool8 Settings_Mode7Interpolate_cpy;
static bool8 Settings_BGLayering_cpy;
static bool8 Settings_DisableGraphicWindows_cpy;
static bool8 Settings_DisableHDMA_cpy;
static bool DirectX_wait_for_vsync_cpy;

//sound
static bool8 Settings_InterpolatedSound_cpy;
static bool8 Settings_NextAPUEnabled_cpy;
static bool8 Settings_DisableSoundEcho_cpy;
static bool8 Settings_ReverseStereo_cpy;
static bool8 Settings_SoundSync_cpy;
static bool8 Settings_DisableMasterVolume_cpy;
static bool8 Settings_SoundEnvelopeHeightReading_cpy;
static uint8 Settings_SoundSkipMethod_cpy;

//CPU
static int32 Settings_ShutdownMaster_cpy;

//controller
static uint32 Settings_ControllerOption_cpy;
static bool8 Settings_MultiPlayer5_cpy;
static bool8 Settings_Mouse_cpy;
static bool8 Settings_SuperScope_cpy;
static bool8 Settings_Justifier_cpy;
static unsigned long Options_JoyReadRateMask_cpy;

//timing
static uint32 Settings_SkipFrames_cpy;
static uint32 Settings_AutoMaxSkipFrames_cpy;
static long Settings_CyclesPercentage_cpy;
static uint32 Settings_FrameTimePAL_cpy;
static uint32 Settings_FrameTimeNTSC_cpy;

//auto save sram
static int32 Settings_AutoSaveDelay_cpy; // Snes9xpp Original

//proc extern
extern "C" void S9xNextController ();

static void SetSnesDefaultPlugedInputDeviceCombo(HWND hWnd);
static uint32 GetControllerOptionByComboindex(LRESULT conboindex);

/*****************
 default settings
*****************/
extern bool8 Settings_Mode7Interpolate_def;
extern bool8 TransparencyFlag_def;
extern bool8 Settings_BGLayering_def;
extern bool8 Settings_DisableGraphicWindows_def;
extern bool8 Settings_DisableHDMA_def;
extern bool8 Settings_InterpolatedSound_def;
extern bool8 Settings_NextAPUEnabled_def;
extern bool8 Settings_DisableSoundEcho_def;
extern bool8 Settings_ReverseStereo_def;
extern bool8 Settings_SoundSync_def;
extern bool8 Settings_DisableMasterVolume_def;
extern bool8 Settings_SoundEnvelopeHeightReading_def;
extern uint8 Settings_SoundSkipMethod_def;
extern bool8 Settings_ShutdownMaster_def;
extern uint32 Settings_ControllerOption_def;
extern bool8 Settings_MultiPlayer5_def;
extern bool8 Settings_Mouse_def;
extern bool8 Settings_SuperScope_def;
extern bool8 Settings_Justifier_def;
extern uint32 Settings_SkipFrames_def;
extern uint32 Settings_AutoMaxSkipFrames_def;
extern long Settings_CyclesPercentage_def;
extern uint32 Settings_FrameTimePAL_def;
extern uint32 Settings_FrameTimeNTSC_def;
extern int32 Settings_AutoSaveDelay_def; // Snes9xpp Original

/***********************************************************************************************
 Joy Dialog
************************************************************************************************/ 

static LRESULT CALLBACK EditGetKeyProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
extern LRESULT CALLBACK SubPropProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam);
static WORD SetJoyButtonText(HWND hWnd, WORD &JoyButton, WORD &JoyButtonOld, bool DuplicateCheckFlag);
static void SetSubClass(HWND hWnd);
static void ReleaseSubClass(HWND hWnd);
static void JoyDlgCtlSet(HWND hWnd);
static struct SJoypad Joypadcpy[5];
static DWORD MapFunc_cpy[12];
static int ComboNo;
static bool TimerActiveFlag = 0;
static HBRUSH SelectedEditBkGreen = 0;
static HBRUSH SelectedEditBkOrange = 0;

static WNDPROC Edit_Left, Edit_Right, Edit_Up, Edit_Down, Edit_Start, Edit_Select,
	   Edit_A, Edit_B, Edit_X, Edit_Y, Edit_L, Edit_R,
	   Edit_MapFunc0, Edit_MapFunc1, Edit_MapFunc2, Edit_MapFunc3,
	   Edit_MapFunc4, Edit_MapFunc5, Edit_MapFunc6, Edit_MapFunc7, Edit_MapFunc8, Edit_MapFunc9, Edit_MapFunc10, Edit_MapFunc11;
static HWND hEdit_Left, hEdit_Right, hEdit_Up, hEdit_Down, hEdit_Start, hEdit_Select,
	   hEdit_A, hEdit_B, hEdit_X, hEdit_Y, hEdit_L, hEdit_R,
	   hEdit_MapFunc0, hEdit_MapFunc1, hEdit_MapFunc2, hEdit_MapFunc3,
	   hEdit_MapFunc4, hEdit_MapFunc5, hEdit_MapFunc6, hEdit_MapFunc7, hEdit_MapFunc8, hEdit_MapFunc9,hEdit_MapFunc10,  hEdit_MapFunc11, hFocus;
extern HWND hProp;
extern WNDPROC PropProc;

static unsigned char DirectX_DIuse_cpy;

static int32 Options_MaxWinJoyButtons_cpy; // Snes9xpp Original
extern int32 Options_MaxWinJoyButtons_def; // Snes9xpp Original

bool SetJoyButtonCode(HWND hFocus, WORD KeyCode);

static struct JJoypadButtonRef {
	WORD *Button[12];
	BOOL *Turbo[12];
} JoypadButtonRef[5];

static HWND hJoypadDlg;

/***********************************************************************************************
 Dir Dialog
************************************************************************************************/ 

extern char ROMFileName[];
extern char InitialROMFileDirectory[];
extern char ScreenShotDirectory[];
extern char SPCDumpDirectory[];
extern void SlashAdd(char *PathStr);
extern void SlashRemove(char *PathStr);
extern void RemoveFileName(char *PathStr);

extern "C" const char *S9xBasename (const char *f);
LRESULT CALLBACK BrowseHook(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam);

static char InitialROMFileDirectory_cpy[_MAX_PATH];
static char GUI_FreezeFileDir_cpy[_MAX_PATH];
static char ScreenShotDirectory_cpy[_MAX_PATH];
static char SPCDumpDirectory_cpy[_MAX_PATH];

/***********************************************************************************************
 Other Dialog
************************************************************************************************/ 

static unsigned char Options_SaveImagefileFormat_cpy;
static unsigned char Options_ReduceTheColorTo256_cpy;
static unsigned char Options_ReduceTheColorTo256IfWithin_cpy;
static unsigned char Options_ImagefileHiresStretch_cpy;
static unsigned char Options_RLEcompression256bitmap_cpy;
static unsigned char Options_DisplayScreenBufferAll_cpy;

static unsigned char Xppcdda_Enable_cpy; // Snes9xpp Original
static unsigned char Xppcdda_Drv_cpy; // Snes9xpp Original
static unsigned char Xppsystem_NormalSnap_cpy; // Snes9xpp Original
static unsigned char Settings_SDD1Pack_cpy; // Snes9xpp Original
static unsigned char Xppsystem_DisableIRQCycleCount_cpy; // Snes9xpp Original
static unsigned char Xppsystem_OldAutoFrame_cpy; // Snes9xpp Original
static unsigned char Xppsystem_OldFrameTimer_cpy; // Snes9xpp Original
static unsigned char Xppsystem_FontType_cpy; // Snes9xpp Original
static unsigned char Xppsystem_InactivateWork_cpy; // Snes9xpp Original

/*****************
 default settings
*****************/
extern bool8 Settings_SDD1Pack_def;
extern bool8 Xppsystem_DisableIRQCycleCount_def;
extern bool8 Xppsystem_OldAutoFrame_def;
extern bool8 Xppsystem_OldFrameTimer_def;
extern bool8 Xppsystem_FontType_def;
extern bool8 Xppsystem_InactivateWork_def;

static void OtherDlgCtrlSet(HWND hWnd);



/***********************************************************************************************
 Cfg Dialog
************************************************************************************************/ 

LRESULT CALLBACK SubPropProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDOK:
					if(!SetJoyButtonCode(GetFocus(), VK_RETURN))
						break;
					return 0;
				default:
					break;
			}
			break;

		default:
			break;
	}
	return (CallWindowProc(PropProc, hWnd, uMsg, wParam, lParam));
}

//vpeB[V[g쐬ɌĂ΂
static int CALLBACK InitPropProc(HWND hWnd, UINT uMsg, LPARAM lParam)
{
	static DWORD Style;

	switch (uMsg) {
		case PSCB_INITIALIZED:
			// ftHgłĂwv{^폜
			Style = GetWindowLong(hWnd,GWL_EXSTYLE);
			Style &= ~(WS_EX_CONTEXTHELP);
			SetWindowLong(hWnd,GWL_EXSTYLE,Style);
			SetWindowText(GetDlgItem(hWnd,IDHELP), GUI.Language ? "ݒ胍[h" : "Load default");
			hProp = hWnd;
			PropProc = (WNDPROC)GetWindowLong(hWnd,GWL_WNDPROC);
			break;
		
		default:
			break;
	}
	return 0;
}

static void ConfigDlgSettingsLoad()
{
	S9xwSettingsLoad();
	SNESEmulationSettingsLoad();
	JoySettingsLoad();
	DirectorySettingsLoad();
	OtherSettingsLoad();
	ConfigDlgSettingsSaved = FALSE;
	ConfigDlgDefaultSettingsLoaded = FALSE;

	ConfigChangePageIndex = 0;
}

void ConfigDlgSettingsSave()
{
	if(ConfigDlgSettingsSaved)
		return;
	ConfigDlgSettingsSaved = TRUE;
	
	if(ConfigChangePageIndex & CONFIG_PAGE_SET) S9xwSettingsSave();

	if(ConfigChangePageIndex & CONFIG_PAGE_SNESSET) SNESEmulationSettingsSave();

	if(ConfigChangePageIndex & CONFIG_PAGE_JOYSET) JoySettingsSave();

	if(ConfigChangePageIndex & CONFIG_PAGE_DIRSET) DirectorySettingsSave();

	if(ConfigChangePageIndex & CONFIG_PAGE_OTHER) OtherSettingsSave();

	ConfigChangePageIndex = 0;

}

void ConfigDlgDefaultSettingsLoad(HWND hWnd)
{
	if(ConfigDlgDefaultSettingsLoaded)
		return;
	ConfigDlgDefaultSettingsLoaded = TRUE;
	S9xwDefaultSettingsLoad();
	SNESEmulationDefaultSettingsLoad();
	JoyDefaultSettingsLoad();
	OtherDefaultSettingsLoad();
	PropSheet_Changed(hProp, hWnd);
	ConfigDlgSettingsSaved = FALSE;
}

void ConfigDlgSettingsChanged(HWND hWnd)
{
	PropSheet_Changed(hProp, hWnd);
	ConfigDlgSettingsSaved = FALSE;
	ConfigDlgDefaultSettingsLoaded = FALSE;
}

void ConfigDlgProp(HWND hWnd)
{
	HINSTANCE hInst;
	PROPSHEETPAGE psp;
	PROPSHEETHEADER psh;
	HPROPSHEETPAGE hpsp[5];

	errConfigSwitch = FALSE;

	hInst = (HINSTANCE)GetWindowLong(hWnd, GWL_HINSTANCE);

	psp.dwSize = sizeof(PROPSHEETPAGE);
	psp.dwFlags = PSP_DEFAULT | PSP_HASHELP;
	psp.hInstance = hInst;

	psp.pszTemplate = MAKEINTRESOURCE(GUI.Language ? SETTINGS_DIALOG_JP : SETTINGS_DIALOG_US);
	psp.pfnDlgProc = (DLGPROC)SettingsDlgProc;
	hpsp[0] = CreatePropertySheetPage(&psp);

	psp.pszTemplate = MAKEINTRESOURCE(GUI.Language ? SNESEMULATION_DIALOG_JP : SNESEMULATION_DIALOG_US);
	psp.pfnDlgProc = (DLGPROC)SNESEmulationDlgProc;
	hpsp[1] = CreatePropertySheetPage(&psp);

	psp.pszTemplate = MAKEINTRESOURCE(GUI.Language ? JOYPAD_DIALOG_JP : JOYPAD_DIALOG_US);
	psp.pfnDlgProc = (DLGPROC)JoypadDlgProc;
	hpsp[2] = CreatePropertySheetPage(&psp);

	psp.pszTemplate = MAKEINTRESOURCE(GUI.Language ? DIRECTORY_DIALOG_JP : DIRECTORY_DIALOG_US);
	psp.pfnDlgProc = (DLGPROC)DirectoryDlgProc;
	hpsp[3] = CreatePropertySheetPage(&psp);

	psp.pszTemplate = MAKEINTRESOURCE(GUI.Language ? OTHER_DIALOG_JP : OTHER_DIALOG_US);
	psp.pfnDlgProc = (DLGPROC)OtherDlgProc;
	hpsp[4] = CreatePropertySheetPage(&psp);

	FillMemory((PVOID)&psh, sizeof(PROPSHEETHEADER), 0);
	psh.dwSize = sizeof(PROPSHEETHEADER);
	psh.dwFlags = PSH_USEICONID | PSH_USECALLBACK;
	psh.hInstance = hInst;
	psh.hwndParent = hWnd;
	psh.pszIcon = MAKEINTRESOURCE(SETTINGS_ICON);
	psh.nPages = 5;
	psh.nStartPage = StartPageIndex;
	psh.phpage = hpsp;
	psh.pfnCallback = InitPropProc;
	psh.pszCaption = (GUI.Language ? "ݒ" : "Configuration Settings");

	ConfigDlgSettingsLoad();

	PropertySheet(&psh);
	return;
}


/***********************************************************************************************
 Set Dialog
************************************************************************************************/ 

void S9xwSettingsLoad()
{
	//sound settings
	GUI_SoundMode_cpy = GUI.SoundMode;
	Settings_SoundBufferSize_cpy = Settings.SoundBufferSize;
	Settings_SoundMixInterval_cpy = Settings.SoundMixInterval;

	//Display settings
	DoubleRenderingModes16_cpy = DoubleRenderingModes16;
	ForceWindowedDirectDrawClipper_cpy = ForceWindowedDirectDrawClipper;
	dRendering_cpy = dRendering;
	DirectX_DisableHardwareScale_cpy = DirectX.DisableHardwareScale;
	GUI_DoubleBuffered_cpy = GUI.DoubleBuffered;
	DirectX_OffScreenVRAM_cpy = DirectX.OffScreenVRAM;
	DirectX_BackBufferVRAM_cpy = DirectX.BackBufferVRAM;

	//Thread
	GUI_PriorityClass_cpy = GUI.PriorityClass;
	GUI_ThreadPriority_cpy = GUI.ThreadPriority;
	
	Options_CpuSleep_cpy = Options.CpuSleep; // Snes9xpp Original
	CPU_DisableSpinWaitLoop_cpy = CPU.DisableSpinWaitLoop; // Snes9xpp Original
}

static HWND hSettingsDlg;

void S9xwDefaultSettingsLoad()
{
	//sound settings
	GUI_SoundMode_cpy = GUI_SoundMode_def;
	Settings_SoundBufferSize_cpy = Settings_SoundBufferSize_def;
	Settings_SoundMixInterval_cpy = Settings_SoundMixInterval_def;

	//Display settings
	DoubleRenderingModes16_cpy = DoubleRenderingModes16_def;
	ForceWindowedDirectDrawClipper_cpy = ForceWindowedDirectDrawClipper_def;
	dRendering_cpy = dRendering_def;
	DirectX_DisableHardwareScale_cpy = DirectX_DisableHardwareScale_def;
	GUI_DoubleBuffered_cpy = GUI_DoubleBuffered_def;
	DirectX_OffScreenVRAM_cpy = DirectX_OffScreenVRAM_def;
	DirectX_BackBufferVRAM_cpy = DirectX_BackBufferVRAM_def;

	//Thread
	GUI_PriorityClass_cpy = GUI_PriorityClass_def;
	GUI_ThreadPriority_cpy = GUI_ThreadPriority_def;
	
	Options_CpuSleep_cpy = Options_CpuSleep_def; // Snes9xpp Original
	CPU_DisableSpinWaitLoop_cpy = CPU_DisableSpinWaitLoop_def; // Snes9xpp Original

	SettingsDlgCtrlSet(hSettingsDlg);
}

void S9xwSettingsSave()
{
	SelectRenderingCodes();
	ForceWindowedDirectDrawClipper = ForceWindowedDirectDrawClipper_cpy;
	dRendering = dRendering_cpy;
	if(dRendering)
		DirectX.ForceWindowedDirectDrawClipper = FALSE;
	else
		DirectX.ForceWindowedDirectDrawClipper = ForceWindowedDirectDrawClipper;
	DirectX.DisableHardwareScale = DirectX_DisableHardwareScale_cpy;
	GUI.DoubleBuffered = GUI_DoubleBuffered_cpy;
	DirectX.BackBufferVRAM = DirectX_BackBufferVRAM_cpy;
	TransparencyFlag = TransparencyFlag_cpy;
	Settings.Transparency = TransparencyFlag;

	if(DoubleRenderingModes16_cpy != DoubleRenderingModes16 
		||  DirectX.OffScreenVRAM != DirectX_OffScreenVRAM_cpy) {
		DoubleRenderingModes16 = DoubleRenderingModes16_cpy;
		DirectX.OffScreenVRAM = DirectX_OffScreenVRAM_cpy;
		if(Options.GFXengine2 != Options_GFXengine2_cpy) {
			Options.GFXengine2 = Options_GFXengine2_cpy;
			S9xSwitchGFXengine(Options.GFXengine2);
		}
		//2xSaI[hւ̐؂芷̏ꍇ̃XN[obt@NA΍
		ResetRenderMethod();
		GUI.Scale = (DoubleRenderingModes16);
		CurFullscreenDisplayModeIndexChange();
		if(GUI.FullScreen && !DirectX.Windowed)
		{
			if(!(errConfigSwitch = !S9xwSwitchScreenMode(FALSE)))
				GUI.FullScreen = TRUE;
				UpdateBackBuffer();
		}
		else
		if(GUI.FullScreen)
		{
			if(!(errConfigSwitch = !S9xwSwitchScreenMode(TRUE)))
				GUI.FullScreen = TRUE;
		}
		else
		{
			if(!(errConfigSwitch = !S9xwSwitchScreenMode(TRUE)))
				GUI.FullScreen = FALSE;
		}
	}

	if(Options.GFXengine2 != Options_GFXengine2_cpy) {
		Options.GFXengine2 = Options_GFXengine2_cpy;
		S9xSwitchGFXengine(Options.GFXengine2);
	}

	if(GUI.SoundMode != GUI_SoundMode_cpy || Settings.SoundBufferSize != Settings_SoundBufferSize_cpy || Settings.SoundMixInterval != Settings_SoundMixInterval_cpy) {
		GUI.SoundMode = GUI_SoundMode_cpy;
		Settings.SoundBufferSize = Settings_SoundBufferSize_cpy;
		Settings.SoundMixInterval = Settings_SoundMixInterval_cpy;
		SetupSound( SoundRates[GUI.SoundMode].rate, SoundRates[GUI.SoundMode].sixteen_bit, SoundRates[GUI.SoundMode].stereo);
	}
	GUI.PriorityClass = GUI_PriorityClass_cpy;
	GUI.ThreadPriority = GUI_ThreadPriority_cpy;
	
	Options.CpuSleep = Options_CpuSleep_cpy; // Snes9xpp Original
	CPU.DisableSpinWaitLoop = CPU_DisableSpinWaitLoop_cpy; // Snes9xpp Original
}


LRESULT CALLBACK SettingsDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam)
{
	static int i;
	static char *SoundModeStr[] = {"8000Hz 8bit mono", "11025Hz 8bit mono", "22050Hz 8bit mono", "32000Hz 8bit mono", "44100Hz 8bit mono", "48000Hz 8bit mono",
									 "8000Hz 16bit mono", "11025Hz 16bit mono", "22050Hz 16bit mono", "32000Hz 16bit mono", "44100Hz 16bit mono", "48000Hz 16bit mono",
									"8000Hz 8bit stereo", "11025Hz 8bit stereo", "22050Hz 8bit stereo", "32000Hz 8bit stereo", "44100Hz 8bit stereo", "48000Hz 8bit stereo",
									"8000Hz 16bit stereo", "11025Hz 16bit stereo", "22050Hz 16bit stereo", "32000Hz 16bit stereo", "44100Hz 16bit stereo", "48000Hz 16bit stereo"};
	static char *SoundBufferLengthStr[] = {"10 ms", "20 ms", "40 ms", "80 ms", "160 ms", "320 ms", "640 ms"};
	static char *SoundMixIntervalStr[] = {"10 ms", "20 ms", "30 ms", "40 ms", "50 ms", "60 ms", "70 ms", "80 ms", "90 ms", "100 ms", "110 ms", "120 ms", "130 ms", "140 ms", "150 ms"};
	static char *PriorityClassStr[] = {"IDLE_PRIORITY_CLASS", "NORMAL_PRIORITY_CLASS", "HIGH_PRIORITY_CLASS", "REALTIME_PRIORITY_CLASS"};
	static char *ThreadPriorityStr[] = {"THREAD_PRIORITY_LOWEST", "THREAD_PRIORITY_BELOW_NORMAL", "THREAD_PRIORITY_NORMAL", "THREAD_PRIORITY_ABOVE_NORMAL", "THREAD_PRIORITY_HIGHEST", "THREAD_PRIORITY_IDLE", "THREAD_PRIORITY_TIME_CRITICAL"};

	switch (uMsg) {
		case WM_INITDIALOG:
			hSettingsDlg = hWnd;
			//TEh[hR{{bNXɃeLXgZbg
			for ( i = 0; i < 24; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDMODE), CB_INSERTSTRING, (WPARAM)i, (LPARAM)SoundModeStr[i]);
			//SoundBufferLengthR{{bNXɃeLXgZbg
			for ( i = 0; i < 7; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDBUFFERLENGTH), CB_INSERTSTRING, (WPARAM)i, (LPARAM)SoundBufferLengthStr[i]);
			//SoundMixIntervalR{{bNXɃeLXgZbg
			for ( i = 0; i < 15; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDMIXINTERVAL), CB_INSERTSTRING, (WPARAM)i, (LPARAM)SoundMixIntervalStr[i]);

			//Thread
			for (i = 0; i < 4; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_PRIORITYCLASS), CB_INSERTSTRING, (WPARAM)i, (LPARAM)PriorityClassStr[i]);
			for (i = 0; i < 7; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_THREADPRIORITY), CB_INSERTSTRING, (WPARAM)i, (LPARAM)ThreadPriorityStr[i]);
			
			SettingsDlgCtrlSet(hWnd);
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDC_RADIO_2X_HARDWARE:
				case IDC_RADIO_2X_SOFTWARE:
				case IDC_RADIO_2X_SCANLINES:
				case IDC_RADIO_2X_EAGLE:
				case IDC_RADIO_2X_SAI:
				case IDC_RADIO_SUPER_2X_SAI:
				case IDC_RADIO_SCALE_2X_SAI:
				case IDC_RADIO_2X_TVMODE:
				case IDC_RADIO_2X_HQ2X:
				case IDC_RADIO_2X_LQ2X:
				case IDC_RADIO_2X_INTERLACE_SCANLINES:
				case IDC_RADIO_2X_INTERLACE:

				case IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER:
				case IDC_CHECK_DOUBLEBUFFERING:
				case IDC_CHECK_DIRECTSCREENRENDERING:
				case IDC_CHECK_DISABLEHARDWARESCALE:
				case IDC_CHECK_OFFSCREENVRAM:
				case IDC_CHECK_BACKBUFFERVRAM:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
						if(LOWORD(wParam) == IDC_RADIO_2X_HARDWARE ||
								LOWORD(wParam) == IDC_RADIO_2X_SOFTWARE ||
								LOWORD(wParam) == IDC_RADIO_2X_SCANLINES ||
								LOWORD(wParam) == IDC_RADIO_2X_INTERLACE_SCANLINES ||
								LOWORD(wParam) == IDC_RADIO_2X_INTERLACE ||
								LOWORD(wParam) == IDC_RADIO_2X_EAGLE ||
								LOWORD(wParam) == IDC_RADIO_2X_SAI ||
								LOWORD(wParam) == IDC_RADIO_SUPER_2X_SAI ||
								LOWORD(wParam) == IDC_RADIO_SCALE_2X_SAI ||
								LOWORD(wParam) == IDC_RADIO_2X_TVMODE ||
								LOWORD(wParam) == IDC_RADIO_2X_HQ2X ||
								LOWORD(wParam) == IDC_RADIO_2X_LQ2X) {
							if(LOWORD(wParam) == IDC_RADIO_2X_HARDWARE) {
									DoubleRenderingModes16_cpy = 1;
							}
							else if(LOWORD(wParam) == IDC_RADIO_2X_SOFTWARE) {
									DoubleRenderingModes16_cpy = 2;
							}
							else if(LOWORD(wParam) == IDC_RADIO_2X_SCANLINES) {
									DoubleRenderingModes16_cpy = 3;
							}
							else if(LOWORD(wParam) == IDC_RADIO_2X_EAGLE)
								DoubleRenderingModes16_cpy = 4;
							else if(LOWORD(wParam) == IDC_RADIO_2X_SAI)
								DoubleRenderingModes16_cpy = 5;
							else if(LOWORD(wParam) == IDC_RADIO_SUPER_2X_SAI)
								DoubleRenderingModes16_cpy = 6;
							else if(LOWORD(wParam) == IDC_RADIO_SCALE_2X_SAI)
								DoubleRenderingModes16_cpy = 7;
							else if(LOWORD(wParam) == IDC_RADIO_2X_TVMODE)
								DoubleRenderingModes16_cpy = 8;
							else if(LOWORD(wParam) == IDC_RADIO_2X_HQ2X)
								DoubleRenderingModes16_cpy = 9;
							else if(LOWORD(wParam) == IDC_RADIO_2X_LQ2X)
								DoubleRenderingModes16_cpy = 10;
							else if(LOWORD(wParam) == IDC_RADIO_2X_INTERLACE_SCANLINES)
								DoubleRenderingModes16_cpy = 11;
							else if(LOWORD(wParam) == IDC_RADIO_2X_INTERLACE)
								DoubleRenderingModes16_cpy = 12;
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_HARDWARE), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_HARDWARE ? 1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SOFTWARE), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_SOFTWARE ? 1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SCANLINES), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_SCANLINES ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_EAGLE), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_EAGLE ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SAI), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_SAI ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_SUPER_2X_SAI), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_SUPER_2X_SAI ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_SCALE_2X_SAI), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_SCALE_2X_SAI ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_TVMODE), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_TVMODE ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_HQ2X), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_HQ2X ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_LQ2X), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_LQ2X ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_INTERLACE_SCANLINES), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_INTERLACE_SCANLINES ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_INTERLACE), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_INTERLACE ?  1 : 0, 0);
							EnableWindow(GetDlgItem(hWnd, IDC_CHECK_DISABLEHARDWARESCALE), ((DoubleRenderingModes16_cpy == 1)) ? FALSE : TRUE);
							if(DoubleRenderingModes16_cpy == DoubleRenderingModes16)
								break;
						}
						else if(LOWORD(wParam) == IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER)
						{
							ForceWindowedDirectDrawClipper_cpy = tmpVal;
						}
						else if(LOWORD(wParam) == IDC_CHECK_DOUBLEBUFFERING)
						{
							GUI_DoubleBuffered_cpy = (tmpVal ? TRUE : FALSE);
						}
						else if(LOWORD(wParam) == IDC_CHECK_DIRECTSCREENRENDERING)
						{
							dRendering_cpy = tmpVal;
							EnableWindow(GetDlgItem(hWnd, IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER), dRendering_cpy ? FALSE : TRUE);
						}
						else if(LOWORD(wParam) == IDC_CHECK_DISABLEHARDWARESCALE)
						{
							DirectX_DisableHardwareScale_cpy = (unsigned char)tmpVal;
						}
						else if(LOWORD(wParam) == IDC_CHECK_OFFSCREENVRAM)
						{
							DirectX_OffScreenVRAM_cpy = (tmpVal ? TRUE : FALSE);
						}
						else if(LOWORD(wParam) == IDC_CHECK_BACKBUFFERVRAM)
						{
							DirectX_BackBufferVRAM_cpy = (tmpVal ? TRUE : FALSE);
						}
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				case IDC_COMBO_SOUNDMODE:
				case IDC_COMBO_SOUNDBUFFERLENGTH:
				case IDC_COMBO_SOUNDMIXINTERVAL:
				case IDC_COMBO_PRIORITYCLASS:
				case IDC_COMBO_THREADPRIORITY:
					if(HIWORD(wParam) == CBN_SELENDOK) {
						i = (int)SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0);
						if(LOWORD(wParam) == IDC_COMBO_SOUNDMODE)
							GUI_SoundMode_cpy = (uint8)i;
						else if(LOWORD(wParam) == IDC_COMBO_SOUNDBUFFERLENGTH)
							Settings_SoundBufferSize_cpy = (1 << i);
						else if(LOWORD(wParam) == IDC_COMBO_SOUNDMIXINTERVAL)
							Settings_SoundMixInterval_cpy = (i + 1) * 10;
						else if(LOWORD(wParam) == IDC_COMBO_PRIORITYCLASS) {
							if(i == 0)
								GUI_PriorityClass_cpy = IDLE_PRIORITY_CLASS;
							else if(i == 1)
								GUI_PriorityClass_cpy = NORMAL_PRIORITY_CLASS;
							else if(i == 2)
								GUI_PriorityClass_cpy = HIGH_PRIORITY_CLASS;
							else if(i == 3)
								GUI_PriorityClass_cpy = REALTIME_PRIORITY_CLASS;
						}
						else if(LOWORD(wParam) == IDC_COMBO_THREADPRIORITY) {
							if(i == 0)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_LOWEST;
							else if(i == 1)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_BELOW_NORMAL;
							else if(i == 2)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_NORMAL;
							else if(i == 3)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_ABOVE_NORMAL;
							else if(i == 4)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_HIGHEST;
							else if(i == 5)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_IDLE;
							else if(i == 6)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_TIME_CRITICAL;
						}
						if(GUI_SoundMode_cpy != GUI.SoundMode || Settings_SoundBufferSize_cpy != Settings.SoundBufferSize ||
							Settings_SoundMixInterval_cpy != Settings.SoundMixInterval || GUI_PriorityClass_cpy != GUI.PriorityClass ||
							GUI_ThreadPriority_cpy != GUI.ThreadPriority) {
							ConfigDlgSettingsChanged(hWnd);
						}
					}
					break;

 // Snes9xpp Original Start
				case IDC_CHECK_CPU_SLEEP:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
                                                if(LOWORD(wParam) == IDC_CHECK_CPU_SLEEP)
							Options_CpuSleep_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;
				case IDC_CHECK_CPU_DISABLE_SPINLOOP:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
                                                if(LOWORD(wParam) == IDC_CHECK_CPU_DISABLE_SPINLOOP)
							CPU_DisableSpinWaitLoop_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;
 // Snes9xpp Original End

				default:
					break;
			}
			break;

		case WM_NOTIFY:
			switch(((NMHDR *)lParam)->code) {
				case PSN_APPLY:
					ConfigChangePageIndex |= CONFIG_PAGE_SET;
					ConfigDlgSettingsSave();
					break;

				case PSN_HELP:
					ConfigDlgDefaultSettingsLoad(hWnd);
					break;

				case PSN_QUERYCANCEL:
					break;

				case PSN_SETACTIVE:
					ConfigChangePageIndex |= CONFIG_PAGE_SET;
						Scale2X16RadioCtrlSet(hWnd);
					EnableWindow(GetDlgItem(hWnd, IDC_CHECK_DISABLEHARDWARESCALE), ((DoubleRenderingModes16_cpy == 1)) ? FALSE : TRUE);
					StartPageIndex = 0;
					break;

				default:
					break;
			}
			break;

		default:
			break;
	}
	return FALSE;
}

static void SettingsDlgCtrlSet(HWND hWnd)
{
	int i, c;
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDMODE), CB_SETCURSEL, (WPARAM)GUI_SoundMode_cpy, 0);
	for(c = i = 1; i != Settings_SoundBufferSize_cpy && i < 64; c++, i <<= 1);
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDBUFFERLENGTH), CB_SETCURSEL, (WPARAM)(c - 1), 0);
	i = ((Settings_SoundMixInterval_cpy / 10) - 1);
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDMIXINTERVAL), CB_SETCURSEL, (WPARAM)i, 0);

		Scale2X16RadioCtrlSet(hWnd);

	SendMessage(GetDlgItem(hWnd, IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER), BM_SETCHECK, (WPARAM)(ForceWindowedDirectDrawClipper_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_DOUBLEBUFFERING), BM_SETCHECK, (WPARAM)(GUI_DoubleBuffered_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_DIRECTSCREENRENDERING), BM_SETCHECK, (WPARAM)(dRendering_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_DISABLEHARDWARESCALE), BM_SETCHECK, (WPARAM)(DirectX_DisableHardwareScale_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_OFFSCREENVRAM), BM_SETCHECK, (WPARAM)(DirectX_OffScreenVRAM_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_BACKBUFFERVRAM), BM_SETCHECK, (WPARAM)(DirectX_BackBufferVRAM_cpy ? 1 : 0), 0);

	EnableWindow(GetDlgItem(hWnd, IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER), dRendering_cpy ? FALSE : TRUE);
	EnableWindow(GetDlgItem(hWnd, IDC_CHECK_DISABLEHARDWARESCALE), ((DoubleRenderingModes16_cpy == 1)) ? FALSE : TRUE);

	if(GUI_PriorityClass_cpy == IDLE_PRIORITY_CLASS)
		i = 0;
	else if(GUI_PriorityClass_cpy == NORMAL_PRIORITY_CLASS)
		i = 1;
	else if(GUI_PriorityClass_cpy == HIGH_PRIORITY_CLASS)
		i = 2;
	else if(GUI_PriorityClass_cpy == REALTIME_PRIORITY_CLASS)
		i = 3;
	else
		i = 1;
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_PRIORITYCLASS), CB_SETCURSEL, (WPARAM)i, 0);

	if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_LOWEST)
		i = 0;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_BELOW_NORMAL)
		i = 1;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_NORMAL)
		i = 2;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_ABOVE_NORMAL)
		i = 3;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_HIGHEST)
		i = 4;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_IDLE)
		i = 5;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_TIME_CRITICAL)
		i = 6;
	else
		i = 2;
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_THREADPRIORITY), CB_SETCURSEL, (WPARAM)i, 0);
	
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_CPU_SLEEP), BM_SETCHECK, (WPARAM)(Options_CpuSleep_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_CPU_DISABLE_SPINLOOP), BM_SETCHECK, (WPARAM)(CPU_DisableSpinWaitLoop_cpy ? 1 : 0), 0);
}

static void Scale2X16RadioCtrlSet(HWND hWnd)
{
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_HARDWARE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 1) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SOFTWARE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 2) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SCANLINES), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 3) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_EAGLE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 4) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SAI), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 5) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_SUPER_2X_SAI), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 6) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_SCALE_2X_SAI), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 7) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_TVMODE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 8) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_HQ2X), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 9) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_LQ2X), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 10) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_INTERLACE_SCANLINES), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 11) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_INTERLACE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 12) ? 1 : 0), 0);
}

/***********************************************************************************************
 Snes Dialog
************************************************************************************************/ 

void SNESEmulationSettingsLoad()
{
	//graphics
	Settings_Mode7Interpolate_cpy = Settings.Mode7Interpolate;
	TransparencyFlag_cpy = TransparencyFlag;
	Settings_BGLayering_cpy = Settings.BGLayering;
	Settings_DisableGraphicWindows_cpy = Settings.DisableGraphicWindows;
	Settings_DisableHDMA_cpy = Settings.DisableHDMA;
	Options_GFXengine2_cpy = Options.GFXengine2;
	DirectX_wait_for_vsync_cpy = DirectX.wait_for_vsync;

	//sound
	Settings_InterpolatedSound_cpy = Settings.InterpolatedSound;
	Settings_NextAPUEnabled_cpy = Settings.NextAPUEnabled;
	Settings_DisableSoundEcho_cpy = Settings.DisableSoundEcho;
	Settings_ReverseStereo_cpy = Settings.ReverseStereo;
	Settings_SoundSync_cpy = Settings.SoundSync;
	Settings_DisableMasterVolume_cpy = Settings.DisableMasterVolume;
	Settings_SoundEnvelopeHeightReading_cpy = Settings.SoundEnvelopeHeightReading;
	Settings_SoundSkipMethod_cpy = Settings.SoundSkipMethod;

	//cpu
	Settings_ShutdownMaster_cpy = Settings.ShutdownMaster;

	//controller
	Settings_ControllerOption_cpy = Settings.ControllerOption;
	Settings_MultiPlayer5_cpy = Settings.MultiPlayer5;
	Settings_Mouse_cpy = Settings.Mouse;
	Settings_SuperScope_cpy = Settings.SuperScope;
	Settings_Justifier_cpy = Settings.Justifier;
	Options_JoyReadRateMask_cpy = Options.JoyReadRateMask;

	//timing
	Settings_SkipFrames_cpy = Settings.SkipFrames;
	Settings_AutoMaxSkipFrames_cpy = Settings.AutoMaxSkipFrames;
	Settings_CyclesPercentage_cpy = Settings.CyclesPercentage;
	Settings_FrameTimePAL_cpy = Settings.FrameTimePAL;
	Settings_FrameTimeNTSC_cpy = Settings.FrameTimeNTSC;

	//AutoSave
	Settings_AutoSaveDelay_cpy = Settings.AutoSaveDelay;

}

static HWND hSNESEmulationDlg;

void SNESEmulationDefaultSettingsLoad()
{
	//graphics
	Settings_Mode7Interpolate_cpy = Settings_Mode7Interpolate_def;
	TransparencyFlag_cpy = TransparencyFlag_def;
	Settings_BGLayering_cpy = Settings_BGLayering_def;
	Settings_DisableGraphicWindows_cpy = Settings_DisableGraphicWindows_def;
	Settings_DisableHDMA_cpy = Settings_DisableHDMA_def;
	Options_GFXengine2_cpy = OPTIONS_GFXENGINE2_DEF;
	DirectX_wait_for_vsync_cpy = DirectX_wait_for_vsync_def;

	//sound
	Settings_InterpolatedSound_cpy = Settings_InterpolatedSound_def;
	Settings_NextAPUEnabled_cpy = Settings_NextAPUEnabled_def;
	Settings_DisableSoundEcho_cpy = Settings_DisableSoundEcho_def;
	Settings_ReverseStereo_cpy = Settings_ReverseStereo_def;
	Settings_SoundSync_cpy = Settings_SoundSync_def;
	Settings_DisableMasterVolume_cpy = Settings_DisableMasterVolume_def;
	Settings_SoundEnvelopeHeightReading_cpy = Settings_SoundEnvelopeHeightReading_def;
	Settings_SoundSkipMethod_cpy = Settings_SoundSkipMethod_def;

	//cpu
	Settings_ShutdownMaster_cpy = Settings_ShutdownMaster_def;

	//controller
	Settings_ControllerOption_cpy = Settings_ControllerOption_def;
	Settings_MultiPlayer5_cpy = Settings_MultiPlayer5_def;
	Settings_Mouse_cpy = Settings_Mouse_def;
	Settings_SuperScope_cpy = Settings_SuperScope_def;
	Settings_Justifier_cpy = Settings_Justifier_def;
	Options_JoyReadRateMask_cpy = OPTIONS_JOYREADRATEMASK_DEF;

	//timing
	Settings_SkipFrames_cpy = Settings_SkipFrames_def;
	Settings_AutoMaxSkipFrames_cpy = Settings_AutoMaxSkipFrames_def;
	Settings_CyclesPercentage_cpy = Settings_CyclesPercentage_def;
	Settings_FrameTimePAL_cpy = Settings_FrameTimePAL_def;
	Settings_FrameTimeNTSC_cpy = Settings_FrameTimeNTSC_def;

	//AutoSave Snes9xpp Original
	Settings_AutoSaveDelay_cpy = Settings_AutoSaveDelay_def;

	SNESEmulationDlgCtrlSet(hSNESEmulationDlg);
}

void SNESEmulationSettingsSave()
{
	static char APUEnabledMessage_US[] = "Sound emulation setting chenged to enabled\nsound will generate after Reset.";
	static char APUEnabledMessage_JP[] = "TEhG~[VLɐݒύX܂\nZbgɏԂɔf܂";
	static char APUDisabledMessage_US[] = "Sound emulation setting chenged to disabled\nsound will generate after Reset in case the setting will be enabled again, would you continue?";
	static char APUDisabledMessage_JP[] = "TEhG~[VɐݒύX܂\n߂ėLɂꍇ̓ZbgɏԂɔfɂȂ܂Aɂ܂H";
	static short tmpVolume;
	static uint32 tmpnumval;

	Settings.Mode7Interpolate = Settings_Mode7Interpolate_cpy;
	Settings.BGLayering = Settings_BGLayering_cpy;
	Settings.DisableGraphicWindows = Settings_DisableGraphicWindows_cpy;
	Settings.DisableHDMA = Settings_DisableHDMA_cpy;
	DirectX.wait_for_vsync = DirectX_wait_for_vsync_cpy;
	Settings.InterpolatedSound = Settings_InterpolatedSound_cpy;
	Settings.AutoMaxSkipFrames = Settings_AutoMaxSkipFrames_cpy;

	if(Settings_NextAPUEnabled_cpy != Settings.NextAPUEnabled) {
		if(!GUI.StopEmulation && Settings_NextAPUEnabled_cpy) {
			MessageBox(hSNESEmulationDlg, GUI.Language ? APUEnabledMessage_JP : APUEnabledMessage_US, "Snes9xpp - Enabled sound emulation", MB_OK | MB_ICONINFORMATION);
			Settings.NextAPUEnabled = TRUE;
		}
		else if(!GUI.StopEmulation && !Settings_NextAPUEnabled_cpy) {
			if(IDOK == MessageBox(hSNESEmulationDlg, GUI.Language ? APUDisabledMessage_JP : APUDisabledMessage_US, "Snes9xpp - Disabled sound emulation", MB_OKCANCEL | MB_ICONQUESTION)) {
				//kill APU executing
				IAPU.APUExecuting = Settings.APUEnabled = Settings.NextAPUEnabled = FALSE;
				S9xResetAPU ();
			}
			else {
				Settings_NextAPUEnabled_cpy = TRUE;
				SendMessage(GetDlgItem(hSNESEmulationDlg, IDC_CHECK_APUENABLED), BM_SETCHECK, (WPARAM)(Settings_NextAPUEnabled_cpy ? 1 : 0), 0);
			}
		}
		else
			Settings.APUEnabled = Settings.NextAPUEnabled = Settings_NextAPUEnabled_cpy;
	}
	Settings.DisableSoundEcho = Settings_DisableSoundEcho_cpy;
	if(Settings_ReverseStereo_cpy != Settings.ReverseStereo) {
		tmpVolume = SoundData.master_volume [0];
		SoundData.master_volume [0] = SoundData.master_volume [1];
		SoundData.master_volume [1] = tmpVolume;
		tmpVolume = SoundData.echo_volume [0];
		SoundData.echo_volume [0] = SoundData.echo_volume [1];
		SoundData.echo_volume [1] = tmpVolume;
		Settings.ReverseStereo = Settings_ReverseStereo_cpy;
	}
	Settings.SoundSync = Settings_SoundSync_cpy;
	Settings.DisableMasterVolume = Settings_DisableMasterVolume_cpy;
	Settings.SoundEnvelopeHeightReading = Settings_SoundEnvelopeHeightReading_cpy;
	if(!GUI.StopEmulation && !(!Settings.Shutdown && Settings.ShutdownMaster))
		Settings.Shutdown = Settings.ShutdownMaster = Settings_ShutdownMaster_cpy;
	else
		Settings.ShutdownMaster = Settings_ShutdownMaster_cpy;
	Settings.SoundSkipMethod = Settings_SoundSkipMethod_cpy;
	if(!GUI.StopEmulation && 
		(Settings.ControllerOption != Settings_ControllerOption_cpy ||
		Settings.MultiPlayer5 != Settings_MultiPlayer5_cpy ||
		Settings.Mouse != Settings_Mouse_cpy || 
		Settings.Justifier != Settings_Justifier_cpy ||
		Settings.SuperScope != Settings_SuperScope_cpy)) {
		if(!(!Settings.MultiPlayer5Master && Settings.MultiPlayer5))
			Settings.MultiPlayer5Master = Settings_MultiPlayer5_cpy;
		if(!(!Settings.MouseMaster && Settings.Mouse))
			Settings.MouseMaster = Settings_Mouse_cpy;
		if(!(!Settings.SuperScopeMaster && Settings.SuperScope))
			Settings.SuperScopeMaster = Settings_SuperScope_cpy;
		if(!(!Settings.JustifierMaster && Settings.Justifier))
			Settings.JustifierMaster = Settings_Justifier_cpy;
		if (Settings_ControllerOption_cpy == 0)
			IPPU.Controller = SNES_MAX_CONTROLLER_OPTIONS - 1;
		else
			IPPU.Controller = Settings_ControllerOption_cpy - 1;
		S9xNextController ();
	}
	Settings.ControllerOption = Settings_ControllerOption_cpy;
	Settings.MultiPlayer5 = Settings_MultiPlayer5_cpy;
	Settings.Mouse = Settings_Mouse_cpy;
	Settings.SuperScope = Settings_SuperScope_cpy;
	Settings.Justifier = Settings_Justifier_cpy;

	Options.JoyReadRateMask = Options_JoyReadRateMask_cpy;
	Settings.SkipFrames = Settings_SkipFrames_cpy;

	//CyclePercentage
	GetWindowText(GetDlgItem(hSNESEmulationDlg, IDC_EDIT_CYCLEPERCENTAGE), tmpcharbuf, 255);
	tmpnumval = (uint32)atoi(tmpcharbuf);
	if(tmpnumval >= 1 && tmpnumval <= 199) {
		Settings_CyclesPercentage_cpy = (long)tmpnumval;
		if(!GUI.StopEmulation && Settings.H_Max == (long)((SNES_CYCLES_PER_SCANLINE * Settings.CyclesPercentage) / 100)) {
			Settings.H_Max = (SNES_CYCLES_PER_SCANLINE * Settings_CyclesPercentage_cpy) / 100;
			Settings.HBlankStart = (256 * Settings.H_Max) / SNES_HCOUNTER_MAX;
		}
		Settings.CyclesPercentage = Settings_CyclesPercentage_cpy;
	}
	else {
		Settings_CyclesPercentage_cpy = Settings.CyclesPercentage;
	}
	wsprintf(tmpcharbuf, "%d", Settings_CyclesPercentage_cpy);
	SetWindowText(GetDlgItem(hSNESEmulationDlg, IDC_EDIT_CYCLEPERCENTAGE), tmpcharbuf);

	//PAL timing ms
	GetWindowText(GetDlgItem(hSNESEmulationDlg, IDC_EDIT_PAL), tmpcharbuf, 255);
	tmpnumval = (uint32)atoi(tmpcharbuf);
	if(tmpnumval >= 1 && tmpnumval <= 999) {
		Settings_FrameTimePAL_cpy = tmpnumval;
		if(!GUI.StopEmulation && Settings.FrameTimePAL == Settings.FrameTime)
			Settings.FrameTime = Settings_FrameTimePAL_cpy;
		Settings.FrameTimePAL = Settings_FrameTimePAL_cpy;
	}
	else {
		Settings_FrameTimePAL_cpy = Settings.FrameTimePAL;
	}
	wsprintf(tmpcharbuf, "%d", Settings_FrameTimePAL_cpy);
	SetWindowText(GetDlgItem(hSNESEmulationDlg, IDC_EDIT_PAL), tmpcharbuf);

	//NTSC timing ms
	GetWindowText(GetDlgItem(hSNESEmulationDlg, IDC_EDIT_NTSC), tmpcharbuf, 255);
	tmpnumval = (uint32)atoi(tmpcharbuf);
	if(tmpnumval >= 1 && tmpnumval <= 999) {
		Settings_FrameTimeNTSC_cpy = tmpnumval;
		if(!GUI.StopEmulation && Settings.FrameTimeNTSC == Settings.FrameTime)
			Settings.FrameTime = Settings_FrameTimeNTSC_cpy;
		Settings.FrameTimeNTSC = Settings_FrameTimeNTSC_cpy;
	}
	else {
		Settings_FrameTimeNTSC_cpy = Settings.FrameTimeNTSC;
	}
	wsprintf(tmpcharbuf, "%d", Settings_FrameTimeNTSC_cpy);
	SetWindowText(GetDlgItem(hSNESEmulationDlg, IDC_EDIT_NTSC), tmpcharbuf);

	GetWindowText(GetDlgItem(hSNESEmulationDlg, IDC_EDIT_AUTOSAVESRAM), tmpcharbuf, 255);
	tmpnumval = (int32)atoi(tmpcharbuf);
	if(tmpnumval >= 0 && tmpnumval <= 9999999) {
		Settings.AutoSaveDelay = Settings_AutoSaveDelay_cpy = tmpnumval;
	}
	else {
		Settings_AutoSaveDelay_cpy = Settings.AutoSaveDelay;
	}
	wsprintf(tmpcharbuf, "%d", Settings_AutoSaveDelay_cpy);
	SetWindowText(GetDlgItem(hSNESEmulationDlg, IDC_EDIT_AUTOSAVESRAM), tmpcharbuf);
}

LRESULT CALLBACK SNESEmulationDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam)
{
	static char *SoundSkipMethodStr[] = {"Method 0", "Method 1", "Method 2", "Method 3"};
	static char *JoyReadRateStr_US[] = {"1 frame", "2 frame", "4 frame", "8 frame"};
	static char *JoyReadRateStr_JP[] = {"1 t[", "2 t[", "4 t[", "8 t["};
	static char *SkipFramesStr_US[] = {"Auto FrameSkip", "0 FrameSkip", "1 FrameSkip", "2 FrameSkip", "3 FrameSkip", "4 FrameSkip", "5 FrameSkip", "6 FrameSkip", "7 FrameSkip", "8 FrameSkip", "9 FrameSkip"};
	static char *SkipFramesStr_JP[] = {"t[XLbv", "0 t[XLbv", "1 t[XLbv", "2 t[XLbv", "3 t[XLbv", "4 t[XLbv", "5 t[XLbv", "6 t[XLbv", "7 t[XLbv", "8 t[XLbv", "9 t[XLbv"};
	static char *AutoMaxSkipFramesStr[] = {"Max 0", "Max 1", "Max 2", "Max 3", "Max 4", "Max 5", "Max 6", "Max 7", "Max 8", "Max 9", "Max 10"};
	static int i, i2;
	switch (uMsg) {
		case WM_INITDIALOG:
			hSNESEmulationDlg = hWnd;
			for(i = 0; i < 4; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDSKIPMETHOD), CB_INSERTSTRING, (WPARAM)i, (LPARAM)SoundSkipMethodStr[i]);
			for(i = 0; i < 4; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_JOYREADRATE), CB_INSERTSTRING, (WPARAM)i, (LPARAM)(GUI.Language ? JoyReadRateStr_JP[i] : JoyReadRateStr_US[i]));
			//timing
			for(i = 0; i < 11; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_FRAMESKIP), CB_INSERTSTRING, (WPARAM)i, (LPARAM)(GUI.Language ? SkipFramesStr_JP[i] : SkipFramesStr_US[i]));
			for(i = 0; i < 11; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_AUTOMAXSKIPFRAMES), CB_INSERTSTRING, (WPARAM)i, (LPARAM)(AutoMaxSkipFramesStr[i]));

			SNESEmulationDlgCtrlSet(hWnd);
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDC_CHECK_MODE7INTERPOLATE:
				case IDC_CHECK_GFX16:
				case IDC_CHECK_TRANSPARENCY:
				case IDC_CHECK_BGLAYERING:
				case IDC_CHECK_GFXW:
				case IDC_CHECK_HDMA:
				case IDC_CHECK_VSYNC:
				case IDC_CHECK_GFXENGINE2:
				case IDC_CHECK_INTERPOLATEDSOUND:

				case IDC_CHECK_APUENABLED:
				case IDC_CHECK_SOUNDECHO:
				case IDC_CHECK_REVERSESTEREO:
				case IDC_CHECK_SOUNDSYNC:
				case IDC_CHECK_MASTERVOLUME:
				case IDC_CHECK_SOUNDENVELOPEHEIGHTREADING:
				
				case IDC_CHECK_SHUTDOWN:
				case IDC_CHECK_SNESMOUSE:
				case IDC_CHECK_SNESSUPERSCOPE:
				case IDC_CHECK_SNESJUSTIFIER:
				case IDC_CHECK_SNESMULTIPLAYER5:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
						if(LOWORD(wParam) == IDC_CHECK_MODE7INTERPOLATE)
							Settings_Mode7Interpolate_cpy = tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_GFX16)
						{
								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_TRANSPARENCY), TRUE);
						}
						else if(LOWORD(wParam) == IDC_CHECK_GFXENGINE2)
						{
							Options_GFXengine2_cpy = tmpVal;
						}
						else if(LOWORD(wParam) == IDC_CHECK_TRANSPARENCY)
						{
							TransparencyFlag_cpy = tmpVal;
						}
						else if(LOWORD(wParam) == IDC_CHECK_BGLAYERING)
						{
							Settings_BGLayering_cpy = tmpVal;
						}
						else if(LOWORD(wParam) == IDC_CHECK_GFXW)
						{
							Settings_DisableGraphicWindows_cpy = !tmpVal;
						}
						else if(LOWORD(wParam) == IDC_CHECK_HDMA)
						{
							Settings_DisableHDMA_cpy = !tmpVal;
						}
						else if(LOWORD(wParam) == IDC_CHECK_VSYNC)
						{
							DirectX_wait_for_vsync_cpy = (tmpVal ? TRUE : FALSE);
						}
						else if(LOWORD(wParam) == IDC_CHECK_INTERPOLATEDSOUND)
							Settings_InterpolatedSound_cpy = tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_APUENABLED) {
							Settings_NextAPUEnabled_cpy = tmpVal;
							EnableWindow(GetDlgItem(hWnd, IDC_COMBO_SOUNDSKIPMETHOD), !Settings_NextAPUEnabled_cpy ? TRUE : FALSE);
						}
						else if(LOWORD(wParam) == IDC_CHECK_SOUNDECHO)
							Settings_DisableSoundEcho_cpy = !tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_REVERSESTEREO)
							Settings_ReverseStereo_cpy = tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_SOUNDSYNC)
							Settings_SoundSync_cpy = tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_MASTERVOLUME)
							Settings_DisableMasterVolume_cpy = !tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_SOUNDENVELOPEHEIGHTREADING)
							Settings_SoundEnvelopeHeightReading_cpy = tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_SHUTDOWN)
							Settings_ShutdownMaster_cpy = tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_SNESMOUSE) {
							Settings_Mouse_cpy = tmpVal;
							SetSnesDefaultPlugedInputDeviceCombo(hWnd);
						}
						else if(LOWORD(wParam) == IDC_CHECK_SNESSUPERSCOPE) {
							Settings_SuperScope_cpy = tmpVal;
							SetSnesDefaultPlugedInputDeviceCombo(hWnd);
						}
						else if(LOWORD(wParam) == IDC_CHECK_SNESJUSTIFIER) {
							Settings_Justifier_cpy = tmpVal;
							SetSnesDefaultPlugedInputDeviceCombo(hWnd);
						}
						else if(LOWORD(wParam) == IDC_CHECK_SNESMULTIPLAYER5) {
							Settings_MultiPlayer5_cpy = tmpVal;
							SetSnesDefaultPlugedInputDeviceCombo(hWnd);
						}
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				case IDC_COMBO_SOUNDSKIPMETHOD:
				case IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE:
				case IDC_COMBO_JOYREADRATE:
				case IDC_COMBO_FRAMESKIP:
				case IDC_COMBO_AUTOMAXSKIPFRAMES:
					if(HIWORD(wParam) == CBN_SELENDOK) {
						i2 = (int)SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0);
						if(LOWORD(wParam) == IDC_COMBO_SOUNDSKIPMETHOD)
							Settings_SoundSkipMethod_cpy = (uint8)i2;
						else if(LOWORD(wParam) == IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE)
							Settings_ControllerOption_cpy = (uint32)GetControllerOptionByComboindex((LRESULT)i2);
						else if(LOWORD(wParam) == IDC_COMBO_JOYREADRATE) {
							if(i2 == 0)
								Options_JoyReadRateMask_cpy = 0;
							else if(i2 == 1)
								Options_JoyReadRateMask_cpy = 1;
							else if(i2 == 2)
								Options_JoyReadRateMask_cpy = 3;
							else if(i2 == 3)
								Options_JoyReadRateMask_cpy = 7;
							else
								Options_JoyReadRateMask_cpy = 3;
						}
						else if(LOWORD(wParam) == IDC_COMBO_FRAMESKIP) {
							if(i2 == 0)
								Settings_SkipFrames_cpy = AUTO_FRAMERATE;
							else if(i2 >= 1 && i2 <= 10)
								Settings_SkipFrames_cpy = (uint32)i2;
							else
								Settings_SkipFrames_cpy = AUTO_FRAMERATE;
							EnableWindow(GetDlgItem(hWnd, IDC_COMBO_AUTOMAXSKIPFRAMES), (Settings_SkipFrames_cpy == AUTO_FRAMERATE) ? TRUE : FALSE);
						}
						else if(LOWORD(wParam) == IDC_COMBO_AUTOMAXSKIPFRAMES) {
							Settings_AutoMaxSkipFrames_cpy = (uint32)i2;
						}
						if(Settings_SoundSkipMethod_cpy != Settings.SoundSkipMethod || Settings_ControllerOption_cpy != Settings.ControllerOption
							|| Options_JoyReadRateMask_cpy != Options.JoyReadRateMask || Settings_SkipFrames_cpy != Settings.SkipFrames
							|| Settings_AutoMaxSkipFrames_cpy != Settings.AutoMaxSkipFrames) {
							ConfigDlgSettingsChanged(hWnd);
						}
					}
					break;

				case IDC_EDIT_CYCLEPERCENTAGE:
				case IDC_EDIT_NTSC:
				case IDC_EDIT_PAL:
				case IDC_EDIT_AUTOSAVESRAM: // Snes9xpp Original
					if(HIWORD(wParam) == EN_CHANGE) {
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				default:
					break;
			}
			break;

		case WM_NOTIFY:
			switch(((NMHDR *)lParam)->code) {
				case PSN_APPLY:
					ConfigChangePageIndex |= CONFIG_PAGE_SNESSET;
					ConfigDlgSettingsSave();
					break;

				case PSN_HELP:
					ConfigDlgDefaultSettingsLoad(hWnd);
					break;

				case PSN_QUERYCANCEL:
					SneseDlgReleaseSubClass();
					break;

				case PSN_SETACTIVE:
					ConfigChangePageIndex |= CONFIG_PAGE_SNESSET;
					StartPageIndex = 1;
					SneseDlgSetSubClass(hWnd);
					break;

				case PSN_KILLACTIVE:
					SneseDlgReleaseSubClass();
					break;

				default:
					break;
			}
			break;

		default:
			break;
	}
	return FALSE;
}

static void SNESEmulationDlgCtrlSet(HWND hWnd)
{
	//sound
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_INTERPOLATEDSOUND), BM_SETCHECK, (WPARAM)(Settings_InterpolatedSound_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_APUENABLED), BM_SETCHECK, (WPARAM)(Settings_NextAPUEnabled_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_SOUNDECHO), BM_SETCHECK, (WPARAM)(Settings_DisableSoundEcho_cpy ? 0 : 1), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_REVERSESTEREO), BM_SETCHECK, (WPARAM)(Settings_ReverseStereo_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_SOUNDSYNC), BM_SETCHECK, (WPARAM)(Settings_SoundSync_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_MASTERVOLUME), BM_SETCHECK, (WPARAM)(Settings_DisableMasterVolume_cpy ? 0 : 1), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_SOUNDENVELOPEHEIGHTREADING), BM_SETCHECK, (WPARAM)(Settings_SoundEnvelopeHeightReading_cpy ? 1 : 0), 0);

	SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDSKIPMETHOD), CB_SETCURSEL, (WPARAM)Settings_SoundSkipMethod_cpy, 0);
	EnableWindow(GetDlgItem(hWnd, IDC_COMBO_SOUNDSKIPMETHOD), !Settings_NextAPUEnabled_cpy ? TRUE : FALSE);

	//cpu
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_SHUTDOWN), BM_SETCHECK, (WPARAM)(Settings_ShutdownMaster_cpy ? 1 : 0), 0);
    EnableWindow(GetDlgItem(hWnd, IDC_CHECK_SHUTDOWN), (GUI.StopEmulation == TRUE) ? TRUE : FALSE);

	//graphics
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_MODE7INTERPOLATE), BM_SETCHECK, (WPARAM)(Settings_Mode7Interpolate_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_GFXENGINE2), BM_SETCHECK, (WPARAM)(Options_GFXengine2_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TRANSPARENCY), BM_SETCHECK, (WPARAM)(TransparencyFlag_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_BGLAYERING), BM_SETCHECK, (WPARAM)(Settings_BGLayering_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_GFXW), BM_SETCHECK, (WPARAM)(Settings_DisableGraphicWindows_cpy ? 0 : 1), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_HDMA), BM_SETCHECK, (WPARAM)(Settings_DisableHDMA_cpy ? 0 : 1), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_VSYNC), BM_SETCHECK, (WPARAM)(DirectX_wait_for_vsync_cpy ? 1 : 0), 0);

	//controller
	SetSnesDefaultPlugedInputDeviceCombo(hWnd);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_SNESMOUSE), BM_SETCHECK, (WPARAM)(Settings_Mouse_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_SNESSUPERSCOPE), BM_SETCHECK, (WPARAM)(Settings_SuperScope_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_SNESJUSTIFIER), BM_SETCHECK, (WPARAM)(Settings_Justifier_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_SNESMULTIPLAYER5), BM_SETCHECK, (WPARAM)(Settings_MultiPlayer5_cpy ? 1 : 0), 0);

	if(Options_JoyReadRateMask_cpy == 0)
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_JOYREADRATE), CB_SETCURSEL, 0, 0);
	else if(Options_JoyReadRateMask_cpy == 1)
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_JOYREADRATE), CB_SETCURSEL, 1, 0);
	else if(Options_JoyReadRateMask_cpy == 3)
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_JOYREADRATE), CB_SETCURSEL, 2, 0);
	else if(Options_JoyReadRateMask_cpy == 7)
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_JOYREADRATE), CB_SETCURSEL, 3, 0);
	else
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_JOYREADRATE), CB_SETCURSEL, 2, 0);

	if(Settings_SkipFrames_cpy == AUTO_FRAMERATE)
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_FRAMESKIP), CB_SETCURSEL, 0, 0);
	else if(Settings_SkipFrames_cpy >= 1 && Settings_SkipFrames_cpy <= 10)
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_FRAMESKIP), CB_SETCURSEL, (WPARAM)(Settings_SkipFrames_cpy), 0);
	else {
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_FRAMESKIP), CB_SETCURSEL, 0, 0);
		Settings_SkipFrames_cpy = Settings.SkipFrames = AUTO_FRAMERATE;
	}

	SendMessage(GetDlgItem(hWnd, IDC_COMBO_AUTOMAXSKIPFRAMES), CB_SETCURSEL, (WPARAM)Settings_AutoMaxSkipFrames_cpy, 0);
	EnableWindow(GetDlgItem(hWnd, IDC_COMBO_AUTOMAXSKIPFRAMES), (Settings_SkipFrames_cpy == AUTO_FRAMERATE) ? TRUE : FALSE);

	//CyclesPercentage
	SendMessage(GetDlgItem(hWnd, IDC_EDIT_CYCLEPERCENTAGE), EM_SETLIMITTEXT, 3, 0);
	SendMessage(GetDlgItem(hWnd, IDC_EDIT_PAL), EM_SETLIMITTEXT, 3, 0);
	SendMessage(GetDlgItem(hWnd, IDC_EDIT_NTSC), EM_SETLIMITTEXT, 3, 0);
	if(Settings_CyclesPercentage_cpy >= 1 && Settings_CyclesPercentage_cpy <= 199) {
		wsprintf(tmpcharbuf, "%d", Settings_CyclesPercentage_cpy);
		SetWindowText(GetDlgItem(hWnd, IDC_EDIT_CYCLEPERCENTAGE), tmpcharbuf);
	}
	else {
		Settings_CyclesPercentage_cpy = Settings.CyclesPercentage = 100;
		wsprintf(tmpcharbuf, "%d", Settings_CyclesPercentage_cpy);
		SetWindowText(GetDlgItem(hWnd, IDC_EDIT_CYCLEPERCENTAGE), tmpcharbuf);
	}
	wsprintf(tmpcharbuf, "%d", Settings_FrameTimePAL_cpy);
	SetWindowText(GetDlgItem(hWnd, IDC_EDIT_PAL), tmpcharbuf);
	wsprintf(tmpcharbuf, "%d", Settings_FrameTimeNTSC_cpy);
	SetWindowText(GetDlgItem(hWnd, IDC_EDIT_NTSC), tmpcharbuf);

// Snes9xpp Original
	SendMessage(GetDlgItem(hWnd, IDC_EDIT_AUTOSAVESRAM), EM_SETLIMITTEXT, 6, 0);
	wsprintf(tmpcharbuf, "%d", Settings_AutoSaveDelay_cpy);
	SetWindowText(GetDlgItem(hWnd, IDC_EDIT_AUTOSAVESRAM), tmpcharbuf);


	//not suported yet
//	EnableWindow(GetDlgItem(hWnd, IDC_CHECK_GFXENGINE2), FALSE);
}

static void SetSnesDefaultPlugedInputDeviceCombo(HWND hWnd)
{
	static char *SnesDeviceStr_US[] = {"MultiPlayer5", "Joypad", "Mouse #1", "Mouse #0", "Super scope","Justifier"};
	static char *SnesDeviceStr_JP[] = {"}`vC[5", "WCpbh", "}EX #1", "}EX #0", "X[p[XR[v","Justifier"};
	int joycomboi;
	int i3 = 0;
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_RESETCONTENT, 0, 0);
	//MultiPlayer5
	if(Settings_MultiPlayer5_cpy) {
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_INSERTSTRING, (WPARAM)i3, (LPARAM)(GUI.Language ? SnesDeviceStr_JP[0] : SnesDeviceStr_US[0]));
		++i3;
	}
	//Joypad
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_INSERTSTRING, (WPARAM)i3, (LPARAM)(GUI.Language ? SnesDeviceStr_JP[1] : SnesDeviceStr_US[1]));
	++i3;
	if(Settings_Mouse_cpy) {
		//Mouse Swapped
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_INSERTSTRING, (WPARAM)i3, (LPARAM)(GUI.Language ? SnesDeviceStr_JP[2] : SnesDeviceStr_US[2]));
		++i3;
		//Mouse
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_INSERTSTRING, (WPARAM)i3, (LPARAM)(GUI.Language ? SnesDeviceStr_JP[3] : SnesDeviceStr_US[3]));
		++i3;
	}
	if(Settings_SuperScope_cpy) {
		//Super scope
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_INSERTSTRING, (WPARAM)i3, (LPARAM)(GUI.Language ? SnesDeviceStr_JP[4] : SnesDeviceStr_US[4]));
		++i3;
	}
	if(Settings_Justifier_cpy) {
		//Justifier
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_INSERTSTRING, (WPARAM)i3, (LPARAM)(GUI.Language ? SnesDeviceStr_JP[5] : SnesDeviceStr_US[5]));
	}
	i3 = 0;
	if(Settings_ControllerOption_cpy == SNES_MULTIPLAYER5 && Settings_MultiPlayer5_cpy) {
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_SETCURSEL, (WPARAM)i3, 0);
		return;
	}
	if(Settings_MultiPlayer5_cpy)
		++i3;
	if(Settings_ControllerOption_cpy == SNES_JOYPAD) {
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_SETCURSEL, (WPARAM)i3, 0);
		return;
	}
	joycomboi = i3;
	++i3;
	if(Settings_ControllerOption_cpy == SNES_MOUSE_SWAPPED && Settings_Mouse_cpy) {
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_SETCURSEL, (WPARAM)i3, 0);
		return;
	}
	if(Settings_Mouse_cpy)
		++i3;
	if(Settings_ControllerOption_cpy == SNES_MOUSE && Settings_Mouse_cpy) {
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_SETCURSEL, (WPARAM)i3, 0);
		return;
	}
	if(Settings_Mouse_cpy)
		++i3;
	if(Settings_ControllerOption_cpy == SNES_SUPERSCOPE && Settings_SuperScope_cpy) {
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_SETCURSEL, (WPARAM)i3, 0);
		return;
	}
	if(Settings_SuperScope_cpy)
		++i3;
	if(Settings_ControllerOption_cpy == SNES_JUSTIFIER && Settings_Justifier_cpy) {
		SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_SETCURSEL, (WPARAM)i3, 0);
		return;
	}
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_SNESDEFAULTPLUGEDINPUTDEVICE), CB_SETCURSEL, (WPARAM)joycomboi, 0);
	Settings_ControllerOption_cpy = Settings.ControllerOption = SNES_JOYPAD;
	if(!GUI.StopEmulation)
		IPPU.Controller = Settings_ControllerOption_cpy;
}

static uint32 GetControllerOptionByComboindex(LRESULT conboindex)
{
	if(Settings_MultiPlayer5_cpy && conboindex == 0)
		return SNES_MULTIPLAYER5;

	if(Settings_MultiPlayer5_cpy && conboindex == 1)
		return SNES_JOYPAD;
	if(!Settings_MultiPlayer5_cpy && conboindex == 0)
		return SNES_JOYPAD;
	
	if(Settings_MultiPlayer5_cpy && Settings_Mouse_cpy && conboindex == 2)
		return SNES_MOUSE_SWAPPED;
	if(!Settings_MultiPlayer5_cpy && Settings_Mouse_cpy && conboindex == 1)
		return SNES_MOUSE_SWAPPED;

	if(Settings_MultiPlayer5_cpy && Settings_Mouse_cpy && conboindex == 3)
		return SNES_MOUSE;
	if(!Settings_MultiPlayer5_cpy && Settings_Mouse_cpy && conboindex == 2)
		return SNES_MOUSE;

	if(Settings_MultiPlayer5_cpy && Settings_Mouse_cpy && Settings_SuperScope_cpy && conboindex == 4)
		return SNES_SUPERSCOPE;
	if(!Settings_MultiPlayer5_cpy && Settings_Mouse_cpy && Settings_SuperScope_cpy && conboindex == 3)
		return SNES_SUPERSCOPE;
	if(Settings_MultiPlayer5_cpy && !Settings_Mouse_cpy && Settings_SuperScope_cpy && conboindex == 2)
		return SNES_SUPERSCOPE;
	if(!Settings_MultiPlayer5_cpy && !Settings_Mouse_cpy && Settings_SuperScope_cpy && conboindex == 1)
		return SNES_SUPERSCOPE;

	if(Settings_MultiPlayer5_cpy && Settings_Mouse_cpy && Settings_SuperScope_cpy && Settings_Justifier_cpy && conboindex == 5)
		return SNES_JUSTIFIER;
	if(!Settings_MultiPlayer5_cpy && Settings_Mouse_cpy && Settings_SuperScope_cpy && Settings_Justifier_cpy && conboindex == 4)
		return SNES_JUSTIFIER;
	if(Settings_MultiPlayer5_cpy && Settings_Mouse_cpy && !Settings_SuperScope_cpy && Settings_Justifier_cpy && conboindex == 4)
		return SNES_JUSTIFIER;
	if(Settings_MultiPlayer5_cpy && !Settings_Mouse_cpy && Settings_SuperScope_cpy && Settings_Justifier_cpy && conboindex == 3)
		return SNES_JUSTIFIER;
	if(!Settings_MultiPlayer5_cpy && Settings_Mouse_cpy && !Settings_SuperScope_cpy && Settings_Justifier_cpy && conboindex == 3)
		return SNES_JUSTIFIER;
	if(!Settings_MultiPlayer5_cpy && !Settings_Mouse_cpy && Settings_SuperScope_cpy && Settings_Justifier_cpy && conboindex == 2)
		return SNES_JUSTIFIER;
	if(Settings_MultiPlayer5_cpy && !Settings_Mouse_cpy && !Settings_SuperScope_cpy && Settings_Justifier_cpy && conboindex == 2)
		return SNES_JUSTIFIER;
	if(!Settings_MultiPlayer5_cpy && !Settings_Mouse_cpy && !Settings_SuperScope_cpy && Settings_Justifier_cpy && conboindex == 1)
		return SNES_JUSTIFIER;

	return SNES_JOYPAD;
}

static WNDPROC Edit_CyclesPercentage;
static WNDPROC Edit_FrameTimePAL;
static WNDPROC Edit_FrameTimeNTSC;
static HWND hEdit_CyclesPercentage;
static HWND hEdit_FrameTimePAL;
static HWND hEdit_FrameTimeNTSC;

// Snes9xpp Original
static WNDPROC Edit_AutoSaveDelay;
static HWND hEdit_AutoSaveDelay;


static LRESULT CALLBACK SneseEditSubProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_CHAR:
			if((LOWORD(wParam) >= '0' && LOWORD(wParam) <= '9') || LOWORD(wParam) == VK_BACK)
				break;
			return 0;

		default:
			break;

	}
	if(hWnd == hEdit_CyclesPercentage)
		return (CallWindowProc(Edit_CyclesPercentage, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_FrameTimePAL)
		return (CallWindowProc(Edit_FrameTimePAL, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_FrameTimeNTSC)
		return (CallWindowProc(Edit_FrameTimeNTSC, hWnd, uMsg, wParam, lParam));
// Snes9xpp Original
	if(hWnd == hEdit_AutoSaveDelay)
		return (CallWindowProc(Edit_AutoSaveDelay, hWnd, uMsg, wParam, lParam));
	return 0;
}

static void SneseDlgSetSubClass(HWND hWnd)
{
	hEdit_CyclesPercentage = GetDlgItem(hWnd, IDC_EDIT_CYCLEPERCENTAGE);
	hEdit_FrameTimePAL = GetDlgItem(hWnd, IDC_EDIT_PAL);
	hEdit_FrameTimeNTSC = GetDlgItem(hWnd, IDC_EDIT_NTSC);
	Edit_CyclesPercentage = (WNDPROC)GetWindowLong(hEdit_CyclesPercentage, GWL_WNDPROC);
	Edit_FrameTimePAL = (WNDPROC)GetWindowLong(hEdit_FrameTimePAL, GWL_WNDPROC);
	Edit_FrameTimeNTSC = (WNDPROC)GetWindowLong(hEdit_FrameTimeNTSC, GWL_WNDPROC);
	SetWindowLong(hEdit_CyclesPercentage, GWL_WNDPROC, (LONG)SneseEditSubProc);
	SetWindowLong(hEdit_FrameTimePAL, GWL_WNDPROC, (LONG)SneseEditSubProc);
	SetWindowLong(hEdit_FrameTimeNTSC, GWL_WNDPROC, (LONG)SneseEditSubProc);

 // Snes9xpp Original
	hEdit_AutoSaveDelay = GetDlgItem(hWnd, IDC_EDIT_AUTOSAVESRAM);
	Edit_AutoSaveDelay = (WNDPROC)GetWindowLong(hEdit_AutoSaveDelay, GWL_WNDPROC);
	SetWindowLong(hEdit_AutoSaveDelay, GWL_WNDPROC, (LONG)SneseEditSubProc);
}

static void SneseDlgReleaseSubClass()
{
	SetWindowLong(hEdit_CyclesPercentage, GWL_WNDPROC, (LONG)Edit_CyclesPercentage);
	SetWindowLong(hEdit_FrameTimePAL, GWL_WNDPROC, (LONG)Edit_FrameTimePAL);
	SetWindowLong(hEdit_FrameTimeNTSC, GWL_WNDPROC, (LONG)Edit_FrameTimeNTSC);

	SetWindowLong(hEdit_AutoSaveDelay, GWL_WNDPROC, (LONG)Edit_AutoSaveDelay); // Snes9xpp Original
}

/***********************************************************************************************
 Joy Dialog
************************************************************************************************/ 

void JoySettingsLoad()
{
	int i;
	for (i = 0; i < 5; i++ ) {
		//CfbNXANZXpɃ|C^
		JoypadButtonRef[i].Button[0] = &Joypadcpy[i].Left;
		JoypadButtonRef[i].Button[1] = &Joypadcpy[i].Right;
		JoypadButtonRef[i].Button[2] = &Joypadcpy[i].Up;
		JoypadButtonRef[i].Button[3] = &Joypadcpy[i].Down;
		JoypadButtonRef[i].Button[4] = &Joypadcpy[i].Start;
		JoypadButtonRef[i].Button[5] = &Joypadcpy[i].Select;
		JoypadButtonRef[i].Button[6] = &Joypadcpy[i].A;
		JoypadButtonRef[i].Button[7] = &Joypadcpy[i].B;
		JoypadButtonRef[i].Button[8] = &Joypadcpy[i].X;
		JoypadButtonRef[i].Button[9] = &Joypadcpy[i].Y;
		JoypadButtonRef[i].Button[10] = &Joypadcpy[i].L;
		JoypadButtonRef[i].Button[11] = &Joypadcpy[i].R;
		JoypadButtonRef[i].Turbo[0] = &Joypadcpy[i].Turbo_Left;
		JoypadButtonRef[i].Turbo[1] = &Joypadcpy[i].Turbo_Right;
		JoypadButtonRef[i].Turbo[2] = &Joypadcpy[i].Turbo_Up;
		JoypadButtonRef[i].Turbo[3] = &Joypadcpy[i].Turbo_Down;
		JoypadButtonRef[i].Turbo[4] = &Joypadcpy[i].Turbo_Start;
		JoypadButtonRef[i].Turbo[5] = &Joypadcpy[i].Turbo_Select;
		JoypadButtonRef[i].Turbo[6] = &Joypadcpy[i].Turbo_A;
		JoypadButtonRef[i].Turbo[7] = &Joypadcpy[i].Turbo_B;
		JoypadButtonRef[i].Turbo[8] = &Joypadcpy[i].Turbo_X;
		JoypadButtonRef[i].Turbo[9] = &Joypadcpy[i].Turbo_Y;
		JoypadButtonRef[i].Turbo[10] = &Joypadcpy[i].Turbo_L;
		JoypadButtonRef[i].Turbo[11] = &Joypadcpy[i].Turbo_R;
		Joypadcpy[i] = Joypad[i];
	}
	for (i = 0; i < 12; i++) {
		MapFunc_cpy[i] = MapFunc[i];
	}
	DirectX_DIuse_cpy = DirectX.DIuse;

	//MaxWinJoyButtons
	Options_MaxWinJoyButtons_cpy = Options.MaxWinJoyButtons;

}

void JoyDefaultSettingsLoad()
{
	DirectX_DIuse_cpy = DirectX_DIuse_def;
	SendMessage(GetDlgItem(hJoypadDlg, IDC_INPUTCOMBO), CB_SETCURSEL, (WPARAM)(DirectX_DIuse_cpy ? 1 : 0), 0);

	Options_MaxWinJoyButtons_cpy = Options_MaxWinJoyButtons_def;

	JoyDlgCtlSet(hJoypadDlg);
}

void JoySettingsSave()
{
	int i;
	static int32 tmpnumval;
	
	for (i = 0; i < 5; i++ )
		Joypad[i] = Joypadcpy[i];
	MakeTurboStatus();
	for (i = 0; i < 12; i++) {
		MapFunc[i] = MapFunc_cpy[i];
	}
	if(DirectX_DIuse_cpy != DirectX.DIuse) {
		DirectX.DIuse = DirectX_DIuse_cpy;
		S9xwJoystickInit();
	}

	GetWindowText(GetDlgItem(hJoypadDlg, IDC_EDIT_MAXWINJOYBUTTONS), tmpcharbuf, 255);
	tmpnumval = (int32)atoi(tmpcharbuf);
	if(tmpnumval >= 0 && tmpnumval <= 32) {
		Options.MaxWinJoyButtons = Options_MaxWinJoyButtons_cpy = tmpnumval;
	}
	else {
		Options_MaxWinJoyButtons_cpy = Options.MaxWinJoyButtons;
	}
	wsprintf(tmpcharbuf, "%d", Options_MaxWinJoyButtons_cpy);
	SetWindowText(GetDlgItem(hJoypadDlg, IDC_EDIT_MAXWINJOYBUTTONS), tmpcharbuf);

}

//Joypad Configurasion
LRESULT CALLBACK JoypadDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam)
{
	static char *combostr[] = {"Joypad #1", "Joypad #2", "Joypad #3", "Joypad #4", "Joypad #5"};
	static char *combostrDIorWinAPI[] = {"WindowsAPI", "DirectInput"};
	static char *combostrMapFunc_JP[] = {"Xbg 1 Z[u",
										"Xbg 2 Z[u",
										"Xbg 1 [h",
										"Xbg 2 [h",
										"^[{[h",
										"^[{ || >>",
										"Zbg",
										"\x83\x5ctgZbg",
										"SPCzo",
										"XN[Vbg",
										"Xpp Key 1", // Snes9xpp Original 
										"Xpp Key 2",
										"Xpp Key 3",
										"Xpp Key 4",
										"Xpp Key 5",
										"Xpp Key 6",
										"Xpp Key 7",
										"Xpp Key 8",
										"Xpp Key 9",
										"Xpp Key 10",
										"Xpp Key 11",
										"Xpp Key 12",
										"Xpp Key A",
										"Xpp Key B",
										"Xpp Key C",
										"Xpp Key D",
										"Xpp Key E",
										"Xpp Key F"};
										//"XN[Vbg"};
	static char *combostrMapFunc_US[] = {"Slot 1 save",
										"Slot 2 save",
										"Slot 1 load",
										"Slot 2 load",
										"Turbo Mode",
										"Turbo || >>",
										"Reset",
										"SoftReset",
										"SPC dump",
										"Screenshot",
										"Xpp Key 1", // Snes9xpp Original 
										"Xpp Key 2",
										"Xpp Key 3",
										"Xpp Key 4",
										"Xpp Key 5",
										"Xpp Key 6",
										"Xpp Key 7",
										"Xpp Key 8",
										"Xpp Key 9",
										"Xpp Key 10",
										"Xpp Key 11",
										"Xpp Key 12",
										"Xpp Key A",
										"Xpp Key B",
										"Xpp Key C",
										"Xpp Key D",
										"Xpp Key E",
										"Xpp Key F"};
										//"Screenshot"};
	static WORD jb;
	static int i2, i3, i4;
	volatile static unsigned char JoyStatusReadingFlag;
	switch (uMsg) {
		case WM_INITDIALOG:
			hJoypadDlg = hWnd;
			//GfBgEChEnhۑiŎg)
			hEdit_Left = GetDlgItem(hWnd, IDC_EDIT_LEFT);
			hEdit_Right = GetDlgItem(hWnd, IDC_EDIT_RIGHT);
			hEdit_Up = GetDlgItem(hWnd, IDC_EDIT_UP);
			hEdit_Down = GetDlgItem(hWnd, IDC_EDIT_DOWN);
			hEdit_Start = GetDlgItem(hWnd, IDC_EDIT_START);
			hEdit_Select = GetDlgItem(hWnd, IDC_EDIT_SELECT);
			hEdit_A = GetDlgItem(hWnd, IDC_EDIT_A);
			hEdit_B = GetDlgItem(hWnd, IDC_EDIT_B);
			hEdit_X = GetDlgItem(hWnd, IDC_EDIT_X);
			hEdit_Y = GetDlgItem(hWnd, IDC_EDIT_Y);
			hEdit_L = GetDlgItem(hWnd, IDC_EDIT_L);
			hEdit_R = GetDlgItem(hWnd, IDC_EDIT_R);
			hEdit_MapFunc0 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC0);
			hEdit_MapFunc1 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC1);
			hEdit_MapFunc2 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC2);
			hEdit_MapFunc3 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC3);
			hEdit_MapFunc4 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC4);
			hEdit_MapFunc5 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC5);
			hEdit_MapFunc6 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC6);
			hEdit_MapFunc7 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC7);
			hEdit_MapFunc8 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC8);
			hEdit_MapFunc9 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC9);
			hEdit_MapFunc10 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC10);
			hEdit_MapFunc11 = GetDlgItem(hWnd, IDC_EDIT_MAPFUNC11);

			for (i2 = 0; i2 < 5; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_CTRLCOMBO), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)combostr[i2]);
			}
			SendMessage(GetDlgItem(hWnd, IDC_CTRLCOMBO), CB_SETCURSEL, (WPARAM)ComboNo, 0);
			//_CAO{bNXRg[ɕ\
			JoyDlgCtlSet(hWnd);

			for (i2 = 0; i2 < 2; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_INPUTCOMBO), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)combostrDIorWinAPI[i2]);
			}
			SendMessage(GetDlgItem(hWnd, IDC_INPUTCOMBO), CB_SETCURSEL, (WPARAM)(DirectX_DIuse_cpy ? 1 : 0), 0);
			

// Snes9xpp Original
			//WCXeBbN@\蓖ăR{{bNXɃeLXgZbg
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO0), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO0), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[0])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO1), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO1), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[1])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO2), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO2), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[2])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO3), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO3), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[3])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO4), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO4), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[4])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO5), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO5), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[5])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO6), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO6), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[6])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO7), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO7), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[7])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO8), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO8), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[8])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO9), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO9), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[9])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO10), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO10), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[10])), 0);
			for (i2 = 0; i2 < 28; i2++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO11), CB_INSERTSTRING, (WPARAM)i2, (LPARAM)(GUI.Language ? combostrMapFunc_JP[i2] : combostrMapFunc_US[i2]));
			}
			SendMessage(GetDlgItem(hWnd, IDC_MAPFUNCCOMBO11), CB_SETCURSEL, (WPARAM)(HIWORD(MapFunc_cpy[11])), 0);
			break;
// Snes9xpp Original 
		case WM_CTLCOLORSTATIC:
			if((HWND)lParam == hEdit_Left || (HWND)lParam == hEdit_Right
				|| (HWND)lParam == hEdit_Up || (HWND)lParam == hEdit_Down
				|| (HWND)lParam == hEdit_Start || (HWND)lParam == hEdit_Select
				|| (HWND)lParam == hEdit_A || (HWND)lParam == hEdit_B
				|| (HWND)lParam == hEdit_X || (HWND)lParam == hEdit_Y
				|| (HWND)lParam == hEdit_L || (HWND)lParam == hEdit_R) {
					if(GetFocus() == (HWND)lParam) {
						SetBkMode((HDC)wParam, TRANSPARENT);
						if(SelectedEditBkGreen)
							return (long)SelectedEditBkGreen;
					}
					return (long)GetStockObject(WHITE_BRUSH);
			}
			else if((HWND)lParam == hEdit_MapFunc0 || (HWND)lParam == hEdit_MapFunc1
				|| (HWND)lParam == hEdit_MapFunc2 || (HWND)lParam == hEdit_MapFunc3
				|| (HWND)lParam == hEdit_MapFunc4 || (HWND)lParam == hEdit_MapFunc5
				|| (HWND)lParam == hEdit_MapFunc6 || (HWND)lParam == hEdit_MapFunc7
				|| (HWND)lParam == hEdit_MapFunc8 || (HWND)lParam == hEdit_MapFunc9
				|| (HWND)lParam == hEdit_MapFunc10 || (HWND)lParam == hEdit_MapFunc11) {
					if(GetFocus() == (HWND)lParam) {
						SetBkMode((HDC)wParam, TRANSPARENT);
						if(SelectedEditBkOrange)
							return (long)SelectedEditBkOrange;
					}
					return (long)GetStockObject(WHITE_BRUSH);
			}
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDC_EDIT_UP:
				case IDC_EDIT_DOWN:
				case IDC_EDIT_LEFT:
				case IDC_EDIT_RIGHT:
				case IDC_EDIT_A:
				case IDC_EDIT_B:
				case IDC_EDIT_X:
				case IDC_EDIT_Y:
				case IDC_EDIT_L:
				case IDC_EDIT_R:
				case IDC_EDIT_SELECT:
				case IDC_EDIT_START:
				case IDC_EDIT_MAPFUNC0:
				case IDC_EDIT_MAPFUNC1:
				case IDC_EDIT_MAPFUNC2:
				case IDC_EDIT_MAPFUNC3:
				case IDC_EDIT_MAPFUNC4:
				case IDC_EDIT_MAPFUNC5:
				case IDC_EDIT_MAPFUNC6:
				case IDC_EDIT_MAPFUNC7:
				case IDC_EDIT_MAPFUNC8:
				case IDC_EDIT_MAPFUNC9:
				case IDC_EDIT_MAPFUNC10:
				case IDC_EDIT_MAPFUNC11:
					if(HIWORD(wParam) == EN_KILLFOCUS)
						InvalidateRgn((HWND)lParam, NULL, TRUE);
					break;

				case IDC_CTRLCOMBO:
					if(HIWORD(wParam) == CBN_SELENDOK) {
						ComboNo = (int)SendMessage(GetDlgItem(hWnd, IDC_CTRLCOMBO), CB_GETCURSEL, 0, 0);
						JoyDlgCtlSet(hWnd);
					}
					break;

				case IDC_JOYENABLE_CHECK:
					if(HIWORD(wParam) == BN_CLICKED) {
						Joypadcpy[ComboNo].Enabled = IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				case IDC_INPUTCOMBO:
					if(HIWORD(wParam) == CBN_SELENDOK) {
						if(SendMessage(GetDlgItem(hWnd, IDC_INPUTCOMBO), CB_GETCURSEL, 0, 0) == 1 &&
							!DirectX_DIuse_cpy) {
							DirectX_DIuse_cpy = TRUE;
							ConfigDlgSettingsChanged(hWnd);
						}
						else if(DirectX_DIuse_cpy) {
							DirectX_DIuse_cpy = FALSE;
							ConfigDlgSettingsChanged(hWnd);
						}
					}
					break;

				case IDC_MAPFUNCCOMBO0:
				case IDC_MAPFUNCCOMBO1:
				case IDC_MAPFUNCCOMBO2:
				case IDC_MAPFUNCCOMBO3:
				case IDC_MAPFUNCCOMBO4:
				case IDC_MAPFUNCCOMBO5:
				case IDC_MAPFUNCCOMBO6:
				case IDC_MAPFUNCCOMBO7:
				case IDC_MAPFUNCCOMBO8:
				case IDC_MAPFUNCCOMBO9:
				case IDC_MAPFUNCCOMBO10:
				case IDC_MAPFUNCCOMBO11:
					if(HIWORD(wParam) == CBN_SELENDOK) {
						if(LOWORD(wParam) == IDC_MAPFUNCCOMBO0)
							MapFunc_cpy[0] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[0]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO1)
							MapFunc_cpy[1] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[1]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO2)
							MapFunc_cpy[2] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[2]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO3)
							MapFunc_cpy[3] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[3]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO4)
							MapFunc_cpy[4] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[4]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO5)
							MapFunc_cpy[5] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[5]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO6)
							MapFunc_cpy[6] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[6]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO7)
							MapFunc_cpy[7] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[7]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO8)
							MapFunc_cpy[8] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[8]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO9)
							MapFunc_cpy[9] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[9]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO10)
							MapFunc_cpy[10] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[10]);
						else if(LOWORD(wParam) == IDC_MAPFUNCCOMBO11)
							MapFunc_cpy[11] = (SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0) << 16) + LOWORD(MapFunc_cpy[11]);
						if(MapFunc_cpy[0] != MapFunc[0] || MapFunc_cpy[1] != MapFunc[1] || MapFunc_cpy[2] != MapFunc[2] || MapFunc_cpy[3] != MapFunc[3] || MapFunc_cpy[4] != MapFunc[4] || MapFunc_cpy[5] != MapFunc[5]
						|| MapFunc_cpy[6] != MapFunc[6] || MapFunc_cpy[7] != MapFunc[7] || MapFunc_cpy[8] != MapFunc[8] || MapFunc_cpy[9] != MapFunc[9] || MapFunc_cpy[10] != MapFunc[10] || MapFunc_cpy[11] != MapFunc[11]) {
							ConfigDlgSettingsChanged(hWnd);
						}
					}
					break;

				//Turbo`FbN
				case IDC_CHECK_TURBO_LEFT:
				case IDC_CHECK_TURBO_RIGHT:
				case IDC_CHECK_TURBO_UP:
				case IDC_CHECK_TURBO_DOWN:
				case IDC_CHECK_TURBO_START:
				case IDC_CHECK_TURBO_SELECT:
				case IDC_CHECK_TURBO_A:
				case IDC_CHECK_TURBO_B:
				case IDC_CHECK_TURBO_X:
				case IDC_CHECK_TURBO_Y:
				case IDC_CHECK_TURBO_L:
				case IDC_CHECK_TURBO_R:
					if(HIWORD(wParam) == BN_CLICKED) {
						if(IDC_CHECK_TURBO_LEFT == LOWORD(wParam))
							i3 = 0;
						else if (IDC_CHECK_TURBO_RIGHT == LOWORD(wParam))
							i3 = 1;
						else if (IDC_CHECK_TURBO_UP == LOWORD(wParam))
							i3 = 2;
						else if (IDC_CHECK_TURBO_DOWN == LOWORD(wParam))
							i3 = 3;
						else if (IDC_CHECK_TURBO_START == LOWORD(wParam))
							i3 = 4;
						else if (IDC_CHECK_TURBO_SELECT == LOWORD(wParam))
							i3 = 5;
						else if (IDC_CHECK_TURBO_A == LOWORD(wParam))
							i3 = 6;
						else if (IDC_CHECK_TURBO_B == LOWORD(wParam))
							i3 = 7;
						else if (IDC_CHECK_TURBO_X == LOWORD(wParam))
							i3 = 8;
						else if (IDC_CHECK_TURBO_Y == LOWORD(wParam))
							i3 = 9;
						else if (IDC_CHECK_TURBO_L == LOWORD(wParam))
							i3 = 10;
						else if (IDC_CHECK_TURBO_R == LOWORD(wParam))
							i3 = 11;
						*JoypadButtonRef[ComboNo].Turbo[i3] = IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				case IDC_EDIT_MAXWINJOYBUTTONS: // Snes9xpp Original
					if(HIWORD(wParam) == EN_CHANGE) {
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				default:
					break;
			}
			break;

		case WM_NOTIFY:
			switch(((NMHDR *)lParam)->code) {
				case PSN_APPLY:
					ConfigChangePageIndex |= CONFIG_PAGE_JOYSET;
					ConfigDlgSettingsSave();
					break;

				case PSN_HELP:
					ConfigDlgDefaultSettingsLoad(hWnd);
					break;

				case PSN_QUERYCANCEL:
					ReleaseSubClass(hWnd);
					break;

				case PSN_SETACTIVE:
					ConfigChangePageIndex |= CONFIG_PAGE_JOYSET;
					StartPageIndex = 2;
					JoyStatusReadingFlag = 0;
					//GfBgEChEIԂŐFςuV쐬
					if(!SelectedEditBkGreen)
						SelectedEditBkGreen = CreateSolidBrush(RGB(0, 255, 0));
					if(!SelectedEditBkOrange)
						SelectedEditBkOrange = CreateSolidBrush(RGB(0xff, 0xa5, 0x00));
					//WCXeBbN̓`FbNp^C}Zbg
					if(!TimerActiveFlag && SetTimer(hWnd, 100, 50, NULL))
						TimerActiveFlag = 1;
					//EChẼTuNX
					SetSubClass(hWnd);
					break;

				case PSN_KILLACTIVE:
					ReleaseSubClass(hWnd);
					break;

				default:
					break;
			}
			break;

		case WM_TIMER:
			if(JoyStatusReadingFlag)
				break;
			JoyStatusReadingFlag = 1;
			for(i4 = 0; i4 < 8; i4++) {
				jb = GetJoyButtonCode();
				if(jb)
					SetJoyButtonCode(GetFocus(), jb);
			}
			JoyStatusReadingFlag = 0;
			break;

		default:
			break;
	}
	return FALSE;
}

static WORD SetJoyButtonText(HWND hWnd, WORD &JoyButton, WORD &JoyButtonOld, bool DuplicateCheckFlag)
{
	static char JoyText[16];
	static char JoyText2[16];
	static char JoyText3[16];
	static char *Result;
	static int J;

	FillMemory((PVOID)JoyText, 16, 0);
	FillMemory((PVOID)JoyText2, 16, 0);
	FillMemory((PVOID)JoyText3, 16, 0);
	Result = NULL;
	J = 0;

	if(JoyButton == VK_DELETE) {
		JoyButton = 0;
		SetWindowText(hWnd, "None");
		return VK_DELETE;
	}

	// Check for windows keys
	if( JoyButton == VK_MENU || JoyButton == VK_CAPITAL || JoyButton == VK_LWIN ||
		JoyButton == VK_RWIN || JoyButton == VK_APPS)
	{
		return 0;
	}

	// Check for Snes9X keys
	if( (JoyButton >= VK_F1 && JoyButton <= VK_F9) || JoyButton == VK_F11 || JoyButton == VK_F12 || JoyButton == VK_ESCAPE || JoyButton == VK_CONTROL || JoyButton == VK_SHIFT || JoyButton == VK_TAB)
	{
		return 0;
	}

	if(DuplicateCheckFlag) {
		if( JoyButton == LOWORD(MapFunc_cpy[0])) {
			MapFunc_cpy[0] = (MapFunc_cpy[0] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc0, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc0, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[1])) {
			MapFunc_cpy[1] = (MapFunc_cpy[1] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc1, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc1, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[2])) {
			MapFunc_cpy[2] = (MapFunc_cpy[2] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc2, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc2, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[3])) {
			MapFunc_cpy[3] = (MapFunc_cpy[3] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc3, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc3, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[4])) {
			MapFunc_cpy[4] = (MapFunc_cpy[4] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc4, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc4, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[5])) {
			MapFunc_cpy[5] = (MapFunc_cpy[5] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc5, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc5, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[6])) {
			MapFunc_cpy[6] = (MapFunc_cpy[6] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc6, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc6, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[7])) {
			MapFunc_cpy[7] = (MapFunc_cpy[7] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc7, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc7, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[8])) {
			MapFunc_cpy[8] = (MapFunc_cpy[8] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc8, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc8, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[9])) {
			MapFunc_cpy[9] = (MapFunc_cpy[9] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc9, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc9, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[10])) {
			MapFunc_cpy[10] = (MapFunc_cpy[10] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc10, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc10, JoyText2);
			return JoyButton;
		}
		if( JoyButton == LOWORD(MapFunc_cpy[11])) {
			MapFunc_cpy[11] = (MapFunc_cpy[11] & 0xffff0000) | JoyButtonOld;
			GetWindowText(hWnd, JoyText2, 16);
			GetWindowText(hEdit_MapFunc11, JoyText3, 16);
			SetWindowText(hWnd, JoyText3);
			SetWindowText(hEdit_MapFunc11, JoyText2);
			return JoyButton;
		}
		for(J = 0; J < 5; J++)
		{
			if( JoyButton == Joypadcpy[J].Left) {
				Joypadcpy[J].Left = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_Left, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_Left, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].Right) {
				Joypadcpy[J].Right = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_Right, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_Right, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].Up) {
				Joypadcpy[J].Up = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_Up, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_Up, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].Down) {
				Joypadcpy[J].Down = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_Down, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_Down, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].Start) {
				Joypadcpy[J].Start = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_Start, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_Start, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].Select) {
				Joypadcpy[J].Select = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_Select, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_Select, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].A) {
				Joypadcpy[J].A = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_A, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_A, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].B) {
				Joypadcpy[J].B = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_B, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_B, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].X) {
				Joypadcpy[J].X = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_X, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_X, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].Y) {
				Joypadcpy[J].Y = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_Y, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_Y, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].L) {
				Joypadcpy[J].L = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_L, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_L, JoyText2);
					return JoyButton;
				}
				break;
			}
			if( JoyButton == Joypadcpy[J].R) {
				Joypadcpy[J].R = JoyButtonOld;
				if(J == ComboNo) {
					GetWindowText(hWnd, JoyText2, 16);
					GetWindowText(hEdit_R, JoyText3, 16);
					SetWindowText(hWnd, JoyText3);
					SetWindowText(hEdit_R, JoyText2);
					return JoyButton;
				}
				break;
			}
		}
	}

	if(JoyButton&0x8000)
	{
		JoyText[0] = 'J';
		JoyText[1] = 0x30 + ((JoyButton >> 8) & 0x0f);
		if(JoyText[1] > 0x39)
			JoyText[1] += 7;
		switch(JoyButton&0xff)
		{
			case 0: JoyText[2] = 'L'; break;
			case 1: JoyText[2] = 'R'; break;
			case 2: JoyText[2] = 'U'; break;
			case 3: JoyText[2] = 'D'; break;
			case 4: JoyText[2] = 'P'; JoyText[3] = 'L'; break;
			case 5: JoyText[2] = 'P'; JoyText[3] = 'R'; break;
			case 6: JoyText[2] = 'P'; JoyText[3] = 'U'; break;
			case 7: JoyText[2] = 'P'; JoyText[3] = 'D'; break;
			case 41: JoyText[2] = 'Z'; JoyText[3] = 'U'; break;
			case 42: JoyText[2] = 'Z'; JoyText[3] = 'D'; break;
			case 43: JoyText[2] = 'R'; JoyText[3] = 'U'; break;
			case 44: JoyText[2] = 'R'; JoyText[3] = 'D'; break;
			case 45: JoyText[2] = 'U'; JoyText[3] = 'U'; break;
			case 46: JoyText[2] = 'U'; JoyText[3] = 'D'; break;
			case 47: JoyText[2] = 'V'; JoyText[3] = 'U'; break;
			case 48: JoyText[2] = 'V'; JoyText[3] = 'D'; break;

			default:
				if( (JoyButton&0xff) > 40)
				{
					itoa((JoyButton&0xff), &JoyText[2], 10);
					break;
				}
				JoyText[2] = 'B';
				itoa(((JoyButton&0xff)-8), &JoyText[3], 10);
				break;
		}
	}

	// 0~9, A~Z
	else if( (JoyButton >= '0' && JoyButton <= '9') || (JoyButton >= 'A' && JoyButton <= 'Z'))
	{
		JoyText[0] = (char)JoyButton;
	}

	else if( JoyButton >= VK_NUMPAD0 && JoyButton <= VK_NUMPAD9)
	{
		JoyText[0] = 'N';
		JoyText[1] = '-';
		JoyText[2] = char('0'+(JoyButton - VK_NUMPAD0));
	}
	else {
		switch(JoyButton)
		{
			case 0: Result = "None"; break;
			case VK_TAB: Result = "Tab"; break;
			case VK_BACK: Result = "Bs"; break;
			case VK_CLEAR: Result = "Delete"; break;
			case VK_RETURN: Result = "Enter"; break;
			case VK_SHIFT:  Result = "Shift"; break;
			case VK_CONTROL: Result = "CTRL"; break;
			case VK_MENU: Result = "Alt"; break;
			case VK_PAUSE: Result = "Pause"; break;
			case VK_CAPITAL: Result = "C-Lock"; break;
			case VK_ESCAPE: Result = "ESC"; break;
			case VK_SPACE: Result = "Space"; break;
			case VK_PRIOR: Result = "PgUp"; break;
			case VK_NEXT: Result = "PgDn"; break;
			case VK_HOME: Result = "Home"; break;
			case VK_END: Result = "End"; break;
			case VK_LEFT: Result = "Left"; break;
			case VK_RIGHT: Result = "Right"; break;
			case VK_UP: Result = "Up"; break;
			case VK_DOWN: Result = "Down"; break;
			case VK_SELECT: Result = "Select"; break;
			case VK_PRINT: Result = "Print"; break;
			case VK_EXECUTE: Result = "Exec"; break;
			case VK_SNAPSHOT: Result = "Snap"; break;
			case VK_INSERT: Result = "Insert"; break;
			case VK_DELETE: Result = "Delete"; break;
			case VK_HELP: Result = "Help"; break;
			case VK_LWIN: Result = "LWin"; break;
			case VK_RWIN: Result = "RWin"; break;
			case VK_APPS: Result = "AppKey"; break;
			case VK_MULTIPLY: Result = "*"; break;
			case VK_ADD: Result = "+"; break;
			case VK_SEPARATOR: Result = "\\"; break;
			case VK_SUBTRACT: Result = "-"; break;
			case VK_DECIMAL: Result = "."; break;
			case VK_DIVIDE: Result = "/"; break;
			case VK_NUMLOCK: Result = "N-Lock"; break;
			case VK_SCROLL: Result = "S-Lock"; break;
			case VK_F10: Result = "F10"; break;
			case VK_F11: Result = "F11"; break;
			case VK_F12: Result = "F12"; break;
		}
		if(Result)
			lstrcpy(JoyText, Result);
		else
			itoa((int)JoyButton, JoyText, 10);
	}
	SetWindowText(hWnd, JoyText);
	return JoyButton;
}

bool SetJoyButtonCode(HWND hFocus, WORD KeyCode)
{
	if(hEdit_Up == hFocus) {
		if(SetJoyButtonText(hEdit_Up, KeyCode, Joypadcpy[ComboNo].Up, TRUE)) {
			Joypadcpy[ComboNo].Up = KeyCode;
			SetFocus(hEdit_Down);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_Down == hFocus) {
		if(SetJoyButtonText(hEdit_Down, KeyCode, Joypadcpy[ComboNo].Down, TRUE)) {
			Joypadcpy[ComboNo].Down = KeyCode;
			SetFocus(hEdit_Left);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_Left == hFocus) {
		if(SetJoyButtonText(hEdit_Left, KeyCode, Joypadcpy[ComboNo].Left, TRUE)) {
			Joypadcpy[ComboNo].Left = KeyCode;
			SetFocus(hEdit_Right);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_Right == hFocus) {
		if(SetJoyButtonText(hEdit_Right, KeyCode, Joypadcpy[ComboNo].Right, TRUE)) {
			Joypadcpy[ComboNo].Right = KeyCode;
			SetFocus(hEdit_A);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_A == hFocus) {
		if(SetJoyButtonText(hEdit_A, KeyCode, Joypadcpy[ComboNo].A, TRUE)) {
			Joypadcpy[ComboNo].A = KeyCode;
			SetFocus(hEdit_B);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_B == hFocus) {
		if(SetJoyButtonText(hEdit_B, KeyCode, Joypadcpy[ComboNo].B, TRUE)) {
			Joypadcpy[ComboNo].B = KeyCode;
			SetFocus(hEdit_X);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_X == hFocus) {
		if(SetJoyButtonText(hEdit_X, KeyCode, Joypadcpy[ComboNo].X, TRUE)) {
			Joypadcpy[ComboNo].X = KeyCode;
			SetFocus(hEdit_Y);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_Y == hFocus) {
		if(SetJoyButtonText(hEdit_Y, KeyCode, Joypadcpy[ComboNo].Y, TRUE)) {
			Joypadcpy[ComboNo].Y = KeyCode;
			SetFocus(hEdit_Select);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_Select == hFocus) {
		if(SetJoyButtonText(hEdit_Select, KeyCode, Joypadcpy[ComboNo].Select, TRUE)) {
			Joypadcpy[ComboNo].Select = KeyCode;
			SetFocus(hEdit_Start);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_Start == hFocus) {
		if(SetJoyButtonText(hEdit_Start, KeyCode, Joypadcpy[ComboNo].Start, TRUE)) {
			Joypadcpy[ComboNo].Start = KeyCode;
			SetFocus(hEdit_L);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_L == hFocus) {
		if(SetJoyButtonText(hEdit_L, KeyCode, Joypadcpy[ComboNo].L, TRUE)) {
			Joypadcpy[ComboNo].L = KeyCode;
			SetFocus(hEdit_R);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_R == hFocus) {
		if(SetJoyButtonText(hEdit_R, KeyCode, Joypadcpy[ComboNo].R, TRUE)) {
			Joypadcpy[ComboNo].R = KeyCode;
			SetFocus(GetDlgItem(hProp, ID_APPLY_NOW));
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc0 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc0, KeyCode, (WORD &)MapFunc_cpy[0], TRUE)) {
			MapFunc_cpy[0] = (MapFunc_cpy[0] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc1);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc1 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc1, KeyCode, (WORD &)MapFunc_cpy[1], TRUE)) {
			MapFunc_cpy[1] = (MapFunc_cpy[1] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc2);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc2 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc2, KeyCode, (WORD &)MapFunc_cpy[2], TRUE)) {
			MapFunc_cpy[2] = (MapFunc_cpy[2] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc3);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc3 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc3, KeyCode, (WORD &)MapFunc_cpy[3], TRUE)) {
			MapFunc_cpy[3] = (MapFunc_cpy[3] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc4);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc4 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc4, KeyCode, (WORD &)MapFunc_cpy[4], TRUE)) {
			MapFunc_cpy[4] = (MapFunc_cpy[4] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc5);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc5 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc5, KeyCode, (WORD &)MapFunc_cpy[5], TRUE)) {
			MapFunc_cpy[5] = (MapFunc_cpy[5] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc6);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc6 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc6, KeyCode, (WORD &)MapFunc_cpy[6], TRUE)) {
			MapFunc_cpy[6] = (MapFunc_cpy[6] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc7);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc7 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc7, KeyCode, (WORD &)MapFunc_cpy[7], TRUE)) {
			MapFunc_cpy[7] = (MapFunc_cpy[7] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc8);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc8 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc8, KeyCode, (WORD &)MapFunc_cpy[8], TRUE)) {
			MapFunc_cpy[8] = (MapFunc_cpy[8] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc9);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc9 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc9, KeyCode, (WORD &)MapFunc_cpy[9], TRUE)) {
			MapFunc_cpy[9] = (MapFunc_cpy[9] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc10);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc10 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc10, KeyCode, (WORD &)MapFunc_cpy[10], TRUE)) {
			MapFunc_cpy[10] = (MapFunc_cpy[10] & 0xffff0000) | KeyCode;
			SetFocus(hEdit_MapFunc11);
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	if(hEdit_MapFunc11 == hFocus) {
		if(SetJoyButtonText(hEdit_MapFunc11, KeyCode, (WORD &)MapFunc_cpy[11], TRUE)) {
			MapFunc_cpy[11] = (MapFunc_cpy[11] & 0xffff0000) | KeyCode;
			SetFocus(GetDlgItem(hProp, ID_APPLY_NOW));
			ConfigDlgSettingsChanged(GetParent(hFocus));
		}
		return TRUE;
	}
	return FALSE;
}

static LRESULT CALLBACK EditGetKeyProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	hFocus = GetFocus();
	switch (uMsg) {

		case WM_KEYDOWN:
			if(!SetJoyButtonCode(hFocus, LOWORD(wParam)))
				break;
			return 0;

		case WM_SYSKEYDOWN:
			if(!SetJoyButtonCode(hFocus, LOWORD(wParam)))
				break;
			return 0;

		default:
			break;

	}
	if(hWnd == hEdit_Left)
		return (CallWindowProc(Edit_Left, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_Right)
		return (CallWindowProc(Edit_Right, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_Up)
		return (CallWindowProc(Edit_Up, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_Down)
		return (CallWindowProc(Edit_Down, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_Start)
		return (CallWindowProc(Edit_Start, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_Select)
		return (CallWindowProc(Edit_Select, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_A)
		return (CallWindowProc(Edit_A, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_B)
		return (CallWindowProc(Edit_B, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_X)
		return (CallWindowProc(Edit_X, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_Y)
		return (CallWindowProc(Edit_Y, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_L)
		return (CallWindowProc(Edit_L, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_R)
		return (CallWindowProc(Edit_R, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc0)
		return (CallWindowProc(Edit_MapFunc0, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc1)
		return (CallWindowProc(Edit_MapFunc1, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc2)
		return (CallWindowProc(Edit_MapFunc2, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc3)
		return (CallWindowProc(Edit_MapFunc3, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc4)
		return (CallWindowProc(Edit_MapFunc4, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc5)
		return (CallWindowProc(Edit_MapFunc5, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc6)
		return (CallWindowProc(Edit_MapFunc6, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc7)
		return (CallWindowProc(Edit_MapFunc7, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc8)
		return (CallWindowProc(Edit_MapFunc8, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc9)
		return (CallWindowProc(Edit_MapFunc9, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc10)
		return (CallWindowProc(Edit_MapFunc10, hWnd, uMsg, wParam, lParam));
	if(hWnd == hEdit_MapFunc11)
		return (CallWindowProc(Edit_MapFunc11, hWnd, uMsg, wParam, lParam));
	return 0;
}

static void JoyDlgCtlSet(HWND hWnd)
{
	//`FbNt
	SendMessage(GetDlgItem(hWnd, IDC_JOYENABLE_CHECK), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Enabled ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_LEFT), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_Left ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_RIGHT), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_Right ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_UP), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_Up ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_DOWN), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_Down ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_START), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_Start ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_SELECT), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_Select ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_A), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_A ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_B), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_B ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_X), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_X ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_Y), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_Y ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_L), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_L ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_TURBO_R), BM_SETCHECK, (WPARAM)(Joypadcpy[ComboNo].Turbo_R ? 1 : 0), 0);

	//GfBbgEChEɃL[{^eLXg\
	SetJoyButtonText(hEdit_Left, Joypadcpy[ComboNo].Left, Joypadcpy[ComboNo].Left, FALSE);
	SetJoyButtonText(hEdit_Right, Joypadcpy[ComboNo].Right, Joypadcpy[ComboNo].Right, FALSE);
	SetJoyButtonText(hEdit_Up, Joypadcpy[ComboNo].Up, Joypadcpy[ComboNo].Up, FALSE);
	SetJoyButtonText(hEdit_Down, Joypadcpy[ComboNo].Down, Joypadcpy[ComboNo].Down, FALSE);
	SetJoyButtonText(hEdit_Start, Joypadcpy[ComboNo].Start, Joypadcpy[ComboNo].Start, FALSE);
	SetJoyButtonText(hEdit_Select, Joypadcpy[ComboNo].Select, Joypadcpy[ComboNo].Select, FALSE);
	SetJoyButtonText(hEdit_A, Joypadcpy[ComboNo].A, Joypadcpy[ComboNo].A, FALSE);
	SetJoyButtonText(hEdit_B, Joypadcpy[ComboNo].B, Joypadcpy[ComboNo].B, FALSE);
	SetJoyButtonText(hEdit_X, Joypadcpy[ComboNo].X, Joypadcpy[ComboNo].X, FALSE);
	SetJoyButtonText(hEdit_Y, Joypadcpy[ComboNo].Y, Joypadcpy[ComboNo].Y, FALSE);
	SetJoyButtonText(hEdit_L, Joypadcpy[ComboNo].L, Joypadcpy[ComboNo].L, FALSE);
	SetJoyButtonText(hEdit_R, Joypadcpy[ComboNo].R, Joypadcpy[ComboNo].R, FALSE);

	//@\蓖ăGfBbgEChEɃL[{^eLXg\
	SetJoyButtonText(hEdit_MapFunc0, (WORD &)MapFunc_cpy[0], (WORD &)MapFunc_cpy[0], FALSE);
	SetJoyButtonText(hEdit_MapFunc1, (WORD &)MapFunc_cpy[1], (WORD &)MapFunc_cpy[1], FALSE);
	SetJoyButtonText(hEdit_MapFunc2, (WORD &)MapFunc_cpy[2], (WORD &)MapFunc_cpy[2], FALSE);
	SetJoyButtonText(hEdit_MapFunc3, (WORD &)MapFunc_cpy[3], (WORD &)MapFunc_cpy[3], FALSE);
	SetJoyButtonText(hEdit_MapFunc4, (WORD &)MapFunc_cpy[4], (WORD &)MapFunc_cpy[4], FALSE);
	SetJoyButtonText(hEdit_MapFunc5, (WORD &)MapFunc_cpy[5], (WORD &)MapFunc_cpy[5], FALSE);
	SetJoyButtonText(hEdit_MapFunc6, (WORD &)MapFunc_cpy[6], (WORD &)MapFunc_cpy[6], FALSE);
	SetJoyButtonText(hEdit_MapFunc7, (WORD &)MapFunc_cpy[7], (WORD &)MapFunc_cpy[7], FALSE);
	SetJoyButtonText(hEdit_MapFunc8, (WORD &)MapFunc_cpy[8], (WORD &)MapFunc_cpy[8], FALSE);
	SetJoyButtonText(hEdit_MapFunc9, (WORD &)MapFunc_cpy[9], (WORD &)MapFunc_cpy[9], FALSE);
	SetJoyButtonText(hEdit_MapFunc10, (WORD &)MapFunc_cpy[10], (WORD &)MapFunc_cpy[10], FALSE);
	SetJoyButtonText(hEdit_MapFunc11, (WORD &)MapFunc_cpy[11], (WORD &)MapFunc_cpy[11], FALSE);

// Snes9xpp Original
	SendMessage(GetDlgItem(hWnd, IDC_EDIT_MAXWINJOYBUTTONS), EM_SETLIMITTEXT, 6, 0);
	wsprintf(tmpcharbuf, "%d", Options_MaxWinJoyButtons_cpy);
	SetWindowText(GetDlgItem(hWnd, IDC_EDIT_MAXWINJOYBUTTONS), tmpcharbuf);

}

// Snes9xpp Original
static WNDPROC Edit_MaxWinJoyButtons;
static HWND hEdit_MaxWinJoyButtons;


static LRESULT CALLBACK JoyEditSubProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_CHAR:
			if((LOWORD(wParam) >= '0' && LOWORD(wParam) <= '9') || LOWORD(wParam) == VK_BACK)
				break;
			return 0;

		default:
			break;

	}
// Snes9xpp Original
	if(hWnd == hEdit_MaxWinJoyButtons)
		return (CallWindowProc(Edit_MaxWinJoyButtons, hWnd, uMsg, wParam, lParam));
	return 0;
}

static void SetSubClass(HWND hWnd)
{
	Edit_Left = (WNDPROC)GetWindowLong(hEdit_Left, GWL_WNDPROC);
	SetWindowLong(hEdit_Left, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_Right = (WNDPROC)GetWindowLong(hEdit_Right, GWL_WNDPROC);
	SetWindowLong(hEdit_Right, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_Up = (WNDPROC)GetWindowLong(hEdit_Up, GWL_WNDPROC);
	SetWindowLong(hEdit_Up, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_Down = (WNDPROC)GetWindowLong(hEdit_Down, GWL_WNDPROC);
	SetWindowLong(hEdit_Down, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_Start = (WNDPROC)GetWindowLong(hEdit_Start, GWL_WNDPROC);
	SetWindowLong(hEdit_Start, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_Select = (WNDPROC)GetWindowLong(hEdit_Select, GWL_WNDPROC);
	SetWindowLong(hEdit_Select, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_A = (WNDPROC)GetWindowLong(hEdit_A, GWL_WNDPROC);
	SetWindowLong(hEdit_A, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_B = (WNDPROC)GetWindowLong(hEdit_B, GWL_WNDPROC);
	SetWindowLong(hEdit_B, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_X = (WNDPROC)GetWindowLong(hEdit_X, GWL_WNDPROC);
	SetWindowLong(hEdit_X, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_Y = (WNDPROC)GetWindowLong(hEdit_Y, GWL_WNDPROC);
	SetWindowLong(hEdit_Y, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_L = (WNDPROC)GetWindowLong(hEdit_L, GWL_WNDPROC);
	SetWindowLong(hEdit_L, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_R = (WNDPROC)GetWindowLong(hEdit_R, GWL_WNDPROC);
	SetWindowLong(hEdit_R, GWL_WNDPROC, (LONG)EditGetKeyProc);

	Edit_MapFunc0 = (WNDPROC)GetWindowLong(hEdit_MapFunc0, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc0, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc1 = (WNDPROC)GetWindowLong(hEdit_MapFunc1, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc1, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc2 = (WNDPROC)GetWindowLong(hEdit_MapFunc2, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc2, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc3 = (WNDPROC)GetWindowLong(hEdit_MapFunc3, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc3, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc4 = (WNDPROC)GetWindowLong(hEdit_MapFunc4, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc4, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc5 = (WNDPROC)GetWindowLong(hEdit_MapFunc5, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc5, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc6 = (WNDPROC)GetWindowLong(hEdit_MapFunc6, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc6, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc7 = (WNDPROC)GetWindowLong(hEdit_MapFunc7, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc7, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc8 = (WNDPROC)GetWindowLong(hEdit_MapFunc8, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc8, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc9 = (WNDPROC)GetWindowLong(hEdit_MapFunc9, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc9, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc10 = (WNDPROC)GetWindowLong(hEdit_MapFunc10, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc10, GWL_WNDPROC, (LONG)EditGetKeyProc);
	Edit_MapFunc11 = (WNDPROC)GetWindowLong(hEdit_MapFunc11, GWL_WNDPROC);
	SetWindowLong(hEdit_MapFunc11, GWL_WNDPROC, (LONG)EditGetKeyProc);

	SetWindowLong(hProp, GWL_WNDPROC, (LONG)SubPropProc);

 // Snes9xpp Original
	hEdit_MaxWinJoyButtons = GetDlgItem(hWnd, IDC_EDIT_MAXWINJOYBUTTONS);
	Edit_MaxWinJoyButtons = (WNDPROC)GetWindowLong(hEdit_MaxWinJoyButtons, GWL_WNDPROC);
	SetWindowLong(hEdit_MaxWinJoyButtons, GWL_WNDPROC, (LONG)JoyEditSubProc);

}

static void ReleaseSubClass(HWND hWnd)
{
	if(TimerActiveFlag) {
		KillTimer(hWnd, 100);
		TimerActiveFlag = 0;
	}
	if(SelectedEditBkGreen) {
		DeleteObject(SelectedEditBkGreen);
		SelectedEditBkGreen = 0;
	}
	if(SelectedEditBkOrange) {
		DeleteObject(SelectedEditBkOrange);
		SelectedEditBkOrange = 0;
	}
	SetWindowLong(hEdit_Left, GWL_WNDPROC, (LONG)Edit_Left);
	SetWindowLong(hEdit_Right, GWL_WNDPROC, (LONG)Edit_Right);
	SetWindowLong(hEdit_Up, GWL_WNDPROC, (LONG)Edit_Up);
	SetWindowLong(hEdit_Down, GWL_WNDPROC, (LONG)Edit_Down);
	SetWindowLong(hEdit_Start, GWL_WNDPROC, (LONG)Edit_Start);
	SetWindowLong(hEdit_Select, GWL_WNDPROC, (LONG)Edit_Select);
	SetWindowLong(hEdit_A, GWL_WNDPROC, (LONG)Edit_A);
	SetWindowLong(hEdit_B, GWL_WNDPROC, (LONG)Edit_B);
	SetWindowLong(hEdit_X, GWL_WNDPROC, (LONG)Edit_X);
	SetWindowLong(hEdit_Y, GWL_WNDPROC, (LONG)Edit_Y);
	SetWindowLong(hEdit_L, GWL_WNDPROC, (LONG)Edit_L);
	SetWindowLong(hEdit_R, GWL_WNDPROC, (LONG)Edit_R);

	SetWindowLong(hEdit_MapFunc0, GWL_WNDPROC, (LONG)Edit_MapFunc0);
	SetWindowLong(hEdit_MapFunc1, GWL_WNDPROC, (LONG)Edit_MapFunc1);
	SetWindowLong(hEdit_MapFunc2, GWL_WNDPROC, (LONG)Edit_MapFunc2);
	SetWindowLong(hEdit_MapFunc3, GWL_WNDPROC, (LONG)Edit_MapFunc3);
	SetWindowLong(hEdit_MapFunc4, GWL_WNDPROC, (LONG)Edit_MapFunc4);
	SetWindowLong(hEdit_MapFunc5, GWL_WNDPROC, (LONG)Edit_MapFunc5);
	SetWindowLong(hEdit_MapFunc6, GWL_WNDPROC, (LONG)Edit_MapFunc6);
	SetWindowLong(hEdit_MapFunc7, GWL_WNDPROC, (LONG)Edit_MapFunc7);
	SetWindowLong(hEdit_MapFunc8, GWL_WNDPROC, (LONG)Edit_MapFunc8);
	SetWindowLong(hEdit_MapFunc9, GWL_WNDPROC, (LONG)Edit_MapFunc9);
	SetWindowLong(hEdit_MapFunc10, GWL_WNDPROC, (LONG)Edit_MapFunc10);
	SetWindowLong(hEdit_MapFunc11, GWL_WNDPROC, (LONG)Edit_MapFunc11);

	SetWindowLong(hEdit_MaxWinJoyButtons, GWL_WNDPROC, (LONG)Edit_MaxWinJoyButtons); // Snes9xpp Original

	SetWindowLong(hProp, GWL_WNDPROC, (LONG)PropProc);
}

/***********************************************************************************************
 Dir Dialog
************************************************************************************************/ 

void DirectorySettingsLoad()
{
	lstrcpy(InitialROMFileDirectory_cpy, InitialROMFileDirectory);
	lstrcpy(GUI_FreezeFileDir_cpy, GUI.FreezeFileDir);
	lstrcpy(ScreenShotDirectory_cpy, ScreenShotDirectory);
	lstrcpy(SPCDumpDirectory_cpy, SPCDumpDirectory);
}

void DirectorySettingsSave()
{
	if(lstrcmp(InitialROMFileDirectory, InitialROMFileDirectory_cpy) && lstrlen(InitialROMFileDirectory_cpy) <= (_MAX_PATH - 7)) {
		lstrcpy(InitialROMFileDirectory, InitialROMFileDirectory_cpy);
		lstrcpy(ROMFileName, InitialROMFileDirectory_cpy);
		SlashRemove(ROMFileName);
		lstrcat(ROMFileName, "\\a.smc");
	}
	lstrcpy(GUI.FreezeFileDir, GUI_FreezeFileDir_cpy);
	lstrcpy(ScreenShotDirectory, ScreenShotDirectory_cpy);
	lstrcpy(SPCDumpDirectory, SPCDumpDirectory_cpy);
}

LRESULT CALLBACK DirectoryDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam)
{
	static char tmpstr[_MAX_PATH];
	switch (uMsg) {
		case WM_INITDIALOG:
			lstrcpy(tmpstr, InitialROMFileDirectory_cpy);
			SetWindowText(GetDlgItem(hWnd, IDC_DIR_ROM), tmpstr);
			lstrcpy(tmpstr, GUI_FreezeFileDir_cpy);
			SlashAdd(tmpstr);
			SetWindowText(GetDlgItem(hWnd, IDC_DIR_SAVE), tmpstr);
			lstrcpy(tmpstr, ScreenShotDirectory_cpy);
			SlashAdd(tmpstr);
			SetWindowText(GetDlgItem(hWnd, IDC_DIR_SCREEN), tmpstr);
			lstrcpy(tmpstr, SPCDumpDirectory_cpy);
			SlashAdd(tmpstr);
			SetWindowText(GetDlgItem(hWnd, IDC_DIR_SPC), tmpstr);
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDC_DIR_ROM:
				case IDC_DIR_SAVE:
				case IDC_DIR_SCREEN:
				case IDC_DIR_SPC:
					if(HIWORD(wParam) == EN_CHANGE) {
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				case IDC_BUTTON_DIR_ROM:
				case IDC_BUTTON_DIR_SAVE:
				case IDC_BUTTON_DIR_SCREEN:
				case IDC_BUTTON_DIR_SPC:
					{
						char BrowseDir[_MAX_PATH];
						char InitialBrowseDir[_MAX_PATH];
						int idctrl = IDC_DIR_ROM;
						BROWSEINFO bi;
						FillMemory(&bi, sizeof(BROWSEINFO), 0x00);

						switch(LOWORD(wParam))
						{
							case IDC_BUTTON_DIR_ROM:
								bi.lpszTitle = (GUI.Language ? "ǂݍ݃tH_I" : "Select ROM Folder");
								idctrl = IDC_DIR_ROM;
								lstrcpy(InitialBrowseDir, InitialROMFileDirectory_cpy);
								break;

							case IDC_BUTTON_DIR_SAVE:
								bi.lpszTitle = (GUI.Language ? "Z[ut@CۑtH_I" : "Select Save Folder");
								idctrl = IDC_DIR_SAVE;
								lstrcpy(InitialBrowseDir, GUI_FreezeFileDir_cpy);
								break;

							case IDC_BUTTON_DIR_SCREEN:
								bi.lpszTitle = (GUI.Language ? "XN[VbgۑtH_I" : "Select Screen shot Save Folder");
								idctrl = IDC_DIR_SCREEN;
								lstrcpy(InitialBrowseDir, ScreenShotDirectory_cpy);
								break;

							case IDC_BUTTON_DIR_SPC:
								bi.lpszTitle = (GUI.Language ? "SPCۑtH_I" : "Select SPC Save Folder");
								idctrl = IDC_DIR_SPC;
								lstrcpy(InitialBrowseDir, SPCDumpDirectory_cpy);
								break;
						}
						SlashAdd(InitialBrowseDir);
						bi.hwndOwner = hWnd;
						bi.ulFlags = BIF_RETURNONLYFSDIRS | BIF_DONTGOBELOWDOMAIN;
						LPITEMIDLIST lpiil;
						bi.lpfn = (BFFCALLBACK)BrowseHook;
						bi.lParam=(LPARAM)InitialBrowseDir;
						
						if((lpiil = SHBrowseForFolder(&bi)))
						{
							if( SHGetPathFromIDList( lpiil, BrowseDir) ) {
								SetWindowText(GetDlgItem(hWnd, idctrl), BrowseDir);
								//ROMǂݍݏfBNgȊOŌ̃pX؂폜
								if(idctrl != IDC_DIR_ROM)
									SlashRemove(BrowseDir);
								if(idctrl == IDC_DIR_ROM)
									lstrcpy(InitialROMFileDirectory_cpy, BrowseDir);
								else if(idctrl == IDC_DIR_SAVE)
									lstrcpy(GUI_FreezeFileDir_cpy, BrowseDir);
								else if(idctrl == IDC_DIR_SCREEN)
									lstrcpy(ScreenShotDirectory_cpy, BrowseDir);
								else if(idctrl == IDC_DIR_SPC)
									lstrcpy(SPCDumpDirectory_cpy, BrowseDir);
							}
							CoTaskMemFree(lpiil );
						}
						break;
					}
					break;

				default:
					break;
			}
			break;

		case WM_CTLCOLORSTATIC:
			if((HWND)lParam == GetDlgItem(hWnd, IDC_DIR_ROM) || (HWND)lParam == GetDlgItem(hWnd, IDC_DIR_SAVE)
				|| (HWND)lParam == GetDlgItem(hWnd, IDC_DIR_SCREEN) || (HWND)lParam == GetDlgItem(hWnd, IDC_DIR_SPC))
				return (long)GetStockObject(WHITE_BRUSH);
			break;

		case WM_NOTIFY:
			switch(((NMHDR *)lParam)->code) {
				case PSN_APPLY:
					ConfigChangePageIndex |= CONFIG_PAGE_DIRSET;
					ConfigDlgSettingsSave();
					break;

				case PSN_HELP:
					ConfigDlgDefaultSettingsLoad(hWnd);
					break;

				case PSN_QUERYCANCEL:
					break;

				case PSN_SETACTIVE:
					ConfigChangePageIndex |= CONFIG_PAGE_DIRSET;
					StartPageIndex = 3;
					break;

				default:
					break;
			}
			break;

		default:
			break;
	}
	return FALSE;
}

LRESULT CALLBACK BrowseHook(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam)
{
//	static DWORD Style;
	switch (uMsg) {
		case BFFM_INITIALIZED:
			SendMessage(hWnd,BFFM_SETSELECTION,(WPARAM)TRUE,lParam);
#if 0
			Style = GetWindowLong(hWnd, GWL_EXSTYLE );
			Style &= ~(WS_EX_CONTEXTHELP);
			SetWindowLong(hWnd, GWL_EXSTYLE, Style);
#endif
			break;

		default:
			break;
	}
	return FALSE;
}

/***********************************************************************************************
 Other Dialog
************************************************************************************************/ 

void OtherSettingsLoad()
{
	Options_SaveImagefileFormat_cpy = Options.SaveImagefileFormat;
	Options_ReduceTheColorTo256_cpy = Options.ReduceTheColorTo256;
	Options_ReduceTheColorTo256IfWithin_cpy = Options.ReduceTheColorTo256IfWithin;
	Options_ImagefileHiresStretch_cpy = Options.ImagefileHiresStretch;
	Options_RLEcompression256bitmap_cpy = Options.RLEcompression256bitmap;
	Options_DisplayScreenBufferAll_cpy = Options.DisplayScreenBufferAll;

	Xppcdda_Enable_cpy = Xppcdda.Enable; // Snes9xpp Original
	Xppcdda_Drv_cpy = Xppcdda.Drv; // Snes9xpp Original
	Xppsystem_NormalSnap_cpy = Xppsystem.NormalSnap; // Snes9xpp Original
	Settings_SDD1Pack_cpy = Settings.SDD1Pack; // Snes9xpp Original
	Xppsystem_DisableIRQCycleCount_cpy = Xppsystem.DisableIRQCycleCount; // Snes9xpp Original
	Xppsystem_OldAutoFrame_cpy = Xppsystem.OldAutoFrame; // Snes9xpp Original
	Xppsystem_OldFrameTimer_cpy = Xppsystem.OldFrameTimer; // Snes9xpp Original
	Xppsystem_FontType_cpy = Xppsystem.FontType; // Snes9xpp Original
	Xppsystem_InactivateWork_cpy = Xppsystem.InactivateWork; // Snes9xpp Original
}

static HWND hOtherDlg;

void OtherDefaultSettingsLoad()
{
	Options_SaveImagefileFormat_cpy = OPTIONS_SAVEIMAGEFILEFORMAT_DEF;
	Options_ReduceTheColorTo256_cpy = OPTIONS_REDUCETHECOLORTO256_DEF;
	Options_ReduceTheColorTo256IfWithin_cpy = OPTIONS_REDUCETHECOLORTO256IFWITHIN_DEF;
	Options_ImagefileHiresStretch_cpy = OPTIONS_IMAGEFILEHIRESSTRETCH_DEF;
	Options_RLEcompression256bitmap_cpy = OPTIONS_RLECOMPRESSION256BITMAP_DEF;
	Options_DisplayScreenBufferAll_cpy = OPTIONS_DISPLAYSCREENBUFFERALL_DEF;

	Xppcdda_Enable_cpy = OPTIONS_XPP_CDDAENABLE; // Snes9xpp Original
	Xppcdda_Drv_cpy = OPTIONS_XPP_CDDADRV; // Snes9xpp Original
	Xppsystem_NormalSnap_cpy = OPTIONS_XPP_NORMALSNAP; // Snes9xpp Original
	Settings_SDD1Pack_cpy = OPTIONS_SDD1PACK; // Snes9xpp Original
	Xppsystem_DisableIRQCycleCount_cpy = OPTIONS_DISABLE_IRQ_CYCLECOUNT; // Snes9xpp Original

	Settings_SDD1Pack_cpy = Settings_SDD1Pack_def; // Snes9xpp Original
	Xppsystem_DisableIRQCycleCount_cpy = Xppsystem_DisableIRQCycleCount_def;
	Xppsystem_OldAutoFrame_cpy = Xppsystem_OldAutoFrame_def;
	Xppsystem_OldFrameTimer_cpy = Xppsystem_OldFrameTimer_def;
	Xppsystem_FontType_cpy = Xppsystem_FontType_def;
	Xppsystem_InactivateWork_cpy = Xppsystem_InactivateWork_def;

	OtherDlgCtrlSet(hOtherDlg);
}

void OtherSettingsSave()
{
	Options.SaveImagefileFormat = Options_SaveImagefileFormat_cpy;
//	Options.ReduceTheColorTo256 = Options_ReduceTheColorTo256_cpy;
	Options.ReduceTheColorTo256IfWithin = Options_ReduceTheColorTo256IfWithin_cpy;
	Options.ImagefileHiresStretch = Options_ImagefileHiresStretch_cpy;
	Options.RLEcompression256bitmap = Options_RLEcompression256bitmap_cpy;
	Options.DisplayScreenBufferAll = Options_DisplayScreenBufferAll_cpy;

// Snes9xpp Original
       if(Xppcdda_Drv_cpy != Xppcdda.Drv)
       {
          cdda_exit();
	   Xppcdda.Drv = Xppcdda_Drv_cpy;
          cdda_open();
	   cdda_stop();
       }

       if(Xppcdda_Enable_cpy != Xppcdda.Enable)
       {
          if(Xppcdda.Enable)
          cdda_exit();
          else
          {
              cdda_open();
		cdda_stop();
          }
	   Xppcdda.Enable = Xppcdda_Enable_cpy;
       }

{
	int NormalSnap_Old;
	NormalSnap_Old = Xppsystem.NormalSnap; // Snes9xpp Original
	Xppsystem.NormalSnap = Xppsystem_NormalSnap_cpy; // Snes9xpp Original
	Settings.SDD1Pack = Settings_SDD1Pack_cpy; // Snes9xpp Original
	Xppsystem.DisableIRQCycleCount = Xppsystem_DisableIRQCycleCount_cpy; // Snes9xpp Original
	Xppsystem.OldAutoFrame = Xppsystem_OldAutoFrame_cpy; // Snes9xpp Original
	Xppsystem.OldFrameTimer = Xppsystem_OldFrameTimer_cpy; // Snes9xpp Original
	Xppsystem.FontType = Xppsystem_FontType_cpy;
	Xppsystem.InactivateWork = Xppsystem_InactivateWork_cpy;

	if(NormalSnap_Old != Xppsystem.NormalSnap &&
	 Xppsystem.NormalSnap != Settings.Normalsnap &&
	 NormalSnap_Old == Settings.Normalsnap)
	 Settings.Normalsnap = Xppsystem.NormalSnap;   // Snes9xpp Original
}

 // Snes9xpp Original
}

LRESULT CALLBACK OtherDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam)
{
	static char *combostrCdDrvFunc[] = {"A:",
										"B:",
										"C:",
										"D:",
										"E:",
										"F:",
										"G:",
										"H:",
										"I:",
										"J:",
										"K:",
										"L:",
										"M:",
										"N:",
										"O:",
										"P:",
										"Q:",
										"R:",
										"S:",
										"T:",
										"U:",
										"V:",
										"W:",
										"X:",
										"Y:",
										"Z:"};

	static int i;

	switch (uMsg) {
		case WM_INITDIALOG:
			hOtherDlg = hWnd;
// Snes9xpp Original
			for (i = 0; i < 26; i++ ) {
				SendMessage(GetDlgItem(hWnd, IDC_CDDRV_COMBO), CB_INSERTSTRING, (WPARAM)i, (LPARAM)(combostrCdDrvFunc[i]));
			}
// Snes9xpp Original
			OtherDlgCtrlSet(hWnd);
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDC_RADIO_SAVEIMAGEFILEFORMAT_BMP:
				case IDC_RADIO_SAVEIMAGEFILEFORMAT_PCX:
				case IDC_RADIO_SAVEIMAGEFILEFORMAT_GIF:
				case IDC_RADIO_SAVEIMAGEFILEFORMAT_PNG:
				case IDC_CHECK_REDUCETHECOLORTO256IFWITHIN:
				case IDC_CHECK_REDUCETHECOLORTO256:
				case IDC_CHECK_HIRESSTRETCH:
				case IDC_CHECK_RLECOMPRESSION256BITMAP:
				case IDC_CHECK_DISPLAYSCREENBUFFERALL:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
						if(LOWORD(wParam) == IDC_RADIO_SAVEIMAGEFILEFORMAT_BMP ||
								LOWORD(wParam) == IDC_RADIO_SAVEIMAGEFILEFORMAT_PCX ||
								LOWORD(wParam) == IDC_RADIO_SAVEIMAGEFILEFORMAT_GIF ||
								LOWORD(wParam) == IDC_RADIO_SAVEIMAGEFILEFORMAT_PNG) {
							if(LOWORD(wParam) == IDC_RADIO_SAVEIMAGEFILEFORMAT_BMP) {
								Options_SaveImagefileFormat_cpy = 0;
								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256IFWITHIN), TRUE);
								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_RLECOMPRESSION256BITMAP), TRUE);
//								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256), TRUE);
							}
							else if(LOWORD(wParam) == IDC_RADIO_SAVEIMAGEFILEFORMAT_PCX) {
								Options_SaveImagefileFormat_cpy = 1;
								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256IFWITHIN), TRUE);
								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_RLECOMPRESSION256BITMAP), FALSE);
//								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256), TRUE);
							}
							else if(LOWORD(wParam) == IDC_RADIO_SAVEIMAGEFILEFORMAT_GIF) {
								Options_SaveImagefileFormat_cpy = 2;
								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256IFWITHIN), FALSE);
								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_RLECOMPRESSION256BITMAP), FALSE);
//								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256), FALSE);
							}
							else if(LOWORD(wParam) == IDC_RADIO_SAVEIMAGEFILEFORMAT_PNG) {
								Options_SaveImagefileFormat_cpy = 3;
								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256IFWITHIN), TRUE);
								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_RLECOMPRESSION256BITMAP), FALSE);
//								EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256), TRUE);
							}
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_SAVEIMAGEFILEFORMAT_BMP), BM_SETCHECK, (WPARAM)(Options_SaveImagefileFormat_cpy == 0 ? 1 : 0), 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_SAVEIMAGEFILEFORMAT_PCX), BM_SETCHECK, (WPARAM)(Options_SaveImagefileFormat_cpy == 1 ? 1 : 0), 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_SAVEIMAGEFILEFORMAT_GIF), BM_SETCHECK, (WPARAM)(Options_SaveImagefileFormat_cpy == 2 ? 1 : 0), 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_SAVEIMAGEFILEFORMAT_PNG), BM_SETCHECK, (WPARAM)(Options_SaveImagefileFormat_cpy == 3 ? 1 : 0), 0);
							if(Options.SaveImagefileFormat == Options_SaveImagefileFormat_cpy)
								break;
						}
						else if(LOWORD(wParam) == IDC_CHECK_REDUCETHECOLORTO256)
							Options_ReduceTheColorTo256_cpy = tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_REDUCETHECOLORTO256IFWITHIN)
							Options_ReduceTheColorTo256IfWithin_cpy = tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_HIRESSTRETCH)
							Options_ImagefileHiresStretch_cpy = tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_RLECOMPRESSION256BITMAP)
							Options_RLEcompression256bitmap_cpy = tmpVal;
						else if(LOWORD(wParam) == IDC_CHECK_DISPLAYSCREENBUFFERALL)
							Options_DisplayScreenBufferAll_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;
// Snes9xpp Original
				case IDC_CHECK_USECDDA:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
                                                if(LOWORD(wParam) == IDC_CHECK_USECDDA)
							Xppcdda_Enable_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				case IDC_CDDRV_COMBO:
					if(HIWORD(wParam) == CBN_SELENDOK) {
						if(LOWORD(wParam) == IDC_CDDRV_COMBO)
						       i = (int)SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0);
							Xppcdda_Drv_cpy = (uint8)i;
 						if(Xppcdda_Drv_cpy != Xppcdda.Drv)
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				case IDC_CHECK_NORMALSNAP:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
                                                if(LOWORD(wParam) == IDC_CHECK_NORMALSNAP)
							Xppsystem_NormalSnap_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;
				case IDC_CHECK_SDD1PACK:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
                                                if(LOWORD(wParam) == IDC_CHECK_SDD1PACK)
							Settings_SDD1Pack_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;
				case IDC_CHECK_DISABLE_IRQ_CYCLECOUNT:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
                                                if(LOWORD(wParam) == IDC_CHECK_DISABLE_IRQ_CYCLECOUNT)
							Xppsystem_DisableIRQCycleCount_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;
				case IDC_CHECK_OLD_AUTOFRAME:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
                                                if(LOWORD(wParam) == IDC_CHECK_OLD_AUTOFRAME)
							Xppsystem_OldAutoFrame_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;
				case IDC_CHECK_OLD_FRAMETIMER:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
                                                if(LOWORD(wParam) == IDC_CHECK_OLD_FRAMETIMER)
							Xppsystem_OldFrameTimer_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;
				case IDC_CHECK_FONTTYPE:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
                                                if(LOWORD(wParam) == IDC_CHECK_FONTTYPE)
							Xppsystem_FontType_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;
				case IDC_CHECK_INACTIVATEWORK:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
                                                if(LOWORD(wParam) == IDC_CHECK_INACTIVATEWORK)
							Xppsystem_InactivateWork_cpy = tmpVal;
						ConfigDlgSettingsChanged(hWnd);
					}
					break;
// Snes9xpp Original
					
				default:
					break;
			}
			break;

		case WM_NOTIFY:
			switch(((NMHDR *)lParam)->code) {
				case PSN_APPLY:
					ConfigChangePageIndex |= CONFIG_PAGE_OTHER;
					ConfigDlgSettingsSave();
					break;

				case PSN_HELP:
					ConfigDlgDefaultSettingsLoad(hWnd);
					break;

				case PSN_QUERYCANCEL:
					break;

				case PSN_SETACTIVE:
					ConfigChangePageIndex |= CONFIG_PAGE_OTHER;
					EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256IFWITHIN), Options_SaveImagefileFormat_cpy != 2 ? TRUE : FALSE);
//					EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256), Options_SaveImagefileFormat_cpy != 2 ? TRUE : FALSE);
					StartPageIndex = 4;
					
					break;

				default:
					break;
			}
			break;

		default:
			break;
	}
	return FALSE;
}

static void OtherDlgCtrlSet(HWND hWnd)
{
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_SAVEIMAGEFILEFORMAT_BMP), BM_SETCHECK, (WPARAM)(Options_SaveImagefileFormat_cpy == 0 ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_SAVEIMAGEFILEFORMAT_PCX), BM_SETCHECK, (WPARAM)(Options_SaveImagefileFormat_cpy == 1 ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_SAVEIMAGEFILEFORMAT_GIF), BM_SETCHECK, (WPARAM)(Options_SaveImagefileFormat_cpy == 2 ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_SAVEIMAGEFILEFORMAT_PNG), BM_SETCHECK, (WPARAM)(Options_SaveImagefileFormat_cpy == 3 ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256), BM_SETCHECK, (WPARAM)(Options_ReduceTheColorTo256_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256IFWITHIN), BM_SETCHECK, (WPARAM)(Options_ReduceTheColorTo256IfWithin_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_HIRESSTRETCH), BM_SETCHECK, (WPARAM)(Options_ImagefileHiresStretch_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_RLECOMPRESSION256BITMAP), BM_SETCHECK, (WPARAM)(Options_RLEcompression256bitmap_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_DISPLAYSCREENBUFFERALL), BM_SETCHECK, (WPARAM)(Options_DisplayScreenBufferAll_cpy ? 1 : 0), 0);

// Snes9xpp Original
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_USECDDA), BM_SETCHECK, (WPARAM)(Xppcdda_Enable_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CDDRV_COMBO), CB_SETCURSEL, (WPARAM)Xppcdda.Drv, 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_NORMALSNAP), BM_SETCHECK, (WPARAM)(Xppsystem_NormalSnap_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_SDD1PACK), BM_SETCHECK, (WPARAM)(Settings_SDD1Pack_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_DISABLE_IRQ_CYCLECOUNT), BM_SETCHECK, (WPARAM)(Xppsystem_DisableIRQCycleCount_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_OLD_AUTOFRAME), BM_SETCHECK, (WPARAM)(Xppsystem_OldAutoFrame_cpy ? 1 : 0), 0);

	SendMessage(GetDlgItem(hWnd, IDC_CHECK_OLD_FRAMETIMER), BM_SETCHECK, (WPARAM)(Xppsystem_OldFrameTimer_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_FONTTYPE), BM_SETCHECK, (WPARAM)(Xppsystem_FontType_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_INACTIVATEWORK), BM_SETCHECK, (WPARAM)(Xppsystem_InactivateWork_cpy ? 1 : 0), 0);

    EnableWindow(GetDlgItem(hWnd, IDC_CHECK_DISABLE_IRQ_CYCLECOUNT), (GUI.StopEmulation == TRUE) ? TRUE : FALSE);
    EnableWindow(GetDlgItem(hWnd, IDC_CHECK_OLD_FRAMETIMER), (GUI.StopEmulation == TRUE) ? TRUE : FALSE);
// Snes9xpp Original

	EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256IFWITHIN), Options_SaveImagefileFormat_cpy != 2 ? TRUE : FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256), Options_SaveImagefileFormat_cpy != 2 ? TRUE : FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_CHECK_RLECOMPRESSION256BITMAP), (Options_SaveImagefileFormat_cpy == 0) ? TRUE : FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_CHECK_DISPLAYSCREENBUFFERALL), TRUE);

	//not supported yet
	EnableWindow(GetDlgItem(hWnd, IDC_CHECK_REDUCETHECOLORTO256), FALSE);
}

